(*
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0 OR ISC OR MIT-0
 *)

(* ========================================================================= *)
(* Scalar multiplication for NIST P-256.                                     *)
(* ========================================================================= *)

needs "arm/proofs/base.ml";;
needs "common/ecencoding.ml";;
needs "EC/jacobian.ml";;
needs "EC/nistp256.ml";;

prioritize_int();;
prioritize_real();;
prioritize_num();;

needs "arm/proofs/bignum_demont_p256.ml";;
needs "arm/proofs/bignum_inv_p256.ml";;
needs "arm/proofs/bignum_montmul_p256_alt.ml";;
needs "arm/proofs/bignum_montsqr_p256_alt.ml";;
needs "arm/proofs/bignum_tomont_p256.ml";;
needs "arm/proofs/p256_montjadd_alt.ml";;
needs "arm/proofs/p256_montjdouble_alt.ml";;
needs "arm/proofs/p256_montjmixadd_alt.ml";;

(* ------------------------------------------------------------------------- *)
(* Code.                                                                     *)
(* ------------------------------------------------------------------------- *)

(**** print_literal_from_elf "arm/p256/p256_scalarmul_alt.o";;
 ****)

let p256_scalarmul_alt_mc = define_assert_from_elf
  "p256_scalarmul_alt_mc" "arm/p256/p256_scalarmul_alt.o"
[
  0xa9bf53f3;       (* arm_STP X19 X20 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xa9bf7bf5;       (* arm_STP X21 X30 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xd10f83ff;       (* arm_SUB SP SP (rvalue (word 992)) *)
  0xaa0003f3;       (* arm_MOV X19 X0 *)
  0xaa0203f5;       (* arm_MOV X21 X2 *)
  0xd284aa2c;       (* arm_MOV X12 (rvalue (word 9553)) *)
  0xf2bf8c6c;       (* arm_MOVK X12 (word 64611) 16 *)
  0xf2d9584c;       (* arm_MOVK X12 (word 51906) 32 *)
  0xf2fe772c;       (* arm_MOVK X12 (word 62393) 48 *)
  0xd293d08d;       (* arm_MOV X13 (rvalue (word 40580)) *)
  0xf2b4e2ed;       (* arm_MOVK X13 (word 42775) 16 *)
  0xf2df55ad;       (* arm_MOVK X13 (word 64173) 32 *)
  0xf2f79ccd;       (* arm_MOVK X13 (word 48358) 48 *)
  0x9280000e;       (* arm_MOVN X14 (word 0) 0 *)
  0xb2607fef;       (* arm_MOV X15 (rvalue (word 18446744069414584320)) *)
  0xa9400c22;       (* arm_LDP X2 X3 X1 (Immediate_Offset (iword (&0))) *)
  0xa9411424;       (* arm_LDP X4 X5 X1 (Immediate_Offset (iword (&16))) *)
  0xeb0c0046;       (* arm_SUBS X6 X2 X12 *)
  0xfa0d0067;       (* arm_SBCS X7 X3 X13 *)
  0xfa0e0088;       (* arm_SBCS X8 X4 X14 *)
  0xfa0f00a9;       (* arm_SBCS X9 X5 X15 *)
  0x9a863042;       (* arm_CSEL X2 X2 X6 Condition_CC *)
  0x9a873063;       (* arm_CSEL X3 X3 X7 Condition_CC *)
  0x9a883084;       (* arm_CSEL X4 X4 X8 Condition_CC *)
  0x9a8930a5;       (* arm_CSEL X5 X5 X9 Condition_CC *)
  0xeb020186;       (* arm_SUBS X6 X12 X2 *)
  0xfa0301a7;       (* arm_SBCS X7 X13 X3 *)
  0xfa0401c8;       (* arm_SBCS X8 X14 X4 *)
  0xda0501e9;       (* arm_SBC X9 X15 X5 *)
  0xf24100bf;       (* arm_TST X5 (rvalue (word 9223372036854775808)) *)
  0x9a860042;       (* arm_CSEL X2 X2 X6 Condition_EQ *)
  0x9a870063;       (* arm_CSEL X3 X3 X7 Condition_EQ *)
  0x9a880084;       (* arm_CSEL X4 X4 X8 Condition_EQ *)
  0x9a8900a5;       (* arm_CSEL X5 X5 X9 Condition_EQ *)
  0x9a9f07f4;       (* arm_CSET X20 Condition_NE *)
  0xb201e3e6;       (* arm_MOV X6 (rvalue (word 9838263505978427528)) *)
  0xab060042;       (* arm_ADDS X2 X2 X6 *)
  0xba060063;       (* arm_ADCS X3 X3 X6 *)
  0x9240ecc7;       (* arm_AND X7 X6 (rvalue (word 1152921504606846975)) *)
  0xba060084;       (* arm_ADCS X4 X4 X6 *)
  0x9a0700a5;       (* arm_ADC X5 X5 X7 *)
  0xa9000fe2;       (* arm_STP X2 X3 SP (Immediate_Offset (iword (&0))) *)
  0xa90117e4;       (* arm_STP X4 X5 SP (Immediate_Offset (iword (&16))) *)
  0x910383e0;       (* arm_ADD X0 SP (rvalue (word 224)) *)
  0xaa1503e1;       (* arm_MOV X1 X21 *)
  0x94000701;       (* arm_BL (word 7172) *)
  0x910082a1;       (* arm_ADD X1 X21 (rvalue (word 32)) *)
  0x910403e0;       (* arm_ADD X0 SP (rvalue (word 256)) *)
  0x940006fe;       (* arm_BL (word 7160) *)
  0xd2800020;       (* arm_MOV X0 (rvalue (word 1)) *)
  0xb2607fe1;       (* arm_MOV X1 (rvalue (word 18446744069414584320)) *)
  0xa91207e0;       (* arm_STP X0 X1 SP (Immediate_Offset (iword (&288))) *)
  0x92800002;       (* arm_MOVN X2 (word 0) 0 *)
  0xb27f7be3;       (* arm_MOV X3 (rvalue (word 4294967294)) *)
  0xa9130fe2;       (* arm_STP X2 X3 SP (Immediate_Offset (iword (&304))) *)
  0xa95017e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&256))) *)
  0xa9511fe6;       (* arm_LDP X6 X7 SP (Immediate_Offset (iword (&272))) *)
  0x92800000;       (* arm_MOVN X0 (word 0) 0 *)
  0xeb040000;       (* arm_SUBS X0 X0 X4 *)
  0xb2407fe1;       (* arm_MOV X1 (rvalue (word 4294967295)) *)
  0xfa050021;       (* arm_SBCS X1 X1 X5 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0xfa0603e2;       (* arm_NGCS X2 X6 *)
  0xda070063;       (* arm_SBC X3 X3 X7 *)
  0xeb1f029f;       (* arm_CMP X20 XZR *)
  0x9a841004;       (* arm_CSEL X4 X0 X4 Condition_NE *)
  0x9a851025;       (* arm_CSEL X5 X1 X5 Condition_NE *)
  0x9a861046;       (* arm_CSEL X6 X2 X6 Condition_NE *)
  0x9a871067;       (* arm_CSEL X7 X3 X7 Condition_NE *)
  0xa91017e4;       (* arm_STP X4 X5 SP (Immediate_Offset (iword (&256))) *)
  0xa9111fe6;       (* arm_STP X6 X7 SP (Immediate_Offset (iword (&272))) *)
  0x910503e0;       (* arm_ADD X0 SP (rvalue (word 320)) *)
  0x910383e1;       (* arm_ADD X1 SP (rvalue (word 224)) *)
  0x94000ee9;       (* arm_BL (word 15268) *)
  0x910683e0;       (* arm_ADD X0 SP (rvalue (word 416)) *)
  0x910503e1;       (* arm_ADD X1 SP (rvalue (word 320)) *)
  0x910383e2;       (* arm_ADD X2 SP (rvalue (word 224)) *)
  0x940012f4;       (* arm_BL (word 19408) *)
  0x910803e0;       (* arm_ADD X0 SP (rvalue (word 512)) *)
  0x910503e1;       (* arm_ADD X1 SP (rvalue (word 320)) *)
  0x94000ee2;       (* arm_BL (word 15240) *)
  0x910983e0;       (* arm_ADD X0 SP (rvalue (word 608)) *)
  0x910803e1;       (* arm_ADD X1 SP (rvalue (word 512)) *)
  0x910383e2;       (* arm_ADD X2 SP (rvalue (word 224)) *)
  0x940012ed;       (* arm_BL (word 19380) *)
  0x910b03e0;       (* arm_ADD X0 SP (rvalue (word 704)) *)
  0x910683e1;       (* arm_ADD X1 SP (rvalue (word 416)) *)
  0x94000edb;       (* arm_BL (word 15212) *)
  0x910c83e0;       (* arm_ADD X0 SP (rvalue (word 800)) *)
  0x910b03e1;       (* arm_ADD X1 SP (rvalue (word 704)) *)
  0x910383e2;       (* arm_ADD X2 SP (rvalue (word 224)) *)
  0x940012e6;       (* arm_BL (word 19352) *)
  0x910e03e0;       (* arm_ADD X0 SP (rvalue (word 896)) *)
  0x910803e1;       (* arm_ADD X1 SP (rvalue (word 512)) *)
  0x94000ed4;       (* arm_BL (word 15184) *)
  0xf9400fee;       (* arm_LDR X14 SP (Immediate_Offset (word 24)) *)
  0xd37cfdce;       (* arm_LSR X14 X14 60 *)
  0xaa1f03e0;       (* arm_MOV X0 XZR *)
  0xaa1f03e1;       (* arm_MOV X1 XZR *)
  0xaa1f03e2;       (* arm_MOV X2 XZR *)
  0xaa1f03e3;       (* arm_MOV X3 XZR *)
  0xaa1f03e4;       (* arm_MOV X4 XZR *)
  0xaa1f03e5;       (* arm_MOV X5 XZR *)
  0xaa1f03e6;       (* arm_MOV X6 XZR *)
  0xaa1f03e7;       (* arm_MOV X7 XZR *)
  0xaa1f03e8;       (* arm_MOV X8 XZR *)
  0xaa1f03e9;       (* arm_MOV X9 XZR *)
  0xaa1f03ea;       (* arm_MOV X10 XZR *)
  0xaa1f03eb;       (* arm_MOV X11 XZR *)
  0x910383ef;       (* arm_ADD X15 SP (rvalue (word 224)) *)
  0xf10005df;       (* arm_CMP X14 (rvalue (word 1)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10009df;       (* arm_CMP X14 (rvalue (word 2)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf1000ddf;       (* arm_CMP X14 (rvalue (word 3)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10011df;       (* arm_CMP X14 (rvalue (word 4)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10015df;       (* arm_CMP X14 (rvalue (word 5)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10019df;       (* arm_CMP X14 (rvalue (word 6)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf1001ddf;       (* arm_CMP X14 (rvalue (word 7)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10021df;       (* arm_CMP X14 (rvalue (word 8)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xa90207e0;       (* arm_STP X0 X1 SP (Immediate_Offset (iword (&32))) *)
  0xa9030fe2;       (* arm_STP X2 X3 SP (Immediate_Offset (iword (&48))) *)
  0xa90417e4;       (* arm_STP X4 X5 SP (Immediate_Offset (iword (&64))) *)
  0xa9051fe6;       (* arm_STP X6 X7 SP (Immediate_Offset (iword (&80))) *)
  0xa90627e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&96))) *)
  0xa9072fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&112))) *)
  0xd2801f94;       (* arm_MOV X20 (rvalue (word 252)) *)
  0xd1001294;       (* arm_SUB X20 X20 (rvalue (word 4)) *)
  0x910083e0;       (* arm_ADD X0 SP (rvalue (word 32)) *)
  0x910083e1;       (* arm_ADD X1 SP (rvalue (word 32)) *)
  0x94000e1a;       (* arm_BL (word 14440) *)
  0x910083e0;       (* arm_ADD X0 SP (rvalue (word 32)) *)
  0x910083e1;       (* arm_ADD X1 SP (rvalue (word 32)) *)
  0x94000e17;       (* arm_BL (word 14428) *)
  0x910083e0;       (* arm_ADD X0 SP (rvalue (word 32)) *)
  0x910083e1;       (* arm_ADD X1 SP (rvalue (word 32)) *)
  0x94000e14;       (* arm_BL (word 14416) *)
  0x910083e0;       (* arm_ADD X0 SP (rvalue (word 32)) *)
  0x910083e1;       (* arm_ADD X1 SP (rvalue (word 32)) *)
  0x94000e11;       (* arm_BL (word 14404) *)
  0xd346fe82;       (* arm_LSR X2 X20 6 *)
  0xf8627bee;       (* arm_LDR X14 SP (Shiftreg_Offset X2 3) *)
  0x9ad425ce;       (* arm_LSRV X14 X14 X20 *)
  0x92400dce;       (* arm_AND X14 X14 (rvalue (word 15)) *)
  0xf10021ce;       (* arm_SUBS X14 X14 (rvalue (word 8)) *)
  0x9a9f27f0;       (* arm_CSET X16 Condition_CC *)
  0xda8e25ce;       (* arm_CNEG X14 X14 Condition_CC *)
  0xaa1f03e0;       (* arm_MOV X0 XZR *)
  0xaa1f03e1;       (* arm_MOV X1 XZR *)
  0xaa1f03e2;       (* arm_MOV X2 XZR *)
  0xaa1f03e3;       (* arm_MOV X3 XZR *)
  0xaa1f03e4;       (* arm_MOV X4 XZR *)
  0xaa1f03e5;       (* arm_MOV X5 XZR *)
  0xaa1f03e6;       (* arm_MOV X6 XZR *)
  0xaa1f03e7;       (* arm_MOV X7 XZR *)
  0xaa1f03e8;       (* arm_MOV X8 XZR *)
  0xaa1f03e9;       (* arm_MOV X9 XZR *)
  0xaa1f03ea;       (* arm_MOV X10 XZR *)
  0xaa1f03eb;       (* arm_MOV X11 XZR *)
  0x910383ef;       (* arm_ADD X15 SP (rvalue (word 224)) *)
  0xf10005df;       (* arm_CMP X14 (rvalue (word 1)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10009df;       (* arm_CMP X14 (rvalue (word 2)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf1000ddf;       (* arm_CMP X14 (rvalue (word 3)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10011df;       (* arm_CMP X14 (rvalue (word 4)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10015df;       (* arm_CMP X14 (rvalue (word 5)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10019df;       (* arm_CMP X14 (rvalue (word 6)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf1001ddf;       (* arm_CMP X14 (rvalue (word 7)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xf10021df;       (* arm_CMP X14 (rvalue (word 8)) *)
  0xa94035ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&0))) *)
  0x9a800180;       (* arm_CSEL X0 X12 X0 Condition_EQ *)
  0x9a8101a1;       (* arm_CSEL X1 X13 X1 Condition_EQ *)
  0xa94135ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&16))) *)
  0x9a820182;       (* arm_CSEL X2 X12 X2 Condition_EQ *)
  0x9a8301a3;       (* arm_CSEL X3 X13 X3 Condition_EQ *)
  0xa94235ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&32))) *)
  0x9a840184;       (* arm_CSEL X4 X12 X4 Condition_EQ *)
  0x9a8501a5;       (* arm_CSEL X5 X13 X5 Condition_EQ *)
  0xa94335ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&48))) *)
  0x9a860186;       (* arm_CSEL X6 X12 X6 Condition_EQ *)
  0x9a8701a7;       (* arm_CSEL X7 X13 X7 Condition_EQ *)
  0xa94435ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&64))) *)
  0x9a880188;       (* arm_CSEL X8 X12 X8 Condition_EQ *)
  0x9a8901a9;       (* arm_CSEL X9 X13 X9 Condition_EQ *)
  0xa94535ec;       (* arm_LDP X12 X13 X15 (Immediate_Offset (iword (&80))) *)
  0x9a8a018a;       (* arm_CSEL X10 X12 X10 Condition_EQ *)
  0x9a8b01ab;       (* arm_CSEL X11 X13 X11 Condition_EQ *)
  0x910181ef;       (* arm_ADD X15 X15 (rvalue (word 96)) *)
  0xa90807e0;       (* arm_STP X0 X1 SP (Immediate_Offset (iword (&128))) *)
  0xa9090fe2;       (* arm_STP X2 X3 SP (Immediate_Offset (iword (&144))) *)
  0x92800000;       (* arm_MOVN X0 (word 0) 0 *)
  0xeb040000;       (* arm_SUBS X0 X0 X4 *)
  0xb2407fe1;       (* arm_MOV X1 (rvalue (word 4294967295)) *)
  0xfa050021;       (* arm_SBCS X1 X1 X5 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0xfa0603e2;       (* arm_NGCS X2 X6 *)
  0xda070063;       (* arm_SBC X3 X3 X7 *)
  0xeb1f021f;       (* arm_CMP X16 XZR *)
  0x9a841004;       (* arm_CSEL X4 X0 X4 Condition_NE *)
  0x9a851025;       (* arm_CSEL X5 X1 X5 Condition_NE *)
  0x9a861046;       (* arm_CSEL X6 X2 X6 Condition_NE *)
  0x9a871067;       (* arm_CSEL X7 X3 X7 Condition_NE *)
  0xa90a17e4;       (* arm_STP X4 X5 SP (Immediate_Offset (iword (&160))) *)
  0xa90b1fe6;       (* arm_STP X6 X7 SP (Immediate_Offset (iword (&176))) *)
  0xa90c27e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&192))) *)
  0xa90d2fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&208))) *)
  0x910083e0;       (* arm_ADD X0 SP (rvalue (word 32)) *)
  0x910083e1;       (* arm_ADD X1 SP (rvalue (word 32)) *)
  0x910203e2;       (* arm_ADD X2 SP (rvalue (word 128)) *)
  0x940005bf;       (* arm_BL (word 5884) *)
  0xb5ffe534;       (* arm_CBNZ X20 (word 2096292) *)
  0x910383e0;       (* arm_ADD X0 SP (rvalue (word 224)) *)
  0x910183e1;       (* arm_ADD X1 SP (rvalue (word 96)) *)
  0x940004e3;       (* arm_BL (word 5004) *)
  0x910403e0;       (* arm_ADD X0 SP (rvalue (word 256)) *)
  0x910383e2;       (* arm_ADD X2 SP (rvalue (word 224)) *)
  0x910183e1;       (* arm_ADD X1 SP (rvalue (word 96)) *)
  0x9400046d;       (* arm_BL (word 4532) *)
  0x910383e0;       (* arm_ADD X0 SP (rvalue (word 224)) *)
  0x910403e1;       (* arm_ADD X1 SP (rvalue (word 256)) *)
  0x94000014;       (* arm_BL (word 80) *)
  0x910403e0;       (* arm_ADD X0 SP (rvalue (word 256)) *)
  0x910383e1;       (* arm_ADD X1 SP (rvalue (word 224)) *)
  0x94000036;       (* arm_BL (word 216) *)
  0x910383e0;       (* arm_ADD X0 SP (rvalue (word 224)) *)
  0x910403e2;       (* arm_ADD X2 SP (rvalue (word 256)) *)
  0x910183e1;       (* arm_ADD X1 SP (rvalue (word 96)) *)
  0x94000463;       (* arm_BL (word 4492) *)
  0x910083e1;       (* arm_ADD X1 SP (rvalue (word 32)) *)
  0x910383e2;       (* arm_ADD X2 SP (rvalue (word 224)) *)
  0xaa1303e0;       (* arm_MOV X0 X19 *)
  0x9400045f;       (* arm_BL (word 4476) *)
  0x91008260;       (* arm_ADD X0 X19 (rvalue (word 32)) *)
  0x910103e1;       (* arm_ADD X1 SP (rvalue (word 64)) *)
  0x910403e2;       (* arm_ADD X2 SP (rvalue (word 256)) *)
  0x9400045b;       (* arm_BL (word 4460) *)
  0x910f83ff;       (* arm_ADD SP SP (rvalue (word 992)) *)
  0xa8c17bf5;       (* arm_LDP X21 X30 SP (Postimmediate_Offset (iword (&16))) *)
  0xa8c153f3;       (* arm_LDP X19 X20 SP (Postimmediate_Offset (iword (&16))) *)
  0xd65f03c0;       (* arm_RET X30 *)
  0xa9400c22;       (* arm_LDP X2 X3 X1 (Immediate_Offset (iword (&0))) *)
  0xa9411424;       (* arm_LDP X4 X5 X1 (Immediate_Offset (iword (&16))) *)
  0xd3607c47;       (* arm_LSL X7 X2 32 *)
  0xeb070048;       (* arm_SUBS X8 X2 X7 *)
  0xd360fc46;       (* arm_LSR X6 X2 32 *)
  0xda060042;       (* arm_SBC X2 X2 X6 *)
  0xab070063;       (* arm_ADDS X3 X3 X7 *)
  0xba060084;       (* arm_ADCS X4 X4 X6 *)
  0xba0800a5;       (* arm_ADCS X5 X5 X8 *)
  0x9a1f0042;       (* arm_ADC X2 X2 XZR *)
  0xd3607c67;       (* arm_LSL X7 X3 32 *)
  0xeb070068;       (* arm_SUBS X8 X3 X7 *)
  0xd360fc66;       (* arm_LSR X6 X3 32 *)
  0xda060063;       (* arm_SBC X3 X3 X6 *)
  0xab070084;       (* arm_ADDS X4 X4 X7 *)
  0xba0600a5;       (* arm_ADCS X5 X5 X6 *)
  0xba080042;       (* arm_ADCS X2 X2 X8 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0xd3607c87;       (* arm_LSL X7 X4 32 *)
  0xeb070088;       (* arm_SUBS X8 X4 X7 *)
  0xd360fc86;       (* arm_LSR X6 X4 32 *)
  0xda060084;       (* arm_SBC X4 X4 X6 *)
  0xab0700a5;       (* arm_ADDS X5 X5 X7 *)
  0xba060042;       (* arm_ADCS X2 X2 X6 *)
  0xba080063;       (* arm_ADCS X3 X3 X8 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0xd3607ca7;       (* arm_LSL X7 X5 32 *)
  0xeb0700a8;       (* arm_SUBS X8 X5 X7 *)
  0xd360fca6;       (* arm_LSR X6 X5 32 *)
  0xda0600a5;       (* arm_SBC X5 X5 X6 *)
  0xab070042;       (* arm_ADDS X2 X2 X7 *)
  0xba060063;       (* arm_ADCS X3 X3 X6 *)
  0xba080084;       (* arm_ADCS X4 X4 X8 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xa9000c02;       (* arm_STP X2 X3 X0 (Immediate_Offset (iword (&0))) *)
  0xa9011404;       (* arm_STP X4 X5 X0 (Immediate_Offset (iword (&16))) *)
  0xd65f03c0;       (* arm_RET X30 *)
  0xa9bf53f3;       (* arm_STP X19 X20 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xa9bf5bf5;       (* arm_STP X21 X22 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xa9bf63f7;       (* arm_STP X23 X24 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xd10283ff;       (* arm_SUB SP SP (rvalue (word 160)) *)
  0xaa0003f4;       (* arm_MOV X20 X0 *)
  0x9280000a;       (* arm_MOVN X10 (word 0) 0 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb26083ed;       (* arm_MOV X13 (rvalue (word 18446744069414584321)) *)
  0xa9002fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&0))) *)
  0xa90137ff;       (* arm_STP XZR X13 SP (Immediate_Offset (iword (&16))) *)
  0xf90013ff;       (* arm_STR XZR SP (Immediate_Offset (word 32)) *)
  0xa9400c22;       (* arm_LDP X2 X3 X1 (Immediate_Offset (iword (&0))) *)
  0xeb0a004a;       (* arm_SUBS X10 X2 X10 *)
  0xfa0b006b;       (* arm_SBCS X11 X3 X11 *)
  0xa9411424;       (* arm_LDP X4 X5 X1 (Immediate_Offset (iword (&16))) *)
  0xfa1f008c;       (* arm_SBCS X12 X4 XZR *)
  0xfa0d00ad;       (* arm_SBCS X13 X5 X13 *)
  0x9a8a3042;       (* arm_CSEL X2 X2 X10 Condition_CC *)
  0x9a8b3063;       (* arm_CSEL X3 X3 X11 Condition_CC *)
  0x9a8c3084;       (* arm_CSEL X4 X4 X12 Condition_CC *)
  0x9a8d30a5;       (* arm_CSEL X5 X5 X13 Condition_CC *)
  0xa9030fe2;       (* arm_STP X2 X3 SP (Immediate_Offset (iword (&48))) *)
  0xa90417e4;       (* arm_STP X4 X5 SP (Immediate_Offset (iword (&64))) *)
  0xf9002bff;       (* arm_STR XZR SP (Immediate_Offset (word 80)) *)
  0xa9067fff;       (* arm_STP XZR XZR SP (Immediate_Offset (iword (&96))) *)
  0xa9077fff;       (* arm_STP XZR XZR SP (Immediate_Offset (iword (&112))) *)
  0xd2e0008a;       (* arm_MOVZ X10 (word 4) 48 *)
  0xa9087fea;       (* arm_STP X10 XZR SP (Immediate_Offset (iword (&128))) *)
  0xa9097fff;       (* arm_STP XZR XZR SP (Immediate_Offset (iword (&144))) *)
  0xd2800155;       (* arm_MOV X21 (rvalue (word 10)) *)
  0xd2800036;       (* arm_MOV X22 (rvalue (word 1)) *)
  0x14000131;       (* arm_B (word 1220) *)
  0xeb1f015f;       (* arm_CMP X10 XZR *)
  0xda9f53ee;       (* arm_CSETM X14 Condition_MI *)
  0xda8a554a;       (* arm_CNEG X10 X10 Condition_MI *)
  0xeb1f017f;       (* arm_CMP X11 XZR *)
  0xda9f53ef;       (* arm_CSETM X15 Condition_MI *)
  0xda8b556b;       (* arm_CNEG X11 X11 Condition_MI *)
  0xeb1f019f;       (* arm_CMP X12 XZR *)
  0xda9f53f0;       (* arm_CSETM X16 Condition_MI *)
  0xda8c558c;       (* arm_CNEG X12 X12 Condition_MI *)
  0xeb1f01bf;       (* arm_CMP X13 XZR *)
  0xda9f53f1;       (* arm_CSETM X17 Condition_MI *)
  0xda8d55ad;       (* arm_CNEG X13 X13 Condition_MI *)
  0x8a0e0140;       (* arm_AND X0 X10 X14 *)
  0x8a0f0161;       (* arm_AND X1 X11 X15 *)
  0x8b010009;       (* arm_ADD X9 X0 X1 *)
  0x8a100180;       (* arm_AND X0 X12 X16 *)
  0x8a1101a1;       (* arm_AND X1 X13 X17 *)
  0x8b010013;       (* arm_ADD X19 X0 X1 *)
  0xf94003e7;       (* arm_LDR X7 SP (Immediate_Offset (word 0)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab000124;       (* arm_ADDS X4 X9 X0 *)
  0x9a0103e2;       (* arm_ADC X2 XZR X1 *)
  0xf9401be8;       (* arm_LDR X8 SP (Immediate_Offset (word 48)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab000084;       (* arm_ADDS X4 X4 X0 *)
  0x9a010042;       (* arm_ADC X2 X2 X1 *)
  0xca1000e1;       (* arm_EOR X1 X7 X16 *)
  0x9b0c7c20;       (* arm_MUL X0 X1 X12 *)
  0x9bcc7c21;       (* arm_UMULH X1 X1 X12 *)
  0xab000265;       (* arm_ADDS X5 X19 X0 *)
  0x9a0103e3;       (* arm_ADC X3 XZR X1 *)
  0xca110101;       (* arm_EOR X1 X8 X17 *)
  0x9b0d7c20;       (* arm_MUL X0 X1 X13 *)
  0x9bcd7c21;       (* arm_UMULH X1 X1 X13 *)
  0xab0000a5;       (* arm_ADDS X5 X5 X0 *)
  0x9a010063;       (* arm_ADC X3 X3 X1 *)
  0xf94007e7;       (* arm_LDR X7 SP (Immediate_Offset (word 8)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0x9a0103e6;       (* arm_ADC X6 XZR X1 *)
  0xf9401fe8;       (* arm_LDR X8 SP (Immediate_Offset (word 56)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0x9a0100c6;       (* arm_ADC X6 X6 X1 *)
  0x93c4ec44;       (* arm_EXTR X4 X2 X4 59 *)
  0xf90003e4;       (* arm_STR X4 SP (Immediate_Offset (word 0)) *)
  0xca1000e1;       (* arm_EOR X1 X7 X16 *)
  0x9b0c7c20;       (* arm_MUL X0 X1 X12 *)
  0x9bcc7c21;       (* arm_UMULH X1 X1 X12 *)
  0xab000063;       (* arm_ADDS X3 X3 X0 *)
  0x9a0103e4;       (* arm_ADC X4 XZR X1 *)
  0xca110101;       (* arm_EOR X1 X8 X17 *)
  0x9b0d7c20;       (* arm_MUL X0 X1 X13 *)
  0x9bcd7c21;       (* arm_UMULH X1 X1 X13 *)
  0xab000063;       (* arm_ADDS X3 X3 X0 *)
  0x9a010084;       (* arm_ADC X4 X4 X1 *)
  0x93c5ec65;       (* arm_EXTR X5 X3 X5 59 *)
  0xf9001be5;       (* arm_STR X5 SP (Immediate_Offset (word 48)) *)
  0xf9400be7;       (* arm_LDR X7 SP (Immediate_Offset (word 16)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab0000c6;       (* arm_ADDS X6 X6 X0 *)
  0x9a0103e5;       (* arm_ADC X5 XZR X1 *)
  0xf94023e8;       (* arm_LDR X8 SP (Immediate_Offset (word 64)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab0000c6;       (* arm_ADDS X6 X6 X0 *)
  0x9a0100a5;       (* arm_ADC X5 X5 X1 *)
  0x93c2ecc2;       (* arm_EXTR X2 X6 X2 59 *)
  0xf90007e2;       (* arm_STR X2 SP (Immediate_Offset (word 8)) *)
  0xca1000e1;       (* arm_EOR X1 X7 X16 *)
  0x9b0c7c20;       (* arm_MUL X0 X1 X12 *)
  0x9bcc7c21;       (* arm_UMULH X1 X1 X12 *)
  0xab000084;       (* arm_ADDS X4 X4 X0 *)
  0x9a0103e2;       (* arm_ADC X2 XZR X1 *)
  0xca110101;       (* arm_EOR X1 X8 X17 *)
  0x9b0d7c20;       (* arm_MUL X0 X1 X13 *)
  0x9bcd7c21;       (* arm_UMULH X1 X1 X13 *)
  0xab000084;       (* arm_ADDS X4 X4 X0 *)
  0x9a010042;       (* arm_ADC X2 X2 X1 *)
  0x93c3ec83;       (* arm_EXTR X3 X4 X3 59 *)
  0xf9001fe3;       (* arm_STR X3 SP (Immediate_Offset (word 56)) *)
  0xf9400fe7;       (* arm_LDR X7 SP (Immediate_Offset (word 24)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0xf94013f7;       (* arm_LDR X23 SP (Immediate_Offset (word 32)) *)
  0xca0e02e3;       (* arm_EOR X3 X23 X14 *)
  0x8a0a0063;       (* arm_AND X3 X3 X10 *)
  0xcb0303e3;       (* arm_NEG X3 X3 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab0000a5;       (* arm_ADDS X5 X5 X0 *)
  0x9a010063;       (* arm_ADC X3 X3 X1 *)
  0xf94027e8;       (* arm_LDR X8 SP (Immediate_Offset (word 72)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0xf9402bf8;       (* arm_LDR X24 SP (Immediate_Offset (word 80)) *)
  0xca0f0300;       (* arm_EOR X0 X24 X15 *)
  0x8a0b0000;       (* arm_AND X0 X0 X11 *)
  0xcb000063;       (* arm_SUB X3 X3 X0 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab0000a5;       (* arm_ADDS X5 X5 X0 *)
  0x9a010063;       (* arm_ADC X3 X3 X1 *)
  0x93c6eca6;       (* arm_EXTR X6 X5 X6 59 *)
  0xf9000be6;       (* arm_STR X6 SP (Immediate_Offset (word 16)) *)
  0x93c5ec65;       (* arm_EXTR X5 X3 X5 59 *)
  0xf9000fe5;       (* arm_STR X5 SP (Immediate_Offset (word 24)) *)
  0x937bfc63;       (* arm_ASR X3 X3 59 *)
  0xf90013e3;       (* arm_STR X3 SP (Immediate_Offset (word 32)) *)
  0xca1000e1;       (* arm_EOR X1 X7 X16 *)
  0xca1002e5;       (* arm_EOR X5 X23 X16 *)
  0x8a0c00a5;       (* arm_AND X5 X5 X12 *)
  0xcb0503e5;       (* arm_NEG X5 X5 *)
  0x9b0c7c20;       (* arm_MUL X0 X1 X12 *)
  0x9bcc7c21;       (* arm_UMULH X1 X1 X12 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0x9a0100a5;       (* arm_ADC X5 X5 X1 *)
  0xca110101;       (* arm_EOR X1 X8 X17 *)
  0xca110300;       (* arm_EOR X0 X24 X17 *)
  0x8a0d0000;       (* arm_AND X0 X0 X13 *)
  0xcb0000a5;       (* arm_SUB X5 X5 X0 *)
  0x9b0d7c20;       (* arm_MUL X0 X1 X13 *)
  0x9bcd7c21;       (* arm_UMULH X1 X1 X13 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0x9a0100a5;       (* arm_ADC X5 X5 X1 *)
  0x93c4ec44;       (* arm_EXTR X4 X2 X4 59 *)
  0xf90023e4;       (* arm_STR X4 SP (Immediate_Offset (word 64)) *)
  0x93c2eca2;       (* arm_EXTR X2 X5 X2 59 *)
  0xf90027e2;       (* arm_STR X2 SP (Immediate_Offset (word 72)) *)
  0x937bfca5;       (* arm_ASR X5 X5 59 *)
  0xf9002be5;       (* arm_STR X5 SP (Immediate_Offset (word 80)) *)
  0xf94033e7;       (* arm_LDR X7 SP (Immediate_Offset (word 96)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab000124;       (* arm_ADDS X4 X9 X0 *)
  0x9a0103e2;       (* arm_ADC X2 XZR X1 *)
  0xf94043e8;       (* arm_LDR X8 SP (Immediate_Offset (word 128)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab000084;       (* arm_ADDS X4 X4 X0 *)
  0xf90033e4;       (* arm_STR X4 SP (Immediate_Offset (word 96)) *)
  0x9a010042;       (* arm_ADC X2 X2 X1 *)
  0xca1000e1;       (* arm_EOR X1 X7 X16 *)
  0x9b0c7c20;       (* arm_MUL X0 X1 X12 *)
  0x9bcc7c21;       (* arm_UMULH X1 X1 X12 *)
  0xab000265;       (* arm_ADDS X5 X19 X0 *)
  0x9a0103e3;       (* arm_ADC X3 XZR X1 *)
  0xca110101;       (* arm_EOR X1 X8 X17 *)
  0x9b0d7c20;       (* arm_MUL X0 X1 X13 *)
  0x9bcd7c21;       (* arm_UMULH X1 X1 X13 *)
  0xab0000a5;       (* arm_ADDS X5 X5 X0 *)
  0xf90043e5;       (* arm_STR X5 SP (Immediate_Offset (word 128)) *)
  0x9a010063;       (* arm_ADC X3 X3 X1 *)
  0xf94037e7;       (* arm_LDR X7 SP (Immediate_Offset (word 104)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0x9a0103e6;       (* arm_ADC X6 XZR X1 *)
  0xf94047e8;       (* arm_LDR X8 SP (Immediate_Offset (word 136)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0xf90037e2;       (* arm_STR X2 SP (Immediate_Offset (word 104)) *)
  0x9a0100c6;       (* arm_ADC X6 X6 X1 *)
  0xca1000e1;       (* arm_EOR X1 X7 X16 *)
  0x9b0c7c20;       (* arm_MUL X0 X1 X12 *)
  0x9bcc7c21;       (* arm_UMULH X1 X1 X12 *)
  0xab000063;       (* arm_ADDS X3 X3 X0 *)
  0x9a0103e4;       (* arm_ADC X4 XZR X1 *)
  0xca110101;       (* arm_EOR X1 X8 X17 *)
  0x9b0d7c20;       (* arm_MUL X0 X1 X13 *)
  0x9bcd7c21;       (* arm_UMULH X1 X1 X13 *)
  0xab000063;       (* arm_ADDS X3 X3 X0 *)
  0xf90047e3;       (* arm_STR X3 SP (Immediate_Offset (word 136)) *)
  0x9a010084;       (* arm_ADC X4 X4 X1 *)
  0xf9403be7;       (* arm_LDR X7 SP (Immediate_Offset (word 112)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab0000c6;       (* arm_ADDS X6 X6 X0 *)
  0x9a0103e5;       (* arm_ADC X5 XZR X1 *)
  0xf9404be8;       (* arm_LDR X8 SP (Immediate_Offset (word 144)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab0000c6;       (* arm_ADDS X6 X6 X0 *)
  0xf9003be6;       (* arm_STR X6 SP (Immediate_Offset (word 112)) *)
  0x9a0100a5;       (* arm_ADC X5 X5 X1 *)
  0xca1000e1;       (* arm_EOR X1 X7 X16 *)
  0x9b0c7c20;       (* arm_MUL X0 X1 X12 *)
  0x9bcc7c21;       (* arm_UMULH X1 X1 X12 *)
  0xab000084;       (* arm_ADDS X4 X4 X0 *)
  0x9a0103e2;       (* arm_ADC X2 XZR X1 *)
  0xca110101;       (* arm_EOR X1 X8 X17 *)
  0x9b0d7c20;       (* arm_MUL X0 X1 X13 *)
  0x9bcd7c21;       (* arm_UMULH X1 X1 X13 *)
  0xab000084;       (* arm_ADDS X4 X4 X0 *)
  0xf9004be4;       (* arm_STR X4 SP (Immediate_Offset (word 144)) *)
  0x9a010042;       (* arm_ADC X2 X2 X1 *)
  0xf9403fe7;       (* arm_LDR X7 SP (Immediate_Offset (word 120)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x8a0a01c3;       (* arm_AND X3 X14 X10 *)
  0xcb0303e3;       (* arm_NEG X3 X3 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab0000a5;       (* arm_ADDS X5 X5 X0 *)
  0x9a010063;       (* arm_ADC X3 X3 X1 *)
  0xf9404fe8;       (* arm_LDR X8 SP (Immediate_Offset (word 152)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x8a0b01e0;       (* arm_AND X0 X15 X11 *)
  0xcb000063;       (* arm_SUB X3 X3 X0 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab0000a5;       (* arm_ADDS X5 X5 X0 *)
  0x9a010063;       (* arm_ADC X3 X3 X1 *)
  0xa94607e0;       (* arm_LDP X0 X1 SP (Immediate_Offset (iword (&96))) *)
  0xf9403be6;       (* arm_LDR X6 SP (Immediate_Offset (word 112)) *)
  0xd2fc000e;       (* arm_MOVZ X14 (word 57344) 48 *)
  0xab0e0000;       (* arm_ADDS X0 X0 X14 *)
  0xfa1f0021;       (* arm_SBCS X1 X1 XZR *)
  0xb24073eb;       (* arm_MOV X11 (rvalue (word 536870911)) *)
  0xba0b00c6;       (* arm_ADCS X6 X6 X11 *)
  0xd2e4000a;       (* arm_MOVZ X10 (word 8192) 48 *)
  0xba0a00a5;       (* arm_ADCS X5 X5 X10 *)
  0xb2637fee;       (* arm_MOV X14 (rvalue (word 2305843008676823040)) *)
  0x9a0e0063;       (* arm_ADC X3 X3 X14 *)
  0xd3607c0b;       (* arm_LSL X11 X0 32 *)
  0xeb0b000e;       (* arm_SUBS X14 X0 X11 *)
  0xd360fc0a;       (* arm_LSR X10 X0 32 *)
  0xda0a0000;       (* arm_SBC X0 X0 X10 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0xba0a00c6;       (* arm_ADCS X6 X6 X10 *)
  0xba0e00a5;       (* arm_ADCS X5 X5 X14 *)
  0xba000063;       (* arm_ADCS X3 X3 X0 *)
  0x9280000e;       (* arm_MOVN X14 (word 0) 0 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0x9a9f21ce;       (* arm_CSEL X14 X14 XZR Condition_CS *)
  0x9a9f216b;       (* arm_CSEL X11 X11 XZR Condition_CS *)
  0x9a9f214a;       (* arm_CSEL X10 X10 XZR Condition_CS *)
  0xeb0e0021;       (* arm_SUBS X1 X1 X14 *)
  0xfa0b00c6;       (* arm_SBCS X6 X6 X11 *)
  0xfa1f00a5;       (* arm_SBCS X5 X5 XZR *)
  0xda0a0063;       (* arm_SBC X3 X3 X10 *)
  0xa9061be1;       (* arm_STP X1 X6 SP (Immediate_Offset (iword (&96))) *)
  0xa9070fe5;       (* arm_STP X5 X3 SP (Immediate_Offset (iword (&112))) *)
  0xca1000e1;       (* arm_EOR X1 X7 X16 *)
  0x8a0c0205;       (* arm_AND X5 X16 X12 *)
  0xcb0503e5;       (* arm_NEG X5 X5 *)
  0x9b0c7c20;       (* arm_MUL X0 X1 X12 *)
  0x9bcc7c21;       (* arm_UMULH X1 X1 X12 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0x9a0100a5;       (* arm_ADC X5 X5 X1 *)
  0xca110101;       (* arm_EOR X1 X8 X17 *)
  0x8a0d0220;       (* arm_AND X0 X17 X13 *)
  0xcb0000a5;       (* arm_SUB X5 X5 X0 *)
  0x9b0d7c20;       (* arm_MUL X0 X1 X13 *)
  0x9bcd7c21;       (* arm_UMULH X1 X1 X13 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0x9a0100a5;       (* arm_ADC X5 X5 X1 *)
  0xa94807e0;       (* arm_LDP X0 X1 SP (Immediate_Offset (iword (&128))) *)
  0xf9404be3;       (* arm_LDR X3 SP (Immediate_Offset (word 144)) *)
  0xd2fc000e;       (* arm_MOVZ X14 (word 57344) 48 *)
  0xab0e0000;       (* arm_ADDS X0 X0 X14 *)
  0xfa1f0021;       (* arm_SBCS X1 X1 XZR *)
  0xb24073eb;       (* arm_MOV X11 (rvalue (word 536870911)) *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0xd2e4000a;       (* arm_MOVZ X10 (word 8192) 48 *)
  0xba0a0042;       (* arm_ADCS X2 X2 X10 *)
  0xb2637fee;       (* arm_MOV X14 (rvalue (word 2305843008676823040)) *)
  0x9a0e00a5;       (* arm_ADC X5 X5 X14 *)
  0xd3607c0b;       (* arm_LSL X11 X0 32 *)
  0xeb0b000e;       (* arm_SUBS X14 X0 X11 *)
  0xd360fc0a;       (* arm_LSR X10 X0 32 *)
  0xda0a0000;       (* arm_SBC X0 X0 X10 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0xba0a0063;       (* arm_ADCS X3 X3 X10 *)
  0xba0e0042;       (* arm_ADCS X2 X2 X14 *)
  0xba0000a5;       (* arm_ADCS X5 X5 X0 *)
  0x9280000e;       (* arm_MOVN X14 (word 0) 0 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0x9a9f21ce;       (* arm_CSEL X14 X14 XZR Condition_CS *)
  0x9a9f216b;       (* arm_CSEL X11 X11 XZR Condition_CS *)
  0x9a9f214a;       (* arm_CSEL X10 X10 XZR Condition_CS *)
  0xeb0e0021;       (* arm_SUBS X1 X1 X14 *)
  0xfa0b0063;       (* arm_SBCS X3 X3 X11 *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xda0a00a5;       (* arm_SBC X5 X5 X10 *)
  0xa9080fe1;       (* arm_STP X1 X3 SP (Immediate_Offset (iword (&128))) *)
  0xa90917e2;       (* arm_STP X2 X5 SP (Immediate_Offset (iword (&144))) *)
  0xaa1603e1;       (* arm_MOV X1 X22 *)
  0xf94003e2;       (* arm_LDR X2 SP (Immediate_Offset (word 0)) *)
  0xf9401be3;       (* arm_LDR X3 SP (Immediate_Offset (word 48)) *)
  0x92404c44;       (* arm_AND X4 X2 (rvalue (word 1048575)) *)
  0xb2575884;       (* arm_ORR X4 X4 (rvalue (word 18446741874686296064)) *)
  0x92404c65;       (* arm_AND X5 X3 (rvalue (word 1048575)) *)
  0xb24204a5;       (* arm_ORR X5 X5 (rvalue (word 13835058055282163712)) *)
  0xf24000bf;       (* arm_TST X5 (rvalue (word 1)) *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x91440088;       (* arm_ADD X8 X4 (rvalue (word 1048576)) *)
  0x9355a508;       (* arm_SBFM X8 X8 21 41 *)
  0xd2a0020b;       (* arm_MOVZ X11 (word 16) 16 *)
  0x8b0b556b;       (* arm_ADD X11 X11 (Shiftedreg X11 LSL 21) *)
  0x8b0b0089;       (* arm_ADD X9 X4 X11 *)
  0x936afd29;       (* arm_ASR X9 X9 42 *)
  0x914400aa;       (* arm_ADD X10 X5 (rvalue (word 1048576)) *)
  0x9355a54a;       (* arm_SBFM X10 X10 21 41 *)
  0x8b0b00ab;       (* arm_ADD X11 X5 X11 *)
  0x936afd6b;       (* arm_ASR X11 X11 42 *)
  0x9b027d06;       (* arm_MUL X6 X8 X2 *)
  0x9b037d27;       (* arm_MUL X7 X9 X3 *)
  0x9b027d42;       (* arm_MUL X2 X10 X2 *)
  0x9b037d63;       (* arm_MUL X3 X11 X3 *)
  0x8b0700c4;       (* arm_ADD X4 X6 X7 *)
  0x8b030045;       (* arm_ADD X5 X2 X3 *)
  0x9354fc82;       (* arm_ASR X2 X4 20 *)
  0x9354fca3;       (* arm_ASR X3 X5 20 *)
  0x92404c44;       (* arm_AND X4 X2 (rvalue (word 1048575)) *)
  0xb2575884;       (* arm_ORR X4 X4 (rvalue (word 18446741874686296064)) *)
  0x92404c65;       (* arm_AND X5 X3 (rvalue (word 1048575)) *)
  0xb24204a5;       (* arm_ORR X5 X5 (rvalue (word 13835058055282163712)) *)
  0xf24000bf;       (* arm_TST X5 (rvalue (word 1)) *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9144008c;       (* arm_ADD X12 X4 (rvalue (word 1048576)) *)
  0x9355a58c;       (* arm_SBFM X12 X12 21 41 *)
  0xd2a0020f;       (* arm_MOVZ X15 (word 16) 16 *)
  0x8b0f55ef;       (* arm_ADD X15 X15 (Shiftedreg X15 LSL 21) *)
  0x8b0f008d;       (* arm_ADD X13 X4 X15 *)
  0x936afdad;       (* arm_ASR X13 X13 42 *)
  0x914400ae;       (* arm_ADD X14 X5 (rvalue (word 1048576)) *)
  0x9355a5ce;       (* arm_SBFM X14 X14 21 41 *)
  0x8b0f00af;       (* arm_ADD X15 X5 X15 *)
  0x936afdef;       (* arm_ASR X15 X15 42 *)
  0x9b027d86;       (* arm_MUL X6 X12 X2 *)
  0x9b037da7;       (* arm_MUL X7 X13 X3 *)
  0x9b027dc2;       (* arm_MUL X2 X14 X2 *)
  0x9b037de3;       (* arm_MUL X3 X15 X3 *)
  0x8b0700c4;       (* arm_ADD X4 X6 X7 *)
  0x8b030045;       (* arm_ADD X5 X2 X3 *)
  0x9354fc82;       (* arm_ASR X2 X4 20 *)
  0x9354fca3;       (* arm_ASR X3 X5 20 *)
  0x92404c44;       (* arm_AND X4 X2 (rvalue (word 1048575)) *)
  0xb2575884;       (* arm_ORR X4 X4 (rvalue (word 18446741874686296064)) *)
  0x92404c65;       (* arm_AND X5 X3 (rvalue (word 1048575)) *)
  0xb24204a5;       (* arm_ORR X5 X5 (rvalue (word 13835058055282163712)) *)
  0xf24000bf;       (* arm_TST X5 (rvalue (word 1)) *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9b087d82;       (* arm_MUL X2 X12 X8 *)
  0x9b097d83;       (* arm_MUL X3 X12 X9 *)
  0x9b087dc6;       (* arm_MUL X6 X14 X8 *)
  0x9b097dc7;       (* arm_MUL X7 X14 X9 *)
  0x9b0a09a8;       (* arm_MADD X8 X13 X10 X2 *)
  0x9b0b0da9;       (* arm_MADD X9 X13 X11 X3 *)
  0x9b0a19f0;       (* arm_MADD X16 X15 X10 X6 *)
  0x9b0b1df1;       (* arm_MADD X17 X15 X11 X7 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0xf27f00bf;       (* arm_TST X5 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9a9f1086;       (* arm_CSEL X6 X4 XZR Condition_NE *)
  0xfa5f1028;       (* arm_CCMP X1 XZR (word 8) Condition_NE *)
  0xda81b421;       (* arm_CNEG X1 X1 Condition_GE *)
  0xda86b4c6;       (* arm_CNEG X6 X6 Condition_GE *)
  0x9a84a0a4;       (* arm_CSEL X4 X5 X4 Condition_GE *)
  0x8b0600a5;       (* arm_ADD X5 X5 X6 *)
  0x91000821;       (* arm_ADD X1 X1 (rvalue (word 2)) *)
  0x9341fca5;       (* arm_ASR X5 X5 1 *)
  0x9144008c;       (* arm_ADD X12 X4 (rvalue (word 1048576)) *)
  0x9356a98c;       (* arm_SBFM X12 X12 22 42 *)
  0xd2a0020f;       (* arm_MOVZ X15 (word 16) 16 *)
  0x8b0f55ef;       (* arm_ADD X15 X15 (Shiftedreg X15 LSL 21) *)
  0x8b0f008d;       (* arm_ADD X13 X4 X15 *)
  0x936bfdad;       (* arm_ASR X13 X13 43 *)
  0x914400ae;       (* arm_ADD X14 X5 (rvalue (word 1048576)) *)
  0x9356a9ce;       (* arm_SBFM X14 X14 22 42 *)
  0x8b0f00af;       (* arm_ADD X15 X5 X15 *)
  0x936bfdef;       (* arm_ASR X15 X15 43 *)
  0x9b08fd82;       (* arm_MNEG X2 X12 X8 *)
  0x9b09fd83;       (* arm_MNEG X3 X12 X9 *)
  0x9b08fdc4;       (* arm_MNEG X4 X14 X8 *)
  0x9b09fdc5;       (* arm_MNEG X5 X14 X9 *)
  0x9b1089aa;       (* arm_MSUB X10 X13 X16 X2 *)
  0x9b118dab;       (* arm_MSUB X11 X13 X17 X3 *)
  0x9b1091ec;       (* arm_MSUB X12 X15 X16 X4 *)
  0x9b1195ed;       (* arm_MSUB X13 X15 X17 X5 *)
  0xaa0103f6;       (* arm_MOV X22 X1 *)
  0xf10006b5;       (* arm_SUBS X21 X21 (rvalue (word 1)) *)
  0x54ff8dc1;       (* arm_BNE (word 2093496) *)
  0xf94003e0;       (* arm_LDR X0 SP (Immediate_Offset (word 0)) *)
  0xf9401be1;       (* arm_LDR X1 SP (Immediate_Offset (word 48)) *)
  0x9b0a7c00;       (* arm_MUL X0 X0 X10 *)
  0x9b0b0021;       (* arm_MADD X1 X1 X11 X0 *)
  0x937ffc20;       (* arm_ASR X0 X1 63 *)
  0xeb1f015f;       (* arm_CMP X10 XZR *)
  0xda9f53ee;       (* arm_CSETM X14 Condition_MI *)
  0xda8a554a;       (* arm_CNEG X10 X10 Condition_MI *)
  0xca0001ce;       (* arm_EOR X14 X14 X0 *)
  0xeb1f017f;       (* arm_CMP X11 XZR *)
  0xda9f53ef;       (* arm_CSETM X15 Condition_MI *)
  0xda8b556b;       (* arm_CNEG X11 X11 Condition_MI *)
  0xca0001ef;       (* arm_EOR X15 X15 X0 *)
  0xeb1f019f;       (* arm_CMP X12 XZR *)
  0xda9f53f0;       (* arm_CSETM X16 Condition_MI *)
  0xda8c558c;       (* arm_CNEG X12 X12 Condition_MI *)
  0xca000210;       (* arm_EOR X16 X16 X0 *)
  0xeb1f01bf;       (* arm_CMP X13 XZR *)
  0xda9f53f1;       (* arm_CSETM X17 Condition_MI *)
  0xda8d55ad;       (* arm_CNEG X13 X13 Condition_MI *)
  0xca000231;       (* arm_EOR X17 X17 X0 *)
  0x8a0e0140;       (* arm_AND X0 X10 X14 *)
  0x8a0f0161;       (* arm_AND X1 X11 X15 *)
  0x8b010009;       (* arm_ADD X9 X0 X1 *)
  0xf94033e7;       (* arm_LDR X7 SP (Immediate_Offset (word 96)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab000124;       (* arm_ADDS X4 X9 X0 *)
  0x9a0103e2;       (* arm_ADC X2 XZR X1 *)
  0xf94043e8;       (* arm_LDR X8 SP (Immediate_Offset (word 128)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab000084;       (* arm_ADDS X4 X4 X0 *)
  0xf90033e4;       (* arm_STR X4 SP (Immediate_Offset (word 96)) *)
  0x9a010042;       (* arm_ADC X2 X2 X1 *)
  0xf94037e7;       (* arm_LDR X7 SP (Immediate_Offset (word 104)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0x9a0103e6;       (* arm_ADC X6 XZR X1 *)
  0xf94047e8;       (* arm_LDR X8 SP (Immediate_Offset (word 136)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab000042;       (* arm_ADDS X2 X2 X0 *)
  0xf90037e2;       (* arm_STR X2 SP (Immediate_Offset (word 104)) *)
  0x9a0100c6;       (* arm_ADC X6 X6 X1 *)
  0xf9403be7;       (* arm_LDR X7 SP (Immediate_Offset (word 112)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab0000c6;       (* arm_ADDS X6 X6 X0 *)
  0x9a0103e5;       (* arm_ADC X5 XZR X1 *)
  0xf9404be8;       (* arm_LDR X8 SP (Immediate_Offset (word 144)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab0000c6;       (* arm_ADDS X6 X6 X0 *)
  0xf9003be6;       (* arm_STR X6 SP (Immediate_Offset (word 112)) *)
  0x9a0100a5;       (* arm_ADC X5 X5 X1 *)
  0xf9403fe7;       (* arm_LDR X7 SP (Immediate_Offset (word 120)) *)
  0xca0e00e1;       (* arm_EOR X1 X7 X14 *)
  0x8a0a01c3;       (* arm_AND X3 X14 X10 *)
  0xcb0303e3;       (* arm_NEG X3 X3 *)
  0x9b0a7c20;       (* arm_MUL X0 X1 X10 *)
  0x9bca7c21;       (* arm_UMULH X1 X1 X10 *)
  0xab0000a5;       (* arm_ADDS X5 X5 X0 *)
  0x9a010063;       (* arm_ADC X3 X3 X1 *)
  0xf9404fe8;       (* arm_LDR X8 SP (Immediate_Offset (word 152)) *)
  0xca0f0101;       (* arm_EOR X1 X8 X15 *)
  0x8a0b01e0;       (* arm_AND X0 X15 X11 *)
  0xcb000063;       (* arm_SUB X3 X3 X0 *)
  0x9b0b7c20;       (* arm_MUL X0 X1 X11 *)
  0x9bcb7c21;       (* arm_UMULH X1 X1 X11 *)
  0xab0000a5;       (* arm_ADDS X5 X5 X0 *)
  0x9a010063;       (* arm_ADC X3 X3 X1 *)
  0xa94607e0;       (* arm_LDP X0 X1 SP (Immediate_Offset (iword (&96))) *)
  0xf9403be2;       (* arm_LDR X2 SP (Immediate_Offset (word 112)) *)
  0xd2fc000e;       (* arm_MOVZ X14 (word 57344) 48 *)
  0xab0e0000;       (* arm_ADDS X0 X0 X14 *)
  0xfa1f0021;       (* arm_SBCS X1 X1 XZR *)
  0xb24073eb;       (* arm_MOV X11 (rvalue (word 536870911)) *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0xd2e4000a;       (* arm_MOVZ X10 (word 8192) 48 *)
  0xba0a00a5;       (* arm_ADCS X5 X5 X10 *)
  0xb2637fee;       (* arm_MOV X14 (rvalue (word 2305843008676823040)) *)
  0x9a0e0063;       (* arm_ADC X3 X3 X14 *)
  0xd3607c0b;       (* arm_LSL X11 X0 32 *)
  0xeb0b000e;       (* arm_SUBS X14 X0 X11 *)
  0xd360fc0a;       (* arm_LSR X10 X0 32 *)
  0xda0a0000;       (* arm_SBC X0 X0 X10 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0xba0a0042;       (* arm_ADCS X2 X2 X10 *)
  0xba0e00a5;       (* arm_ADCS X5 X5 X14 *)
  0xba000063;       (* arm_ADCS X3 X3 X0 *)
  0x9280000e;       (* arm_MOVN X14 (word 0) 0 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0x9a9f21ce;       (* arm_CSEL X14 X14 XZR Condition_CS *)
  0x9a9f216b;       (* arm_CSEL X11 X11 XZR Condition_CS *)
  0x9a9f214a;       (* arm_CSEL X10 X10 XZR Condition_CS *)
  0xeb0e0021;       (* arm_SUBS X1 X1 X14 *)
  0xfa0b0042;       (* arm_SBCS X2 X2 X11 *)
  0xfa1f00a5;       (* arm_SBCS X5 X5 XZR *)
  0xda0a0063;       (* arm_SBC X3 X3 X10 *)
  0x9280000a;       (* arm_MOVN X10 (word 0) 0 *)
  0xeb0a002a;       (* arm_SUBS X10 X1 X10 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xfa0b004b;       (* arm_SBCS X11 X2 X11 *)
  0xb26083ed;       (* arm_MOV X13 (rvalue (word 18446744069414584321)) *)
  0xfa1f00ac;       (* arm_SBCS X12 X5 XZR *)
  0xfa0d006d;       (* arm_SBCS X13 X3 X13 *)
  0x9a8a302a;       (* arm_CSEL X10 X1 X10 Condition_CC *)
  0x9a8b304b;       (* arm_CSEL X11 X2 X11 Condition_CC *)
  0x9a8c30ac;       (* arm_CSEL X12 X5 X12 Condition_CC *)
  0x9a8d306d;       (* arm_CSEL X13 X3 X13 Condition_CC *)
  0xa9002e8a;       (* arm_STP X10 X11 X20 (Immediate_Offset (iword (&0))) *)
  0xa901368c;       (* arm_STP X12 X13 X20 (Immediate_Offset (iword (&16))) *)
  0x910283ff;       (* arm_ADD SP SP (rvalue (word 160)) *)
  0xa8c163f7;       (* arm_LDP X23 X24 SP (Postimmediate_Offset (iword (&16))) *)
  0xa8c15bf5;       (* arm_LDP X21 X22 SP (Postimmediate_Offset (iword (&16))) *)
  0xa8c153f3;       (* arm_LDP X19 X20 SP (Postimmediate_Offset (iword (&16))) *)
  0xd65f03c0;       (* arm_RET X30 *)
  0xa9401023;       (* arm_LDP X3 X4 X1 (Immediate_Offset (iword (&0))) *)
  0xa9402047;       (* arm_LDP X7 X8 X2 (Immediate_Offset (iword (&0))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9412849;       (* arm_LDP X9 X10 X2 (Immediate_Offset (iword (&16))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c6f;       (* arm_UMULH X15 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c70;       (* arm_UMULH X16 X3 X10 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0xa9411825;       (* arm_LDP X5 X6 X1 (Immediate_Offset (iword (&16))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e81ef;       (* arm_ADDS X15 X15 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab0f818c;       (* arm_ADDS X12 X12 (Shiftedreg X15 LSL 32) *)
  0xd360fdeb;       (* arm_LSR X11 X15 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7deb;       (* arm_MUL X11 X15 X10 *)
  0x9bca7def;       (* arm_UMULH X15 X15 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f01ef;       (* arm_ADC X15 X15 XZR *)
  0xab10018c;       (* arm_ADDS X12 X12 X16 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba0501ef;       (* arm_ADCS X15 X15 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000590;       (* arm_ADDS X16 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a01e5;       (* arm_SBCS X5 X15 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a90318c;       (* arm_CSEL X12 X12 X16 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a8531ef;       (* arm_CSEL X15 X15 X5 Condition_CC *)
  0xa900340c;       (* arm_STP X12 X13 X0 (Immediate_Offset (iword (&0))) *)
  0xa9013c0e;       (* arm_STP X14 X15 X0 (Immediate_Offset (iword (&16))) *)
  0xd65f03c0;       (* arm_RET X30 *)
  0xa9400c22;       (* arm_LDP X2 X3 X1 (Immediate_Offset (iword (&0))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa9411424;       (* arm_LDP X4 X5 X1 (Immediate_Offset (iword (&16))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd02;       (* arm_LSR X2 X8 32 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9b057d02;       (* arm_MUL X2 X8 X5 *)
  0x9bc57d08;       (* arm_UMULH X8 X8 X5 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd22;       (* arm_LSR X2 X9 32 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9b057d22;       (* arm_MUL X2 X9 X5 *)
  0x9bc57d29;       (* arm_UMULH X9 X9 X5 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd42;       (* arm_LSR X2 X10 32 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9b057d42;       (* arm_MUL X2 X10 X5 *)
  0x9bc57d4a;       (* arm_UMULH X10 X10 X5 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd62;       (* arm_LSR X2 X11 32 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9b057d62;       (* arm_MUL X2 X11 X5 *)
  0x9bc57d6b;       (* arm_UMULH X11 X11 X5 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0xb100050c;       (* arm_ADDS X12 X8 (rvalue (word 1)) *)
  0xfa03012d;       (* arm_SBCS X13 X9 X3 *)
  0xfa1f014e;       (* arm_SBCS X14 X10 XZR *)
  0xfa050167;       (* arm_SBCS X7 X11 X5 *)
  0xfa1f005f;       (* arm_SBCS XZR X2 XZR *)
  0x9a8c3108;       (* arm_CSEL X8 X8 X12 Condition_CC *)
  0x9a8d3129;       (* arm_CSEL X9 X9 X13 Condition_CC *)
  0x9a8e314a;       (* arm_CSEL X10 X10 X14 Condition_CC *)
  0x9a87316b;       (* arm_CSEL X11 X11 X7 Condition_CC *)
  0xa9002408;       (* arm_STP X8 X9 X0 (Immediate_Offset (iword (&0))) *)
  0xa9012c0a;       (* arm_STP X10 X11 X0 (Immediate_Offset (iword (&16))) *)
  0xd65f03c0;       (* arm_RET X30 *)
  0xa9400c22;       (* arm_LDP X2 X3 X1 (Immediate_Offset (iword (&0))) *)
  0xa9411424;       (* arm_LDP X4 X5 X1 (Immediate_Offset (iword (&16))) *)
  0x92800001;       (* arm_MOVN X1 (word 0) 0 *)
  0xb2407fe7;       (* arm_MOV X7 (rvalue (word 4294967295)) *)
  0xb26083e9;       (* arm_MOV X9 (rvalue (word 18446744069414584321)) *)
  0xeb010041;       (* arm_SUBS X1 X2 X1 *)
  0xfa070067;       (* arm_SBCS X7 X3 X7 *)
  0xfa1f0088;       (* arm_SBCS X8 X4 XZR *)
  0xfa0900a9;       (* arm_SBCS X9 X5 X9 *)
  0x9a813042;       (* arm_CSEL X2 X2 X1 Condition_CC *)
  0x9a873063;       (* arm_CSEL X3 X3 X7 Condition_CC *)
  0x9a883084;       (* arm_CSEL X4 X4 X8 Condition_CC *)
  0x9a8930a5;       (* arm_CSEL X5 X5 X9 Condition_CC *)
  0xeb1f03ff;       (* arm_CMP XZR XZR *)
  0x93c480a9;       (* arm_EXTR X9 X5 X4 32 *)
  0xba09009f;       (* arm_ADCS XZR X4 X9 *)
  0xd360fca9;       (* arm_LSR X9 X5 32 *)
  0xba0900a9;       (* arm_ADCS X9 X5 X9 *)
  0xda9f33e6;       (* arm_CSETM X6 Condition_CS *)
  0xaa060129;       (* arm_ORR X9 X9 X6 *)
  0xd3607d27;       (* arm_LSL X7 X9 32 *)
  0xd360fd28;       (* arm_LSR X8 X9 32 *)
  0xab070084;       (* arm_ADDS X4 X4 X7 *)
  0x9a0800a5;       (* arm_ADC X5 X5 X8 *)
  0xeb0903e6;       (* arm_NEGS X6 X9 *)
  0xfa1f00e7;       (* arm_SBCS X7 X7 XZR *)
  0xda1f0108;       (* arm_SBC X8 X8 XZR *)
  0xeb0603e6;       (* arm_NEGS X6 X6 *)
  0xfa070042;       (* arm_SBCS X2 X2 X7 *)
  0xfa080063;       (* arm_SBCS X3 X3 X8 *)
  0xfa090084;       (* arm_SBCS X4 X4 X9 *)
  0xfa0900a5;       (* arm_SBCS X5 X5 X9 *)
  0xab0500c6;       (* arm_ADDS X6 X6 X5 *)
  0xb2407fe7;       (* arm_MOV X7 (rvalue (word 4294967295)) *)
  0x8a0500e7;       (* arm_AND X7 X7 X5 *)
  0xba070042;       (* arm_ADCS X2 X2 X7 *)
  0xba1f0063;       (* arm_ADCS X3 X3 XZR *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x8a0500e7;       (* arm_AND X7 X7 X5 *)
  0x9a070084;       (* arm_ADC X4 X4 X7 *)
  0xeb1f03ff;       (* arm_CMP XZR XZR *)
  0x93c38089;       (* arm_EXTR X9 X4 X3 32 *)
  0xba09007f;       (* arm_ADCS XZR X3 X9 *)
  0xd360fc89;       (* arm_LSR X9 X4 32 *)
  0xba090089;       (* arm_ADCS X9 X4 X9 *)
  0xda9f33e5;       (* arm_CSETM X5 Condition_CS *)
  0xaa050129;       (* arm_ORR X9 X9 X5 *)
  0xd3607d27;       (* arm_LSL X7 X9 32 *)
  0xd360fd28;       (* arm_LSR X8 X9 32 *)
  0xab070063;       (* arm_ADDS X3 X3 X7 *)
  0x9a080084;       (* arm_ADC X4 X4 X8 *)
  0xeb0903e5;       (* arm_NEGS X5 X9 *)
  0xfa1f00e7;       (* arm_SBCS X7 X7 XZR *)
  0xda1f0108;       (* arm_SBC X8 X8 XZR *)
  0xeb0503e5;       (* arm_NEGS X5 X5 *)
  0xfa0700c6;       (* arm_SBCS X6 X6 X7 *)
  0xfa080042;       (* arm_SBCS X2 X2 X8 *)
  0xfa090063;       (* arm_SBCS X3 X3 X9 *)
  0xfa090084;       (* arm_SBCS X4 X4 X9 *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xb2407fe7;       (* arm_MOV X7 (rvalue (word 4294967295)) *)
  0x8a0400e7;       (* arm_AND X7 X7 X4 *)
  0xba0700c6;       (* arm_ADCS X6 X6 X7 *)
  0xba1f0042;       (* arm_ADCS X2 X2 XZR *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x8a0400e7;       (* arm_AND X7 X7 X4 *)
  0x9a070063;       (* arm_ADC X3 X3 X7 *)
  0xeb1f03ff;       (* arm_CMP XZR XZR *)
  0x93c28069;       (* arm_EXTR X9 X3 X2 32 *)
  0xba09005f;       (* arm_ADCS XZR X2 X9 *)
  0xd360fc69;       (* arm_LSR X9 X3 32 *)
  0xba090069;       (* arm_ADCS X9 X3 X9 *)
  0xda9f33e4;       (* arm_CSETM X4 Condition_CS *)
  0xaa040129;       (* arm_ORR X9 X9 X4 *)
  0xd3607d27;       (* arm_LSL X7 X9 32 *)
  0xd360fd28;       (* arm_LSR X8 X9 32 *)
  0xab070042;       (* arm_ADDS X2 X2 X7 *)
  0x9a080063;       (* arm_ADC X3 X3 X8 *)
  0xeb0903e4;       (* arm_NEGS X4 X9 *)
  0xfa1f00e7;       (* arm_SBCS X7 X7 XZR *)
  0xda1f0108;       (* arm_SBC X8 X8 XZR *)
  0xeb0403e4;       (* arm_NEGS X4 X4 *)
  0xfa0700a5;       (* arm_SBCS X5 X5 X7 *)
  0xfa0800c6;       (* arm_SBCS X6 X6 X8 *)
  0xfa090042;       (* arm_SBCS X2 X2 X9 *)
  0xfa090063;       (* arm_SBCS X3 X3 X9 *)
  0xab030084;       (* arm_ADDS X4 X4 X3 *)
  0xb2407fe7;       (* arm_MOV X7 (rvalue (word 4294967295)) *)
  0x8a0300e7;       (* arm_AND X7 X7 X3 *)
  0xba0700a5;       (* arm_ADCS X5 X5 X7 *)
  0xba1f00c6;       (* arm_ADCS X6 X6 XZR *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x8a0300e7;       (* arm_AND X7 X7 X3 *)
  0x9a070042;       (* arm_ADC X2 X2 X7 *)
  0xeb1f03ff;       (* arm_CMP XZR XZR *)
  0x93c68049;       (* arm_EXTR X9 X2 X6 32 *)
  0xba0900df;       (* arm_ADCS XZR X6 X9 *)
  0xd360fc49;       (* arm_LSR X9 X2 32 *)
  0xba090049;       (* arm_ADCS X9 X2 X9 *)
  0xda9f33e3;       (* arm_CSETM X3 Condition_CS *)
  0xaa030129;       (* arm_ORR X9 X9 X3 *)
  0xd3607d27;       (* arm_LSL X7 X9 32 *)
  0xd360fd28;       (* arm_LSR X8 X9 32 *)
  0xab0700c6;       (* arm_ADDS X6 X6 X7 *)
  0x9a080042;       (* arm_ADC X2 X2 X8 *)
  0xeb0903e3;       (* arm_NEGS X3 X9 *)
  0xfa1f00e7;       (* arm_SBCS X7 X7 XZR *)
  0xda1f0108;       (* arm_SBC X8 X8 XZR *)
  0xeb0303e3;       (* arm_NEGS X3 X3 *)
  0xfa070084;       (* arm_SBCS X4 X4 X7 *)
  0xfa0800a5;       (* arm_SBCS X5 X5 X8 *)
  0xfa0900c6;       (* arm_SBCS X6 X6 X9 *)
  0xfa090042;       (* arm_SBCS X2 X2 X9 *)
  0xab020063;       (* arm_ADDS X3 X3 X2 *)
  0xb2407fe7;       (* arm_MOV X7 (rvalue (word 4294967295)) *)
  0x8a0200e7;       (* arm_AND X7 X7 X2 *)
  0xba070084;       (* arm_ADCS X4 X4 X7 *)
  0xba1f00a5;       (* arm_ADCS X5 X5 XZR *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x8a0200e7;       (* arm_AND X7 X7 X2 *)
  0x9a0700c6;       (* arm_ADC X6 X6 X7 *)
  0xa9001003;       (* arm_STP X3 X4 X0 (Immediate_Offset (iword (&0))) *)
  0xa9011805;       (* arm_STP X5 X6 X0 (Immediate_Offset (iword (&16))) *)
  0xd65f03c0;       (* arm_RET X30 *)
  0xd10383ff;       (* arm_SUB SP SP (rvalue (word 224)) *)
  0xaa0003ef;       (* arm_MOV X15 X0 *)
  0xaa0103f0;       (* arm_MOV X16 X1 *)
  0xaa0203f1;       (* arm_MOV X17 X2 *)
  0xa9440e02;       (* arm_LDP X2 X3 X16 (Immediate_Offset (iword (&64))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa9451604;       (* arm_LDP X4 X5 X16 (Immediate_Offset (iword (&80))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d02;       (* arm_MUL X2 X8 X3 *)
  0x9bc37d08;       (* arm_UMULH X8 X8 X3 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d22;       (* arm_MUL X2 X9 X3 *)
  0x9bc37d29;       (* arm_UMULH X9 X9 X3 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d42;       (* arm_MUL X2 X10 X3 *)
  0x9bc37d4a;       (* arm_UMULH X10 X10 X3 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d62;       (* arm_MUL X2 X11 X3 *)
  0x9bc37d6b;       (* arm_UMULH X11 X11 X3 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x92800002;       (* arm_MOVN X2 (word 0) 0 *)
  0x9a8233e2;       (* arm_CSEL X2 XZR X2 Condition_CC *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0x9a8333e3;       (* arm_CSEL X3 XZR X3 Condition_CC *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9a8533e5;       (* arm_CSEL X5 XZR X5 Condition_CC *)
  0xeb020108;       (* arm_SUBS X8 X8 X2 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda05016b;       (* arm_SBC X11 X11 X5 *)
  0xa90027e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&0))) *)
  0xa9012fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&16))) *)
  0xa9440e22;       (* arm_LDP X2 X3 X17 (Immediate_Offset (iword (&64))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa9451624;       (* arm_LDP X4 X5 X17 (Immediate_Offset (iword (&80))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d02;       (* arm_MUL X2 X8 X3 *)
  0x9bc37d08;       (* arm_UMULH X8 X8 X3 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d22;       (* arm_MUL X2 X9 X3 *)
  0x9bc37d29;       (* arm_UMULH X9 X9 X3 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d42;       (* arm_MUL X2 X10 X3 *)
  0x9bc37d4a;       (* arm_UMULH X10 X10 X3 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d62;       (* arm_MUL X2 X11 X3 *)
  0x9bc37d6b;       (* arm_UMULH X11 X11 X3 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x92800002;       (* arm_MOVN X2 (word 0) 0 *)
  0x9a8233e2;       (* arm_CSEL X2 XZR X2 Condition_CC *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0x9a8333e3;       (* arm_CSEL X3 XZR X3 Condition_CC *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9a8533e5;       (* arm_CSEL X5 XZR X5 Condition_CC *)
  0xeb020108;       (* arm_SUBS X8 X8 X2 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda05016b;       (* arm_SBC X11 X11 X5 *)
  0xa90a27e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&160))) *)
  0xa90b2fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&176))) *)
  0xa9441223;       (* arm_LDP X3 X4 X17 (Immediate_Offset (iword (&64))) *)
  0xa9422207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&32))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9432a09;       (* arm_LDP X9 X10 X16 (Immediate_Offset (iword (&48))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9451a25;       (* arm_LDP X5 X6 X17 (Immediate_Offset (iword (&80))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90c37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&192))) *)
  0xa90d03ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&208))) *)
  0xa9441203;       (* arm_LDP X3 X4 X16 (Immediate_Offset (iword (&64))) *)
  0xa9422227;       (* arm_LDP X7 X8 X17 (Immediate_Offset (iword (&32))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9432a29;       (* arm_LDP X9 X10 X17 (Immediate_Offset (iword (&48))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9451a05;       (* arm_LDP X5 X6 X16 (Immediate_Offset (iword (&80))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90237ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&32))) *)
  0xa90303ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&48))) *)
  0xa94013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&0))) *)
  0xa9402227;       (* arm_LDP X7 X8 X17 (Immediate_Offset (iword (&0))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9412a29;       (* arm_LDP X9 X10 X17 (Immediate_Offset (iword (&16))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9411be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&16))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90437ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&64))) *)
  0xa90503ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&80))) *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0xa9402207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&0))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9412a09;       (* arm_LDP X9 X10 X16 (Immediate_Offset (iword (&16))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa94b1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&176))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90837ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&128))) *)
  0xa90903ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&144))) *)
  0xa94013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&0))) *)
  0xa94223e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&32))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9432be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&48))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9411be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&16))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90237ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&32))) *)
  0xa90303ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&48))) *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0xa94c23e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&192))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa94d2be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&208))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa94b1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&176))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90c37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&192))) *)
  0xa90d03ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&208))) *)
  0xa9441be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa9480fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&128))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94523e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa9490fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&144))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa90a1be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&160))) *)
  0xa90b23e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&176))) *)
  0xa9421be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&32))) *)
  0xa94c0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&192))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94323e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&48))) *)
  0xa94d0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&208))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9021be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&32))) *)
  0xa90323e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&48))) *)
  0xa94a0fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&160))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa94b17e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&176))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d02;       (* arm_MUL X2 X8 X3 *)
  0x9bc37d08;       (* arm_UMULH X8 X8 X3 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d22;       (* arm_MUL X2 X9 X3 *)
  0x9bc37d29;       (* arm_UMULH X9 X9 X3 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d42;       (* arm_MUL X2 X10 X3 *)
  0x9bc37d4a;       (* arm_UMULH X10 X10 X3 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d62;       (* arm_MUL X2 X11 X3 *)
  0x9bc37d6b;       (* arm_UMULH X11 X11 X3 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x92800002;       (* arm_MOVN X2 (word 0) 0 *)
  0x9a8233e2;       (* arm_CSEL X2 XZR X2 Condition_CC *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0x9a8333e3;       (* arm_CSEL X3 XZR X3 Condition_CC *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9a8533e5;       (* arm_CSEL X5 XZR X5 Condition_CC *)
  0xeb020108;       (* arm_SUBS X8 X8 X2 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda05016b;       (* arm_SBC X11 X11 X5 *)
  0xa90627e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&96))) *)
  0xa9072fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&112))) *)
  0xa9420fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&32))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa94317e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&48))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d02;       (* arm_MUL X2 X8 X3 *)
  0x9bc37d08;       (* arm_UMULH X8 X8 X3 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d22;       (* arm_MUL X2 X9 X3 *)
  0x9bc37d29;       (* arm_UMULH X9 X9 X3 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d42;       (* arm_MUL X2 X10 X3 *)
  0x9bc37d4a;       (* arm_UMULH X10 X10 X3 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d62;       (* arm_MUL X2 X11 X3 *)
  0x9bc37d6b;       (* arm_UMULH X11 X11 X3 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0xb100050c;       (* arm_ADDS X12 X8 (rvalue (word 1)) *)
  0xfa03012d;       (* arm_SBCS X13 X9 X3 *)
  0xfa1f014e;       (* arm_SBCS X14 X10 XZR *)
  0xfa050167;       (* arm_SBCS X7 X11 X5 *)
  0xfa1f005f;       (* arm_SBCS XZR X2 XZR *)
  0x9a8c3108;       (* arm_CSEL X8 X8 X12 Condition_CC *)
  0x9a8d3129;       (* arm_CSEL X9 X9 X13 Condition_CC *)
  0x9a8e314a;       (* arm_CSEL X10 X10 X14 Condition_CC *)
  0x9a87316b;       (* arm_CSEL X11 X11 X7 Condition_CC *)
  0xa90027e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&0))) *)
  0xa9012fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&16))) *)
  0xa94613e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&96))) *)
  0xa94823e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&128))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9492be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&144))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9471be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&112))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90837ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&128))) *)
  0xa90903ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&144))) *)
  0xa94613e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&96))) *)
  0xa94423e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&64))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9452be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&80))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9471be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&112))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90437ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&64))) *)
  0xa90503ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&80))) *)
  0xa9401be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa9480fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&128))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94123e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9490fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&144))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9001be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa90123e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9441be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa9480fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&128))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94523e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa9490fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&144))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9061be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&96))) *)
  0xa90723e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&112))) *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0xa9442207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&64))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9452a09;       (* arm_LDP X9 X10 X16 (Immediate_Offset (iword (&80))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa94b1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&176))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90a37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&160))) *)
  0xa90b03ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&176))) *)
  0xa9401be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa9440fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&64))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94123e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9450fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&80))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9001be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa90123e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9481be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&128))) *)
  0xa9400fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&0))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94923e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&144))) *)
  0xa9410fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&16))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9081be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&128))) *)
  0xa90923e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&144))) *)
  0xa94613e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&96))) *)
  0xa94c23e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&192))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa94d2be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&208))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9471be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&112))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90637ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&96))) *)
  0xa90703ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&112))) *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0xa9442227;       (* arm_LDP X7 X8 X17 (Immediate_Offset (iword (&64))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9452a29;       (* arm_LDP X9 X10 X17 (Immediate_Offset (iword (&80))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa94b1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&176))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90a37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&160))) *)
  0xa90b03ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&176))) *)
  0xa94213e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&32))) *)
  0xa94823e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&128))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9492be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&144))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9431be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&48))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90837ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&128))) *)
  0xa90903ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&144))) *)
  0xa9481be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&128))) *)
  0xa9460fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&96))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94923e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&144))) *)
  0xa9470fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&112))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9081be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&128))) *)
  0xa90923e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&144))) *)
  0xa9440600;       (* arm_LDP X0 X1 X16 (Immediate_Offset (iword (&64))) *)
  0xa9450e02;       (* arm_LDP X2 X3 X16 (Immediate_Offset (iword (&80))) *)
  0xaa01000c;       (* arm_ORR X12 X0 X1 *)
  0xaa03004d;       (* arm_ORR X13 X2 X3 *)
  0xaa0d018c;       (* arm_ORR X12 X12 X13 *)
  0xeb1f019f;       (* arm_CMP X12 XZR *)
  0x9a9f07ec;       (* arm_CSET X12 Condition_NE *)
  0xa9441624;       (* arm_LDP X4 X5 X17 (Immediate_Offset (iword (&64))) *)
  0xa9451e26;       (* arm_LDP X6 X7 X17 (Immediate_Offset (iword (&80))) *)
  0xaa05008d;       (* arm_ORR X13 X4 X5 *)
  0xaa0700ce;       (* arm_ORR X14 X6 X7 *)
  0xaa0e01ad;       (* arm_ORR X13 X13 X14 *)
  0xeb1f01bf;       (* arm_CMP X13 XZR *)
  0x9a9f07ed;       (* arm_CSET X13 Condition_NE *)
  0xeb0c01bf;       (* arm_CMP X13 X12 *)
  0xa94a27e8;       (* arm_LDP X8 X9 SP (Immediate_Offset (iword (&160))) *)
  0x9a883008;       (* arm_CSEL X8 X0 X8 Condition_CC *)
  0x9a893029;       (* arm_CSEL X9 X1 X9 Condition_CC *)
  0x9a888088;       (* arm_CSEL X8 X4 X8 Condition_HI *)
  0x9a8980a9;       (* arm_CSEL X9 X5 X9 Condition_HI *)
  0xa94b2fea;       (* arm_LDP X10 X11 SP (Immediate_Offset (iword (&176))) *)
  0x9a8a304a;       (* arm_CSEL X10 X2 X10 Condition_CC *)
  0x9a8b306b;       (* arm_CSEL X11 X3 X11 Condition_CC *)
  0x9a8a80ca;       (* arm_CSEL X10 X6 X10 Condition_HI *)
  0x9a8b80eb;       (* arm_CSEL X11 X7 X11 Condition_HI *)
  0xa940360c;       (* arm_LDP X12 X13 X16 (Immediate_Offset (iword (&0))) *)
  0xa94007e0;       (* arm_LDP X0 X1 SP (Immediate_Offset (iword (&0))) *)
  0x9a803180;       (* arm_CSEL X0 X12 X0 Condition_CC *)
  0x9a8131a1;       (* arm_CSEL X1 X13 X1 Condition_CC *)
  0xa940362c;       (* arm_LDP X12 X13 X17 (Immediate_Offset (iword (&0))) *)
  0x9a808180;       (* arm_CSEL X0 X12 X0 Condition_HI *)
  0x9a8181a1;       (* arm_CSEL X1 X13 X1 Condition_HI *)
  0xa941360c;       (* arm_LDP X12 X13 X16 (Immediate_Offset (iword (&16))) *)
  0xa9410fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&16))) *)
  0x9a823182;       (* arm_CSEL X2 X12 X2 Condition_CC *)
  0x9a8331a3;       (* arm_CSEL X3 X13 X3 Condition_CC *)
  0xa941362c;       (* arm_LDP X12 X13 X17 (Immediate_Offset (iword (&16))) *)
  0x9a828182;       (* arm_CSEL X2 X12 X2 Condition_HI *)
  0x9a8381a3;       (* arm_CSEL X3 X13 X3 Condition_HI *)
  0xa942360c;       (* arm_LDP X12 X13 X16 (Immediate_Offset (iword (&32))) *)
  0xa94817e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&128))) *)
  0x9a843184;       (* arm_CSEL X4 X12 X4 Condition_CC *)
  0x9a8531a5;       (* arm_CSEL X5 X13 X5 Condition_CC *)
  0xa942362c;       (* arm_LDP X12 X13 X17 (Immediate_Offset (iword (&32))) *)
  0x9a848184;       (* arm_CSEL X4 X12 X4 Condition_HI *)
  0x9a8581a5;       (* arm_CSEL X5 X13 X5 Condition_HI *)
  0xa943360c;       (* arm_LDP X12 X13 X16 (Immediate_Offset (iword (&48))) *)
  0xa9491fe6;       (* arm_LDP X6 X7 SP (Immediate_Offset (iword (&144))) *)
  0x9a863186;       (* arm_CSEL X6 X12 X6 Condition_CC *)
  0x9a8731a7;       (* arm_CSEL X7 X13 X7 Condition_CC *)
  0xa943362c;       (* arm_LDP X12 X13 X17 (Immediate_Offset (iword (&48))) *)
  0x9a868186;       (* arm_CSEL X6 X12 X6 Condition_HI *)
  0x9a8781a7;       (* arm_CSEL X7 X13 X7 Condition_HI *)
  0xa90005e0;       (* arm_STP X0 X1 X15 (Immediate_Offset (iword (&0))) *)
  0xa9010de2;       (* arm_STP X2 X3 X15 (Immediate_Offset (iword (&16))) *)
  0xa90215e4;       (* arm_STP X4 X5 X15 (Immediate_Offset (iword (&32))) *)
  0xa9031de6;       (* arm_STP X6 X7 X15 (Immediate_Offset (iword (&48))) *)
  0xa90425e8;       (* arm_STP X8 X9 X15 (Immediate_Offset (iword (&64))) *)
  0xa9052dea;       (* arm_STP X10 X11 X15 (Immediate_Offset (iword (&80))) *)
  0x910383ff;       (* arm_ADD SP SP (rvalue (word 224)) *)
  0xd65f03c0;       (* arm_RET X30 *)
  0xd10303ff;       (* arm_SUB SP SP (rvalue (word 192)) *)
  0xaa0003ef;       (* arm_MOV X15 X0 *)
  0xaa0103f0;       (* arm_MOV X16 X1 *)
  0xa9440e02;       (* arm_LDP X2 X3 X16 (Immediate_Offset (iword (&64))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa9451604;       (* arm_LDP X4 X5 X16 (Immediate_Offset (iword (&80))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0x9b057d02;       (* arm_MUL X2 X8 X5 *)
  0x9bc57d08;       (* arm_UMULH X8 X8 X5 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0x9b057d22;       (* arm_MUL X2 X9 X5 *)
  0x9bc57d29;       (* arm_UMULH X9 X9 X5 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0x9b057d42;       (* arm_MUL X2 X10 X5 *)
  0x9bc57d4a;       (* arm_UMULH X10 X10 X5 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9b057d62;       (* arm_MUL X2 X11 X5 *)
  0x9bc57d6b;       (* arm_UMULH X11 X11 X5 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0xb100050c;       (* arm_ADDS X12 X8 (rvalue (word 1)) *)
  0xfa03012d;       (* arm_SBCS X13 X9 X3 *)
  0xfa1f014e;       (* arm_SBCS X14 X10 XZR *)
  0xfa050167;       (* arm_SBCS X7 X11 X5 *)
  0xfa1f005f;       (* arm_SBCS XZR X2 XZR *)
  0x9a8c3108;       (* arm_CSEL X8 X8 X12 Condition_CC *)
  0x9a8d3129;       (* arm_CSEL X9 X9 X13 Condition_CC *)
  0x9a8e314a;       (* arm_CSEL X10 X10 X14 Condition_CC *)
  0x9a87316b;       (* arm_CSEL X11 X11 X7 Condition_CC *)
  0xa90027e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&0))) *)
  0xa9012fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&16))) *)
  0xa9420e02;       (* arm_LDP X2 X3 X16 (Immediate_Offset (iword (&32))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa9431604;       (* arm_LDP X4 X5 X16 (Immediate_Offset (iword (&48))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0x9b057d02;       (* arm_MUL X2 X8 X5 *)
  0x9bc57d08;       (* arm_UMULH X8 X8 X5 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0x9b057d22;       (* arm_MUL X2 X9 X5 *)
  0x9bc57d29;       (* arm_UMULH X9 X9 X5 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0x9b057d42;       (* arm_MUL X2 X10 X5 *)
  0x9bc57d4a;       (* arm_UMULH X10 X10 X5 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9b057d62;       (* arm_MUL X2 X11 X5 *)
  0x9bc57d6b;       (* arm_UMULH X11 X11 X5 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0xb100050c;       (* arm_ADDS X12 X8 (rvalue (word 1)) *)
  0xfa03012d;       (* arm_SBCS X13 X9 X3 *)
  0xfa1f014e;       (* arm_SBCS X14 X10 XZR *)
  0xfa050167;       (* arm_SBCS X7 X11 X5 *)
  0xfa1f005f;       (* arm_SBCS XZR X2 XZR *)
  0x9a8c3108;       (* arm_CSEL X8 X8 X12 Condition_CC *)
  0x9a8d3129;       (* arm_CSEL X9 X9 X13 Condition_CC *)
  0x9a8e314a;       (* arm_CSEL X10 X10 X14 Condition_CC *)
  0x9a87316b;       (* arm_CSEL X11 X11 X7 Condition_CC *)
  0xa90227e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&32))) *)
  0xa9032fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&48))) *)
  0xa9401a05;       (* arm_LDP X5 X6 X16 (Immediate_Offset (iword (&0))) *)
  0xa9400fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&0))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa9412207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&16))) *)
  0xa9410fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&16))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0x92407c64;       (* arm_AND X4 X3 (rvalue (word 4294967295)) *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0x92608064;       (* arm_AND X4 X3 (rvalue (word 18446744069414584321)) *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9061be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&96))) *)
  0xa90723e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&112))) *)
  0xa9401a05;       (* arm_LDP X5 X6 X16 (Immediate_Offset (iword (&0))) *)
  0xa9400fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&0))) *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xba0300c6;       (* arm_ADCS X6 X6 X3 *)
  0xa9412207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&16))) *)
  0xa9410fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&16))) *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xda9f33e3;       (* arm_CSETM X3 Condition_CS *)
  0xeb0300a5;       (* arm_SUBS X5 X5 X3 *)
  0x92407c61;       (* arm_AND X1 X3 (rvalue (word 4294967295)) *)
  0xfa0100c6;       (* arm_SBCS X6 X6 X1 *)
  0xfa1f00e7;       (* arm_SBCS X7 X7 XZR *)
  0x92608062;       (* arm_AND X2 X3 (rvalue (word 18446744069414584321)) *)
  0xda020108;       (* arm_SBC X8 X8 X2 *)
  0xa9041be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa90523e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa94413e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&64))) *)
  0xa94623e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&96))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9472be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&112))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9451be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&80))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90637ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&96))) *)
  0xa90703ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&112))) *)
  0xa9421a05;       (* arm_LDP X5 X6 X16 (Immediate_Offset (iword (&32))) *)
  0xa9440e04;       (* arm_LDP X4 X3 X16 (Immediate_Offset (iword (&64))) *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xba0300c6;       (* arm_ADCS X6 X6 X3 *)
  0xa9432207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&48))) *)
  0xa9450e04;       (* arm_LDP X4 X3 X16 (Immediate_Offset (iword (&80))) *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xb10004bf;       (* arm_CMN X5 (rvalue (word 1)) *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0xfa0400df;       (* arm_SBCS XZR X6 X4 *)
  0xfa1f00ff;       (* arm_SBCS XZR X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0xfa04011f;       (* arm_SBCS XZR X8 X4 *)
  0xba1f0063;       (* arm_ADCS X3 X3 XZR *)
  0xda9f03e3;       (* arm_CSETM X3 Condition_NE *)
  0xeb0300a5;       (* arm_SUBS X5 X5 X3 *)
  0x92407c64;       (* arm_AND X4 X3 (rvalue (word 4294967295)) *)
  0xfa0400c6;       (* arm_SBCS X6 X6 X4 *)
  0xfa1f00e7;       (* arm_SBCS X7 X7 XZR *)
  0x92608064;       (* arm_AND X4 X3 (rvalue (word 18446744069414584321)) *)
  0xda040108;       (* arm_SBC X8 X8 X4 *)
  0xa9041be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa90523e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa9401203;       (* arm_LDP X3 X4 X16 (Immediate_Offset (iword (&0))) *)
  0xa94223e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&32))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9432be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&48))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9411a05;       (* arm_LDP X5 X6 X16 (Immediate_Offset (iword (&16))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90837ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&128))) *)
  0xa90903ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&144))) *)
  0xa9460fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&96))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa94717e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&112))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0x9b057d02;       (* arm_MUL X2 X8 X5 *)
  0x9bc57d08;       (* arm_UMULH X8 X8 X5 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0x9b057d22;       (* arm_MUL X2 X9 X5 *)
  0x9bc57d29;       (* arm_UMULH X9 X9 X5 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0x9b057d42;       (* arm_MUL X2 X10 X5 *)
  0x9bc57d4a;       (* arm_UMULH X10 X10 X5 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9b057d62;       (* arm_MUL X2 X11 X5 *)
  0x9bc57d6b;       (* arm_UMULH X11 X11 X5 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0xb100050c;       (* arm_ADDS X12 X8 (rvalue (word 1)) *)
  0xfa03012d;       (* arm_SBCS X13 X9 X3 *)
  0xfa1f014e;       (* arm_SBCS X14 X10 XZR *)
  0xfa050167;       (* arm_SBCS X7 X11 X5 *)
  0xfa1f005f;       (* arm_SBCS XZR X2 XZR *)
  0x9a8c3108;       (* arm_CSEL X8 X8 X12 Condition_CC *)
  0x9a8d3129;       (* arm_CSEL X9 X9 X13 Condition_CC *)
  0x9a8e314a;       (* arm_CSEL X10 X10 X14 Condition_CC *)
  0x9a87316b;       (* arm_CSEL X11 X11 X7 Condition_CC *)
  0xa90a27e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&160))) *)
  0xa90b2fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&176))) *)
  0xa9440fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&64))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa94517e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&80))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0x9b057d02;       (* arm_MUL X2 X8 X5 *)
  0x9bc57d08;       (* arm_UMULH X8 X8 X5 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0x9b057d22;       (* arm_MUL X2 X9 X5 *)
  0x9bc57d29;       (* arm_UMULH X9 X9 X5 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0x9b057d42;       (* arm_MUL X2 X10 X5 *)
  0x9bc57d4a;       (* arm_UMULH X10 X10 X5 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9b057d62;       (* arm_MUL X2 X11 X5 *)
  0x9bc57d6b;       (* arm_UMULH X11 X11 X5 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0xb100050c;       (* arm_ADDS X12 X8 (rvalue (word 1)) *)
  0xfa03012d;       (* arm_SBCS X13 X9 X3 *)
  0xfa1f014e;       (* arm_SBCS X14 X10 XZR *)
  0xfa050167;       (* arm_SBCS X7 X11 X5 *)
  0xfa1f005f;       (* arm_SBCS XZR X2 XZR *)
  0x9a8c3108;       (* arm_CSEL X8 X8 X12 Condition_CC *)
  0x9a8d3129;       (* arm_CSEL X9 X9 X13 Condition_CC *)
  0x9a8e314a;       (* arm_CSEL X10 X10 X14 Condition_CC *)
  0x9a87316b;       (* arm_CSEL X11 X11 X7 Condition_CC *)
  0xa90427e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&64))) *)
  0xa9052fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&80))) *)
  0xd2800121;       (* arm_MOV X1 (rvalue (word 9)) *)
  0x92800002;       (* arm_MOVN X2 (word 0) 0 *)
  0xa94a2be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&160))) *)
  0xeb090049;       (* arm_SUBS X9 X2 X9 *)
  0xb2407fe2;       (* arm_MOV X2 (rvalue (word 4294967295)) *)
  0xfa0a004a;       (* arm_SBCS X10 X2 X10 *)
  0xa94b33eb;       (* arm_LDP X11 X12 SP (Immediate_Offset (iword (&176))) *)
  0xfa0b03eb;       (* arm_NGCS X11 X11 *)
  0xb26083e2;       (* arm_MOV X2 (rvalue (word 18446744069414584321)) *)
  0xda0c004c;       (* arm_SBC X12 X2 X12 *)
  0x9b097c23;       (* arm_MUL X3 X1 X9 *)
  0x9b0a7c24;       (* arm_MUL X4 X1 X10 *)
  0x9b0b7c25;       (* arm_MUL X5 X1 X11 *)
  0x9b0c7c26;       (* arm_MUL X6 X1 X12 *)
  0x9bc97c29;       (* arm_UMULH X9 X1 X9 *)
  0x9bca7c2a;       (* arm_UMULH X10 X1 X10 *)
  0x9bcb7c2b;       (* arm_UMULH X11 X1 X11 *)
  0x9bcc7c27;       (* arm_UMULH X7 X1 X12 *)
  0xab090084;       (* arm_ADDS X4 X4 X9 *)
  0xba0a00a5;       (* arm_ADCS X5 X5 X10 *)
  0xba0b00c6;       (* arm_ADCS X6 X6 X11 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xd2800181;       (* arm_MOV X1 (rvalue (word 12)) *)
  0xa9482be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&128))) *)
  0x9b017d28;       (* arm_MUL X8 X9 X1 *)
  0x9bc17d29;       (* arm_UMULH X9 X9 X1 *)
  0xab080063;       (* arm_ADDS X3 X3 X8 *)
  0x9b017d48;       (* arm_MUL X8 X10 X1 *)
  0x9bc17d4a;       (* arm_UMULH X10 X10 X1 *)
  0xba080084;       (* arm_ADCS X4 X4 X8 *)
  0xa94933eb;       (* arm_LDP X11 X12 SP (Immediate_Offset (iword (&144))) *)
  0x9b017d68;       (* arm_MUL X8 X11 X1 *)
  0x9bc17d6b;       (* arm_UMULH X11 X11 X1 *)
  0xba0800a5;       (* arm_ADCS X5 X5 X8 *)
  0x9b017d88;       (* arm_MUL X8 X12 X1 *)
  0x9bc17d8c;       (* arm_UMULH X12 X12 X1 *)
  0xba0800c6;       (* arm_ADCS X6 X6 X8 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab090084;       (* arm_ADDS X4 X4 X9 *)
  0xba0a00a5;       (* arm_ADCS X5 X5 X10 *)
  0xba0b00c6;       (* arm_ADCS X6 X6 X11 *)
  0x9a0c00e7;       (* arm_ADC X7 X7 X12 *)
  0x910004e8;       (* arm_ADD X8 X7 (rvalue (word 1)) *)
  0xd3607d0a;       (* arm_LSL X10 X8 32 *)
  0xab0a00c6;       (* arm_ADDS X6 X6 X10 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xcb0803e9;       (* arm_NEG X9 X8 *)
  0xd100054a;       (* arm_SUB X10 X10 (rvalue (word 1)) *)
  0xeb090063;       (* arm_SUBS X3 X3 X9 *)
  0xfa0a0084;       (* arm_SBCS X4 X4 X10 *)
  0xfa1f00a5;       (* arm_SBCS X5 X5 XZR *)
  0xfa0800c6;       (* arm_SBCS X6 X6 X8 *)
  0xda0800e8;       (* arm_SBC X8 X7 X8 *)
  0xab080063;       (* arm_ADDS X3 X3 X8 *)
  0x92407d09;       (* arm_AND X9 X8 (rvalue (word 4294967295)) *)
  0xba090084;       (* arm_ADCS X4 X4 X9 *)
  0xba1f00a5;       (* arm_ADCS X5 X5 XZR *)
  0xcb0903ea;       (* arm_NEG X10 X9 *)
  0x9a0a00c6;       (* arm_ADC X6 X6 X10 *)
  0xa90a13e3;       (* arm_STP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0xa90b1be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&176))) *)
  0xa9441be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa9400fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&0))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94523e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa9410fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&16))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0x92407c64;       (* arm_AND X4 X3 (rvalue (word 4294967295)) *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0x92608064;       (* arm_AND X4 X3 (rvalue (word 18446744069414584321)) *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9041be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa90523e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa9420fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&32))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa94317e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&48))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0x9b057d02;       (* arm_MUL X2 X8 X5 *)
  0x9bc57d08;       (* arm_UMULH X8 X8 X5 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0x9b057d22;       (* arm_MUL X2 X9 X5 *)
  0x9bc57d29;       (* arm_UMULH X9 X9 X5 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0x9b057d42;       (* arm_MUL X2 X10 X5 *)
  0x9bc57d4a;       (* arm_UMULH X10 X10 X5 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9b057d62;       (* arm_MUL X2 X11 X5 *)
  0x9bc57d6b;       (* arm_UMULH X11 X11 X5 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0xb100050c;       (* arm_ADDS X12 X8 (rvalue (word 1)) *)
  0xfa03012d;       (* arm_SBCS X13 X9 X3 *)
  0xfa1f014e;       (* arm_SBCS X14 X10 XZR *)
  0xfa050167;       (* arm_SBCS X7 X11 X5 *)
  0xfa1f005f;       (* arm_SBCS XZR X2 XZR *)
  0x9a8c3108;       (* arm_CSEL X8 X8 X12 Condition_CC *)
  0x9a8d3129;       (* arm_CSEL X9 X9 X13 Condition_CC *)
  0x9a8e314a;       (* arm_CSEL X10 X10 X14 Condition_CC *)
  0x9a87316b;       (* arm_CSEL X11 X11 X7 Condition_CC *)
  0xa90027e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&0))) *)
  0xa9012fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&16))) *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0xa94623e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&96))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9472be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&112))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa94b1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&176))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90637ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&96))) *)
  0xa90703ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&112))) *)
  0xa9441be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa9420fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&32))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94523e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa9430fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&48))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0x92407c64;       (* arm_AND X4 X3 (rvalue (word 4294967295)) *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0x92608064;       (* arm_AND X4 X3 (rvalue (word 18446744069414584321)) *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa90419e5;       (* arm_STP X5 X6 X15 (Immediate_Offset (iword (&64))) *)
  0xa90521e7;       (* arm_STP X7 X8 X15 (Immediate_Offset (iword (&80))) *)
  0xa9480be1;       (* arm_LDP X1 X2 SP (Immediate_Offset (iword (&128))) *)
  0xd37ef420;       (* arm_LSL X0 X1 2 *)
  0xa94a1fe6;       (* arm_LDP X6 X7 SP (Immediate_Offset (iword (&160))) *)
  0xeb060000;       (* arm_SUBS X0 X0 X6 *)
  0x93c1f841;       (* arm_EXTR X1 X2 X1 62 *)
  0xfa070021;       (* arm_SBCS X1 X1 X7 *)
  0xa94913e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&144))) *)
  0x93c2f862;       (* arm_EXTR X2 X3 X2 62 *)
  0xa94b1fe6;       (* arm_LDP X6 X7 SP (Immediate_Offset (iword (&176))) *)
  0xfa060042;       (* arm_SBCS X2 X2 X6 *)
  0x93c3f883;       (* arm_EXTR X3 X4 X3 62 *)
  0xfa070063;       (* arm_SBCS X3 X3 X7 *)
  0xd37efc84;       (* arm_LSR X4 X4 62 *)
  0xda1f0084;       (* arm_SBC X4 X4 XZR *)
  0x91000485;       (* arm_ADD X5 X4 (rvalue (word 1)) *)
  0xd3607ca8;       (* arm_LSL X8 X5 32 *)
  0xeb0803e6;       (* arm_NEGS X6 X8 *)
  0xfa1f03e7;       (* arm_NGCS X7 XZR *)
  0xda050108;       (* arm_SBC X8 X8 X5 *)
  0xab050000;       (* arm_ADDS X0 X0 X5 *)
  0xba060021;       (* arm_ADCS X1 X1 X6 *)
  0xba070042;       (* arm_ADCS X2 X2 X7 *)
  0xba080063;       (* arm_ADCS X3 X3 X8 *)
  0xda9f23e5;       (* arm_CSETM X5 Condition_CC *)
  0xab050000;       (* arm_ADDS X0 X0 X5 *)
  0x92407ca6;       (* arm_AND X6 X5 (rvalue (word 4294967295)) *)
  0xba060021;       (* arm_ADCS X1 X1 X6 *)
  0xba1f0042;       (* arm_ADCS X2 X2 XZR *)
  0xcb0603e7;       (* arm_NEG X7 X6 *)
  0x9a070063;       (* arm_ADC X3 X3 X7 *)
  0xa90005e0;       (* arm_STP X0 X1 X15 (Immediate_Offset (iword (&0))) *)
  0xa9010de2;       (* arm_STP X2 X3 X15 (Immediate_Offset (iword (&16))) *)
  0xd2800101;       (* arm_MOV X1 (rvalue (word 8)) *)
  0x92800002;       (* arm_MOVN X2 (word 0) 0 *)
  0xa9402be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&0))) *)
  0xeb090049;       (* arm_SUBS X9 X2 X9 *)
  0xb2407fe2;       (* arm_MOV X2 (rvalue (word 4294967295)) *)
  0xfa0a004a;       (* arm_SBCS X10 X2 X10 *)
  0xa94133eb;       (* arm_LDP X11 X12 SP (Immediate_Offset (iword (&16))) *)
  0xfa0b03eb;       (* arm_NGCS X11 X11 *)
  0xb26083e2;       (* arm_MOV X2 (rvalue (word 18446744069414584321)) *)
  0xda0c004c;       (* arm_SBC X12 X2 X12 *)
  0xd37df123;       (* arm_LSL X3 X9 3 *)
  0x93c9f544;       (* arm_EXTR X4 X10 X9 61 *)
  0x93caf565;       (* arm_EXTR X5 X11 X10 61 *)
  0x93cbf586;       (* arm_EXTR X6 X12 X11 61 *)
  0xd37dfd87;       (* arm_LSR X7 X12 61 *)
  0xd2800061;       (* arm_MOV X1 (rvalue (word 3)) *)
  0xa9462be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&96))) *)
  0x9b017d28;       (* arm_MUL X8 X9 X1 *)
  0x9bc17d29;       (* arm_UMULH X9 X9 X1 *)
  0xab080063;       (* arm_ADDS X3 X3 X8 *)
  0x9b017d48;       (* arm_MUL X8 X10 X1 *)
  0x9bc17d4a;       (* arm_UMULH X10 X10 X1 *)
  0xba080084;       (* arm_ADCS X4 X4 X8 *)
  0xa94733eb;       (* arm_LDP X11 X12 SP (Immediate_Offset (iword (&112))) *)
  0x9b017d68;       (* arm_MUL X8 X11 X1 *)
  0x9bc17d6b;       (* arm_UMULH X11 X11 X1 *)
  0xba0800a5;       (* arm_ADCS X5 X5 X8 *)
  0x9b017d88;       (* arm_MUL X8 X12 X1 *)
  0x9bc17d8c;       (* arm_UMULH X12 X12 X1 *)
  0xba0800c6;       (* arm_ADCS X6 X6 X8 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab090084;       (* arm_ADDS X4 X4 X9 *)
  0xba0a00a5;       (* arm_ADCS X5 X5 X10 *)
  0xba0b00c6;       (* arm_ADCS X6 X6 X11 *)
  0x9a0c00e7;       (* arm_ADC X7 X7 X12 *)
  0x910004e8;       (* arm_ADD X8 X7 (rvalue (word 1)) *)
  0xd3607d0a;       (* arm_LSL X10 X8 32 *)
  0xab0a00c6;       (* arm_ADDS X6 X6 X10 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xcb0803e9;       (* arm_NEG X9 X8 *)
  0xd100054a;       (* arm_SUB X10 X10 (rvalue (word 1)) *)
  0xeb090063;       (* arm_SUBS X3 X3 X9 *)
  0xfa0a0084;       (* arm_SBCS X4 X4 X10 *)
  0xfa1f00a5;       (* arm_SBCS X5 X5 XZR *)
  0xfa0800c6;       (* arm_SBCS X6 X6 X8 *)
  0xda0800e8;       (* arm_SBC X8 X7 X8 *)
  0xab080063;       (* arm_ADDS X3 X3 X8 *)
  0x92407d09;       (* arm_AND X9 X8 (rvalue (word 4294967295)) *)
  0xba090084;       (* arm_ADCS X4 X4 X9 *)
  0xba1f00a5;       (* arm_ADCS X5 X5 XZR *)
  0xcb0903ea;       (* arm_NEG X10 X9 *)
  0x9a0a00c6;       (* arm_ADC X6 X6 X10 *)
  0xa90211e3;       (* arm_STP X3 X4 X15 (Immediate_Offset (iword (&32))) *)
  0xa90319e5;       (* arm_STP X5 X6 X15 (Immediate_Offset (iword (&48))) *)
  0x910303ff;       (* arm_ADD SP SP (rvalue (word 192)) *)
  0xd65f03c0;       (* arm_RET X30 *)
  0xd10303ff;       (* arm_SUB SP SP (rvalue (word 192)) *)
  0xaa0003ef;       (* arm_MOV X15 X0 *)
  0xaa0103f0;       (* arm_MOV X16 X1 *)
  0xaa0203f1;       (* arm_MOV X17 X2 *)
  0xa9440e02;       (* arm_LDP X2 X3 X16 (Immediate_Offset (iword (&64))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa9451604;       (* arm_LDP X4 X5 X16 (Immediate_Offset (iword (&80))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d02;       (* arm_MUL X2 X8 X3 *)
  0x9bc37d08;       (* arm_UMULH X8 X8 X3 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d22;       (* arm_MUL X2 X9 X3 *)
  0x9bc37d29;       (* arm_UMULH X9 X9 X3 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d42;       (* arm_MUL X2 X10 X3 *)
  0x9bc37d4a;       (* arm_UMULH X10 X10 X3 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d62;       (* arm_MUL X2 X11 X3 *)
  0x9bc37d6b;       (* arm_UMULH X11 X11 X3 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x92800002;       (* arm_MOVN X2 (word 0) 0 *)
  0x9a8233e2;       (* arm_CSEL X2 XZR X2 Condition_CC *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0x9a8333e3;       (* arm_CSEL X3 XZR X3 Condition_CC *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9a8533e5;       (* arm_CSEL X5 XZR X5 Condition_CC *)
  0xeb020108;       (* arm_SUBS X8 X8 X2 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda05016b;       (* arm_SBC X11 X11 X5 *)
  0xa90027e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&0))) *)
  0xa9012fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&16))) *)
  0xa9441203;       (* arm_LDP X3 X4 X16 (Immediate_Offset (iword (&64))) *)
  0xa9422227;       (* arm_LDP X7 X8 X17 (Immediate_Offset (iword (&32))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9432a29;       (* arm_LDP X9 X10 X17 (Immediate_Offset (iword (&48))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9451a05;       (* arm_LDP X5 X6 X16 (Immediate_Offset (iword (&80))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90237ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&32))) *)
  0xa90303ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&48))) *)
  0xa94013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&0))) *)
  0xa9402227;       (* arm_LDP X7 X8 X17 (Immediate_Offset (iword (&0))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9412a29;       (* arm_LDP X9 X10 X17 (Immediate_Offset (iword (&16))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9411be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&16))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90437ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&64))) *)
  0xa90503ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&80))) *)
  0xa94013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&0))) *)
  0xa94223e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&32))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9432be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&48))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9411be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&16))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90237ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&32))) *)
  0xa90303ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&48))) *)
  0xa9441be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa9400e04;       (* arm_LDP X4 X3 X16 (Immediate_Offset (iword (&0))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94523e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa9410e04;       (* arm_LDP X4 X3 X16 (Immediate_Offset (iword (&16))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa90a1be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&160))) *)
  0xa90b23e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&176))) *)
  0xa9421be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&32))) *)
  0xa9420e04;       (* arm_LDP X4 X3 X16 (Immediate_Offset (iword (&32))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94323e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&48))) *)
  0xa9430e04;       (* arm_LDP X4 X3 X16 (Immediate_Offset (iword (&48))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9021be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&32))) *)
  0xa90323e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&48))) *)
  0xa94a0fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&160))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa94b17e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&176))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d02;       (* arm_MUL X2 X8 X3 *)
  0x9bc37d08;       (* arm_UMULH X8 X8 X3 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d22;       (* arm_MUL X2 X9 X3 *)
  0x9bc37d29;       (* arm_UMULH X9 X9 X3 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d42;       (* arm_MUL X2 X10 X3 *)
  0x9bc37d4a;       (* arm_UMULH X10 X10 X3 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d62;       (* arm_MUL X2 X11 X3 *)
  0x9bc37d6b;       (* arm_UMULH X11 X11 X3 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x92800002;       (* arm_MOVN X2 (word 0) 0 *)
  0x9a8233e2;       (* arm_CSEL X2 XZR X2 Condition_CC *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0x9a8333e3;       (* arm_CSEL X3 XZR X3 Condition_CC *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9a8533e5;       (* arm_CSEL X5 XZR X5 Condition_CC *)
  0xeb020108;       (* arm_SUBS X8 X8 X2 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda05016b;       (* arm_SBC X11 X11 X5 *)
  0xa90627e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&96))) *)
  0xa9072fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&112))) *)
  0xa9420fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&32))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa94317e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&48))) *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b047c46;       (* arm_MUL X6 X2 X4 *)
  0x9bc47c47;       (* arm_UMULH X7 X2 X4 *)
  0xab06014a;       (* arm_ADDS X10 X10 X6 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9b047c66;       (* arm_MUL X6 X3 X4 *)
  0x9bc47c67;       (* arm_UMULH X7 X3 X4 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06016b;       (* arm_ADDS X11 X11 X6 *)
  0x9b057c8d;       (* arm_MUL X13 X4 X5 *)
  0x9bc57c8e;       (* arm_UMULH X14 X4 X5 *)
  0xba07018c;       (* arm_ADCS X12 X12 X7 *)
  0x9b057c66;       (* arm_MUL X6 X3 X5 *)
  0x9bc57c67;       (* arm_UMULH X7 X3 X5 *)
  0x9a1f00e7;       (* arm_ADC X7 X7 XZR *)
  0xab06018c;       (* arm_ADDS X12 X12 X6 *)
  0xba0701ad;       (* arm_ADCS X13 X13 X7 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0x9a9f37e7;       (* arm_CSET X7 Condition_CS *)
  0x9bc27c46;       (* arm_UMULH X6 X2 X2 *)
  0x9b027c48;       (* arm_MUL X8 X2 X2 *)
  0xab060129;       (* arm_ADDS X9 X9 X6 *)
  0x9b037c66;       (* arm_MUL X6 X3 X3 *)
  0xba06014a;       (* arm_ADCS X10 X10 X6 *)
  0x9bc37c66;       (* arm_UMULH X6 X3 X3 *)
  0xba06016b;       (* arm_ADCS X11 X11 X6 *)
  0x9b047c86;       (* arm_MUL X6 X4 X4 *)
  0xba06018c;       (* arm_ADCS X12 X12 X6 *)
  0x9bc47c86;       (* arm_UMULH X6 X4 X4 *)
  0xba0601ad;       (* arm_ADCS X13 X13 X6 *)
  0x9b057ca6;       (* arm_MUL X6 X5 X5 *)
  0xba0601ce;       (* arm_ADCS X14 X14 X6 *)
  0x9bc57ca6;       (* arm_UMULH X6 X5 X5 *)
  0x9a0600e7;       (* arm_ADC X7 X7 X6 *)
  0xab088129;       (* arm_ADDS X9 X9 (Shiftedreg X8 LSL 32) *)
  0xd360fd03;       (* arm_LSR X3 X8 32 *)
  0xba03014a;       (* arm_ADCS X10 X10 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d02;       (* arm_MUL X2 X8 X3 *)
  0x9bc37d08;       (* arm_UMULH X8 X8 X3 *)
  0xba02016b;       (* arm_ADCS X11 X11 X2 *)
  0x9a1f0108;       (* arm_ADC X8 X8 XZR *)
  0xab09814a;       (* arm_ADDS X10 X10 (Shiftedreg X9 LSL 32) *)
  0xd360fd23;       (* arm_LSR X3 X9 32 *)
  0xba03016b;       (* arm_ADCS X11 X11 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d22;       (* arm_MUL X2 X9 X3 *)
  0x9bc37d29;       (* arm_UMULH X9 X9 X3 *)
  0xba020108;       (* arm_ADCS X8 X8 X2 *)
  0x9a1f0129;       (* arm_ADC X9 X9 XZR *)
  0xab0a816b;       (* arm_ADDS X11 X11 (Shiftedreg X10 LSL 32) *)
  0xd360fd43;       (* arm_LSR X3 X10 32 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d42;       (* arm_MUL X2 X10 X3 *)
  0x9bc37d4a;       (* arm_UMULH X10 X10 X3 *)
  0xba020129;       (* arm_ADCS X9 X9 X2 *)
  0x9a1f014a;       (* arm_ADC X10 X10 XZR *)
  0xab0b8108;       (* arm_ADDS X8 X8 (Shiftedreg X11 LSL 32) *)
  0xd360fd63;       (* arm_LSR X3 X11 32 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0xb26083e3;       (* arm_MOV X3 (rvalue (word 18446744069414584321)) *)
  0x9b037d62;       (* arm_MUL X2 X11 X3 *)
  0x9bc37d6b;       (* arm_UMULH X11 X11 X3 *)
  0xba02014a;       (* arm_ADCS X10 X10 X2 *)
  0x9a1f016b;       (* arm_ADC X11 X11 XZR *)
  0xab0c0108;       (* arm_ADDS X8 X8 X12 *)
  0xba0d0129;       (* arm_ADCS X9 X9 X13 *)
  0xba0e014a;       (* arm_ADCS X10 X10 X14 *)
  0xba07016b;       (* arm_ADCS X11 X11 X7 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0xb2407fe3;       (* arm_MOV X3 (rvalue (word 4294967295)) *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0xb100050c;       (* arm_ADDS X12 X8 (rvalue (word 1)) *)
  0xfa03012d;       (* arm_SBCS X13 X9 X3 *)
  0xfa1f014e;       (* arm_SBCS X14 X10 XZR *)
  0xfa050167;       (* arm_SBCS X7 X11 X5 *)
  0xfa1f005f;       (* arm_SBCS XZR X2 XZR *)
  0x9a8c3108;       (* arm_CSEL X8 X8 X12 Condition_CC *)
  0x9a8d3129;       (* arm_CSEL X9 X9 X13 Condition_CC *)
  0x9a8e314a;       (* arm_CSEL X10 X10 X14 Condition_CC *)
  0x9a87316b;       (* arm_CSEL X11 X11 X7 Condition_CC *)
  0xa90027e8;       (* arm_STP X8 X9 SP (Immediate_Offset (iword (&0))) *)
  0xa9012fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&16))) *)
  0xa94613e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&96))) *)
  0xa9402207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&0))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9412a09;       (* arm_LDP X9 X10 X16 (Immediate_Offset (iword (&16))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9471be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&112))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90837ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&128))) *)
  0xa90903ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&144))) *)
  0xa94613e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&96))) *)
  0xa94423e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&64))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9452be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&80))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9471be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&112))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90437ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&64))) *)
  0xa90503ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&80))) *)
  0xa9401be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa9480fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&128))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94123e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9490fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&144))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9001be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa90123e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9441be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&64))) *)
  0xa9480fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&128))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94523e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&80))) *)
  0xa9490fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&144))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9061be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&96))) *)
  0xa90723e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&112))) *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0xa9442207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&64))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9452a09;       (* arm_LDP X9 X10 X16 (Immediate_Offset (iword (&80))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa94b1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&176))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90a37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&160))) *)
  0xa90b03ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&176))) *)
  0xa9401be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa9440fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&64))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94123e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9450fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&80))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9001be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa90123e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9481be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&128))) *)
  0xa9400fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&0))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94923e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&144))) *)
  0xa9410fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&16))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9081be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&128))) *)
  0xa90923e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&144))) *)
  0xa94613e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&96))) *)
  0xa9422207;       (* arm_LDP X7 X8 X16 (Immediate_Offset (iword (&32))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9432a09;       (* arm_LDP X9 X10 X16 (Immediate_Offset (iword (&48))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9471be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&112))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90637ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&96))) *)
  0xa90703ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&112))) *)
  0xa94213e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&32))) *)
  0xa94823e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&128))) *)
  0x9b077c6c;       (* arm_MUL X12 X3 X7 *)
  0x9bc77c6d;       (* arm_UMULH X13 X3 X7 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c6e;       (* arm_UMULH X14 X3 X8 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9492be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&144))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c60;       (* arm_UMULH X0 X3 X9 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c61;       (* arm_UMULH X1 X3 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f0021;       (* arm_ADC X1 X1 XZR *)
  0xa9431be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&48))) *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bca7c83;       (* arm_UMULH X3 X4 X10 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9a1f0063;       (* arm_ADC X3 X3 XZR *)
  0x9b077cab;       (* arm_MUL X11 X5 X7 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b087cab;       (* arm_MUL X11 X5 X8 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b097cab;       (* arm_MUL X11 X5 X9 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b0a7cab;       (* arm_MUL X11 X5 X10 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bca7ca4;       (* arm_UMULH X4 X5 X10 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9bc77cab;       (* arm_UMULH X11 X5 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9bc87cab;       (* arm_UMULH X11 X5 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9bc97cab;       (* arm_UMULH X11 X5 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9a1f0084;       (* arm_ADC X4 X4 XZR *)
  0x9b077ccb;       (* arm_MUL X11 X6 X7 *)
  0xab0b0000;       (* arm_ADDS X0 X0 X11 *)
  0x9b087ccb;       (* arm_MUL X11 X6 X8 *)
  0xba0b0021;       (* arm_ADCS X1 X1 X11 *)
  0x9b097ccb;       (* arm_MUL X11 X6 X9 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9b0a7ccb;       (* arm_MUL X11 X6 X10 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9bca7cc5;       (* arm_UMULH X5 X6 X10 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xb26083ea;       (* arm_MOV X10 (rvalue (word 18446744069414584321)) *)
  0xab0c81ad;       (* arm_ADDS X13 X13 (Shiftedreg X12 LSL 32) *)
  0xd360fd8b;       (* arm_LSR X11 X12 32 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b0a7d8b;       (* arm_MUL X11 X12 X10 *)
  0x9bca7d8c;       (* arm_UMULH X12 X12 X10 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9a1f018c;       (* arm_ADC X12 X12 XZR *)
  0x9bc77ccb;       (* arm_UMULH X11 X6 X7 *)
  0xab0b0021;       (* arm_ADDS X1 X1 X11 *)
  0x9bc87ccb;       (* arm_UMULH X11 X6 X8 *)
  0xba0b0063;       (* arm_ADCS X3 X3 X11 *)
  0x9bc97ccb;       (* arm_UMULH X11 X6 X9 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f00a5;       (* arm_ADC X5 X5 XZR *)
  0xab0d81ce;       (* arm_ADDS X14 X14 (Shiftedreg X13 LSL 32) *)
  0xd360fdab;       (* arm_LSR X11 X13 32 *)
  0xba0b0000;       (* arm_ADCS X0 X0 X11 *)
  0x9b0a7dab;       (* arm_MUL X11 X13 X10 *)
  0x9bca7dad;       (* arm_UMULH X13 X13 X10 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xab0e8000;       (* arm_ADDS X0 X0 (Shiftedreg X14 LSL 32) *)
  0xd360fdcb;       (* arm_LSR X11 X14 32 *)
  0xba0b018c;       (* arm_ADCS X12 X12 X11 *)
  0x9b0a7dcb;       (* arm_MUL X11 X14 X10 *)
  0x9bca7dce;       (* arm_UMULH X14 X14 X10 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9a1f01ce;       (* arm_ADC X14 X14 XZR *)
  0xab00818c;       (* arm_ADDS X12 X12 (Shiftedreg X0 LSL 32) *)
  0xd360fc0b;       (* arm_LSR X11 X0 32 *)
  0xba0b01ad;       (* arm_ADCS X13 X13 X11 *)
  0x9b0a7c0b;       (* arm_MUL X11 X0 X10 *)
  0x9bca7c00;       (* arm_UMULH X0 X0 X10 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9a1f0000;       (* arm_ADC X0 X0 XZR *)
  0xab01018c;       (* arm_ADDS X12 X12 X1 *)
  0xba0301ad;       (* arm_ADCS X13 X13 X3 *)
  0xba0401ce;       (* arm_ADCS X14 X14 X4 *)
  0xba050000;       (* arm_ADCS X0 X0 X5 *)
  0x9a9f37e8;       (* arm_CSET X8 Condition_CS *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xb1000581;       (* arm_ADDS X1 X12 (rvalue (word 1)) *)
  0xfa0b01a3;       (* arm_SBCS X3 X13 X11 *)
  0xfa1f01c4;       (* arm_SBCS X4 X14 XZR *)
  0xfa0a0005;       (* arm_SBCS X5 X0 X10 *)
  0xfa1f011f;       (* arm_SBCS XZR X8 XZR *)
  0x9a81318c;       (* arm_CSEL X12 X12 X1 Condition_CC *)
  0x9a8331ad;       (* arm_CSEL X13 X13 X3 Condition_CC *)
  0x9a8431ce;       (* arm_CSEL X14 X14 X4 Condition_CC *)
  0x9a853000;       (* arm_CSEL X0 X0 X5 Condition_CC *)
  0xa90837ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&128))) *)
  0xa90903ee;       (* arm_STP X14 X0 SP (Immediate_Offset (iword (&144))) *)
  0xa9481be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&128))) *)
  0xa9460fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&96))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94923e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&144))) *)
  0xa9470fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&112))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0xba1f00e7;       (* arm_ADCS X7 X7 XZR *)
  0xb26083e4;       (* arm_MOV X4 (rvalue (word 18446744069414584321)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0x9a040108;       (* arm_ADC X8 X8 X4 *)
  0xa9081be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&128))) *)
  0xa90923e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&144))) *)
  0xa9440600;       (* arm_LDP X0 X1 X16 (Immediate_Offset (iword (&64))) *)
  0xa9450e02;       (* arm_LDP X2 X3 X16 (Immediate_Offset (iword (&80))) *)
  0xaa010004;       (* arm_ORR X4 X0 X1 *)
  0xaa030045;       (* arm_ORR X5 X2 X3 *)
  0xaa050084;       (* arm_ORR X4 X4 X5 *)
  0xeb1f009f;       (* arm_CMP X4 XZR *)
  0xa94007e0;       (* arm_LDP X0 X1 SP (Immediate_Offset (iword (&0))) *)
  0xa940362c;       (* arm_LDP X12 X13 X17 (Immediate_Offset (iword (&0))) *)
  0x9a8c1000;       (* arm_CSEL X0 X0 X12 Condition_NE *)
  0x9a8d1021;       (* arm_CSEL X1 X1 X13 Condition_NE *)
  0xa9410fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&16))) *)
  0xa941362c;       (* arm_LDP X12 X13 X17 (Immediate_Offset (iword (&16))) *)
  0x9a8c1042;       (* arm_CSEL X2 X2 X12 Condition_NE *)
  0x9a8d1063;       (* arm_CSEL X3 X3 X13 Condition_NE *)
  0xa94817e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&128))) *)
  0xa942362c;       (* arm_LDP X12 X13 X17 (Immediate_Offset (iword (&32))) *)
  0x9a8c1084;       (* arm_CSEL X4 X4 X12 Condition_NE *)
  0x9a8d10a5;       (* arm_CSEL X5 X5 X13 Condition_NE *)
  0xa9491fe6;       (* arm_LDP X6 X7 SP (Immediate_Offset (iword (&144))) *)
  0xa943362c;       (* arm_LDP X12 X13 X17 (Immediate_Offset (iword (&48))) *)
  0x9a8c10c6;       (* arm_CSEL X6 X6 X12 Condition_NE *)
  0x9a8d10e7;       (* arm_CSEL X7 X7 X13 Condition_NE *)
  0xa94a27e8;       (* arm_LDP X8 X9 SP (Immediate_Offset (iword (&160))) *)
  0xd280002c;       (* arm_MOV X12 (rvalue (word 1)) *)
  0xb2607fed;       (* arm_MOV X13 (rvalue (word 18446744069414584320)) *)
  0x9a8c1108;       (* arm_CSEL X8 X8 X12 Condition_NE *)
  0x9a8d1129;       (* arm_CSEL X9 X9 X13 Condition_NE *)
  0xa94b2fea;       (* arm_LDP X10 X11 SP (Immediate_Offset (iword (&176))) *)
  0x9280000c;       (* arm_MOVN X12 (word 0) 0 *)
  0xb27f7bed;       (* arm_MOV X13 (rvalue (word 4294967294)) *)
  0x9a8c114a;       (* arm_CSEL X10 X10 X12 Condition_NE *)
  0x9a8d116b;       (* arm_CSEL X11 X11 X13 Condition_NE *)
  0xa90005e0;       (* arm_STP X0 X1 X15 (Immediate_Offset (iword (&0))) *)
  0xa9010de2;       (* arm_STP X2 X3 X15 (Immediate_Offset (iword (&16))) *)
  0xa90215e4;       (* arm_STP X4 X5 X15 (Immediate_Offset (iword (&32))) *)
  0xa9031de6;       (* arm_STP X6 X7 X15 (Immediate_Offset (iword (&48))) *)
  0xa90425e8;       (* arm_STP X8 X9 X15 (Immediate_Offset (iword (&64))) *)
  0xa9052dea;       (* arm_STP X10 X11 X15 (Immediate_Offset (iword (&80))) *)
  0x910303ff;       (* arm_ADD SP SP (rvalue (word 192)) *)
  0xd65f03c0        (* arm_RET X30 *)
];;

let P256_SCALARMUL_ALT_EXEC = ARM_MK_EXEC_RULE p256_scalarmul_alt_mc;;

(* ------------------------------------------------------------------------- *)
(* Local versions of the subroutines.                                        *)
(* ------------------------------------------------------------------------- *)

let LOCAL_DEMONT_TAC =
  ARM_SUBROUTINE_SIM_TAC
   (p256_scalarmul_alt_mc,P256_SCALARMUL_ALT_EXEC,
    0x828,bignum_demont_p256_mc,BIGNUM_DEMONT_P256_SUBROUTINE_CORRECT)
  [`read X0 s`; `read X1 s`;
   `read(memory :> bytes(read X1 s,8 * 4)) s`;
   `pc + 0x828`; `read X30 s`];;

let LOCAL_INV_TAC =
  ARM_SUBROUTINE_SIM_TAC
   (p256_scalarmul_alt_mc,P256_SCALARMUL_ALT_EXEC,
    0x8bc,bignum_inv_p256_mc,BIGNUM_INV_P256_SUBROUTINE_CORRECT)
  [`read X0 s`; `read X1 s`;
   `read(memory :> bytes(read X1 s,8 * 4)) s`;
   `pc + 0x8bc`; `read SP s`; `read X30 s`];;

let LOCAL_MUL_TAC =
  ARM_SUBROUTINE_SIM_TAC
   (p256_scalarmul_alt_mc,P256_SCALARMUL_ALT_EXEC,
    0x1980,bignum_montmul_p256_alt_mc,
    BIGNUM_MONTMUL_P256_ALT_SUBROUTINE_CORRECT)
  [`read X0 s`; `read X1 s`; `read X2 s`;
   `read(memory :> bytes(read X1 s,8 * 4)) s`;
   `read(memory :> bytes(read X2 s,8 * 4)) s`;
   `pc + 0x1980`; `read X30 s`];;

let LOCAL_SQR_TAC =
  ARM_SUBROUTINE_SIM_TAC
   (p256_scalarmul_alt_mc,P256_SCALARMUL_ALT_EXEC,
    0x1b48,bignum_montsqr_p256_alt_mc,
    BIGNUM_MONTSQR_P256_ALT_SUBROUTINE_CORRECT)
  [`read X0 s`; `read X1 s`;
   `read(memory :> bytes(read X1 s,8 * 4)) s`;
   `pc + 0x1b48`; `read X30 s`];;

let LOCAL_TOMONT_TAC =
  ARM_SUBROUTINE_SIM_TAC
   (p256_scalarmul_alt_mc,P256_SCALARMUL_ALT_EXEC,
    0x1cb8,bignum_tomont_p256_mc,BIGNUM_TOMONT_P256_SUBROUTINE_CORRECT)
  [`read X0 s`; `read X1 s`;
   `read(memory :> bytes(read X1 s,8 * 4)) s`;
   `pc + 0x1cb8`; `read X30 s`];;

let LOCAL_JADD_TAC =
  let th =
    CONV_RULE(ONCE_DEPTH_CONV NUM_MULT_CONV)
      (REWRITE_RULE[bignum_triple_from_memory; bignum_pair_from_memory]
       P256_MONTJADD_ALT_SUBROUTINE_CORRECT) in
  ARM_SUBROUTINE_SIM_TAC
   (p256_scalarmul_alt_mc,P256_SCALARMUL_ALT_EXEC,
    0x1ea8,p256_montjadd_alt_mc,th)
  [`read X0 s`; `read X1 s`;
   `read(memory :> bytes(read X1 s,8 * 4)) s,
    read(memory :> bytes(word_add (read X1 s) (word 32),8 * 4)) s,
    read(memory :> bytes(word_add (read X1 s) (word 64),8 * 4)) s`;
   `read X2 s`;
   `read(memory :> bytes(read X2 s,8 * 4)) s,
    read(memory :> bytes(word_add (read X2 s) (word 32),8 * 4)) s,
    read(memory :> bytes(word_add (read X2 s) (word 64),8 * 4)) s`;
   `pc + 0x1ea8`; `read SP s`; `read X30 s`];;

let LOCAL_JDOUBLE_TAC =
  let th =
    CONV_RULE(ONCE_DEPTH_CONV NUM_MULT_CONV)
      (REWRITE_RULE[bignum_triple_from_memory; bignum_pair_from_memory]
       P256_MONTJDOUBLE_ALT_SUBROUTINE_CORRECT) in
  ARM_SUBROUTINE_SIM_TAC
   (p256_scalarmul_alt_mc,P256_SCALARMUL_ALT_EXEC,
    0x3cc8,p256_montjdouble_alt_mc,th)
  [`read X0 s`; `read X1 s`;
   `read(memory :> bytes(read X1 s,8 * 4)) s,
    read(memory :> bytes(word_add (read X1 s) (word 32),8 * 4)) s,
    read(memory :> bytes(word_add (read X1 s) (word 64),8 * 4)) s`;
   `pc + 0x3cc8`; `read SP s`; `read X30 s`];;

let LOCAL_JMIXADD_TAC =
  let th =
    CONV_RULE(ONCE_DEPTH_CONV NUM_MULT_CONV)
      (REWRITE_RULE[bignum_triple_from_memory; bignum_pair_from_memory]
       P256_MONTJMIXADD_ALT_SUBROUTINE_CORRECT) in
  ARM_SUBROUTINE_SIM_TAC
   (p256_scalarmul_alt_mc,P256_SCALARMUL_ALT_EXEC,
    0x4d04,p256_montjmixadd_alt_mc,th)
  [`read X0 s`; `read X1 s`;
   `read(memory :> bytes(read X1 s,8 * 4)) s,
    read(memory :> bytes(word_add (read X1 s) (word 32),8 * 4)) s,
    read(memory :> bytes(word_add (read X1 s) (word 64),8 * 4)) s`;
   `read X2 s`;
   `read(memory :> bytes(read X2 s,8 * 4)) s,
    read(memory :> bytes(word_add (read X2 s) (word 32),8 * 4)) s`;
   `pc + 0x4d04`; `read SP s`; `read X30 s`];;

(* ------------------------------------------------------------------------- *)
(* Overall point operation proof.                                            *)
(* ------------------------------------------------------------------------- *)

let affinepoint_p256 = new_definition
 `affinepoint_p256 (x,y) P <=>
        P = SOME(paired (modular_decode (256,p_256)) (x,y))`;;

let affinepointz_p256 = new_definition
 `affinepointz_p256 (x,y) P <=>
        if x = 0 /\ y = 0 then P = NONE
        else P = SOME(paired (modular_decode (256,p_256)) (x,y))`;;

let REPRESENTS2_P256_NONZERO = prove
 (`!P x y. represents2_p256 P (x,y) ==> ~(P = group_id p256_group)`,
  REWRITE_TAC[represents2_p256; P256_GROUP] THEN MESON_TAC[option_DISTINCT]);;

let REPRESENTS_P256_2 = prove
 (`!P x y. represents_p256 P (x,y,(2 EXP 256) MOD p_256) <=>
           represents2_p256 P (x,y)`,
  REPEAT GEN_TAC THEN REWRITE_TAC[represents2_p256; represents_p256] THEN
  REWRITE_TAC[paired; tripled; weierstrass_of_jacobian] THEN
  REWRITE_TAC[montgomery_decode; INTEGER_MOD_RING_CLAUSES; p_256] THEN
  CONV_TAC(DEPTH_CONV(NUM_RED_CONV ORELSEC INVERSE_MOD_CONV)) THEN
  CONV_TAC INT_REDUCE_CONV THEN
  REWRITE_TAC[ring_div; RING_INV_INTEGER_MOD_RING;
              INTEGER_MOD_RING_CLAUSES; COPRIME_1] THEN
  CONV_TAC(DEPTH_CONV(NUM_RED_CONV ORELSEC INVERSE_MOD_CONV)) THEN
  REWRITE_TAC[INT_MUL_RID] THEN CONV_TAC INT_REM_DOWN_CONV THEN
  REWRITE_TAC[]);;

let REPRESENTS_P256_NEGATION = prove
 (`!P x y z.
        represents_p256 P (x,y,z)
        ==> P IN group_carrier p256_group /\ ~(P = group_id p256_group)
            ==> represents_p256 (group_inv p256_group P) (x,p_256 - y,z)`,
  REWRITE_TAC[FORALL_OPTION_THM; FORALL_PAIR_THM; P256_GROUP] THEN
  REWRITE_TAC[IN; option_DISTINCT] THEN REWRITE_TAC[weierstrass_curve] THEN
  REWRITE_TAC[represents_p256; weierstrass_of_jacobian; tripled] THEN
  MAP_EVERY X_GEN_TAC [`X:int`; `Y:int`; `x:num`; `y:num`; `z:num`] THEN
  COND_CASES_TAC THEN ASM_REWRITE_TAC[option_DISTINCT] THEN
  POP_ASSUM(K ALL_TAC) THEN
  REWRITE_TAC[montgomery_decode; option_INJ; PAIR_EQ; weierstrass_neg] THEN
  REWRITE_TAC[INTEGER_MOD_RING_CLAUSES] THEN
  CONV_TAC INT_REM_DOWN_CONV THEN REWRITE_TAC[INT_REM_EQ] THEN
  REWRITE_TAC[INT_ARITH `a + --(&3) * b + c:int = a - &3 * b + c`] THEN
  ONCE_REWRITE_TAC[INT_CONG_SYM] THEN
  ASM_CASES_TAC `Y:int = &0` THENL
   [ASM_REWRITE_TAC[] THEN CONV_TAC INT_REDUCE_CONV THEN
    REWRITE_TAC[NO_ROOTS_P256];
    ALL_TAC] THEN
  ASM_CASES_TAC `y = 0` THENL
   [ASM_REWRITE_TAC[ring_div; RING_INV_INTEGER_MOD_RING; INT_OF_NUM_CLAUSES;
                    INT_OF_NUM_REM; INTEGER_MOD_RING_CLAUSES] THEN
    REWRITE_TAC[MULT_CLAUSES; p_256] THEN CONV_TAC NUM_REDUCE_CONV THEN
    ASM_REWRITE_TAC[INT_MUL_LZERO; INT_REM_ZERO];
    ALL_TAC] THEN
  REPLICATE_TAC 3 (DISCH_THEN(CONJUNCTS_THEN2 ASSUME_TAC MP_TAC)) THEN
  DISCH_THEN(CONJUNCTS_THEN(SUBST1_TAC o SYM)) THEN
  DISCH_THEN(K ALL_TAC) THEN ASM_REWRITE_TAC[] THEN
  CONJ_TAC THENL [ASM_ARITH_TAC; REWRITE_TAC[ring_div]] THEN
  REWRITE_TAC[RING_INV_INTEGER_MOD_RING; INT_OF_NUM_CLAUSES;
              INT_OF_NUM_REM; INTEGER_MOD_RING_CLAUSES] THEN
  COND_CASES_TAC THEN
  ASM_REWRITE_TAC[INT_MUL_RZERO; INT_REM_ZERO; INT_NEG_0] THEN
  REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES; GSYM INT_OF_NUM_REM] THEN
  ASM_SIMP_TAC[GSYM INT_OF_NUM_SUB; LT_IMP_LE] THEN
  CONV_TAC INT_REM_DOWN_CONV THEN REWRITE_TAC[INT_REM_EQ] THEN
  REWRITE_TAC[INTEGER_RULE
   `((i * (p - y)) * j:int == --((i * y) * j)) (mod p)`]);;

let REPRESENTS_P256_Y_NONZERO = prove
 (`!P x y z.
        represents_p256 P (x,y,z)
        ==> P IN group_carrier p256_group /\ ~(z = 0)
            ==> ~(y = 0)`,
  REPEAT STRIP_TAC THEN
  FIRST_ASSUM(MP_TAC o MATCH_MP REPRESENTS_P256_NEGATION) THEN
  FIRST_ASSUM(MP_TAC o GEN_REWRITE_RULE I [represents_p256]) THEN
  REPEAT(DISCH_THEN(CONJUNCTS_THEN2 ASSUME_TAC MP_TAC)) THEN
  ASM_REWRITE_TAC[montgomery_decode; tripled; weierstrass_of_jacobian] THEN
  REWRITE_TAC[represents_p256; SUB_0; LT_REFL; P256_GROUP] THEN
  REWRITE_TAC[INTEGER_MOD_RING_CLAUSES; INT_REM_EQ_0] THEN
  MATCH_MP_TAC(MESON[]
   `~(x = y) /\ ~p ==> (if p then x else y) = a ==> ~(a = x)`) THEN
  REWRITE_TAC[option_DISTINCT; GSYM INT_OF_NUM_CLAUSES] THEN
  DISCH_THEN(MP_TAC o SPEC `(&2:int) pow 256` o MATCH_MP (INTEGER_RULE
    `p divides i * z ==> !q:int. (q * i == &1) (mod p) ==> p divides z`)) THEN
  REWRITE_TAC[GSYM num_divides; INT_OF_NUM_CLAUSES; GSYM num_congruent] THEN
  REWRITE_TAC[INVERSE_MOD_RMUL_EQ; COPRIME_REXP; COPRIME_2; NOT_IMP] THEN
  CONJ_TAC THENL [REWRITE_TAC[p_256] THEN ARITH_TAC; ALL_TAC] THEN
  DISCH_THEN(MP_TAC o MATCH_MP DIVIDES_LE) THEN
  MP_TAC p_256 THEN ASM_ARITH_TAC);;

let unilemma0 = prove
 (`x = a MOD p_256 ==> x < p_256 /\ &x = &a rem &p_256`,
  REWRITE_TAC[INT_OF_NUM_REM; p_256] THEN ARITH_TAC);;

let unilemma1 = prove
 (`&x = a rem &p_256 ==> x < p_256 /\ &x = a rem &p_256`,
  SIMP_TAC[GSYM INT_OF_NUM_LT; INT_LT_REM_EQ; p_256] THEN INT_ARITH_TAC);;

let fdivlemma = prove
  (`!f a b c:A.
        field f /\
        a IN ring_carrier f /\ b IN ring_carrier f /\ c IN ring_carrier f /\
        ~(b = ring_0 f) /\
        ring_mul f b c = a
        ==> ring_div f a b = c`,
  FIELD_TAC);;

let P256_SCALARMUL_ALT_CORRECT = time prove
 (`!res scalar point n xy pc stackpointer.
        aligned 16 stackpointer /\
        ALL (nonoverlapping (stackpointer,1216))
            [(word pc,0x624c); (res,64); (scalar,32); (point,64)] /\
        nonoverlapping (res,64) (word pc,0x624c)
        ==> ensures arm
             (\s. aligned_bytes_loaded s (word pc) p256_scalarmul_alt_mc /\
                  read PC s = word(pc + 0xc) /\
                  read SP s = word_add stackpointer (word 224) /\
                  C_ARGUMENTS [res;scalar;point] s /\
                  bignum_from_memory (scalar,4) s = n /\
                  bignum_pair_from_memory (point,4) s = xy)
             (\s. read PC s = word (pc + 0x818) /\
                  !P. P IN group_carrier p256_group /\
                      affinepoint_p256 xy P
                      ==> affinepointz_p256
                            (bignum_pair_from_memory(res,4) s)
                            (group_pow p256_group P n))
          (MAYCHANGE_REGS_AND_FLAGS_PERMITTED_BY_ABI ,,
           MAYCHANGE [X19; X20; X21; X30] ,,
           MAYCHANGE [memory :> bytes(res,64);
                      memory :> bytes(stackpointer,1216)])`,
  REWRITE_TAC[FORALL_PAIR_THM] THEN
  REWRITE_TAC[GSYM SEQ_ASSOC; MAYCHANGE_REGS_AND_FLAGS_PERMITTED_BY_ABI] THEN
  MAP_EVERY X_GEN_TAC
   [`res:int64`; `scalar:int64`; `point:int64`;
    `n_input:num`; `x:num`; `y:num`;
    `pc:num`; `stackpointer:int64`] THEN
  REWRITE_TAC[ALLPAIRS; ALL; NONOVERLAPPING_CLAUSES] THEN STRIP_TAC THEN
  REWRITE_TAC[C_ARGUMENTS; SOME_FLAGS; PAIR_EQ; bignum_pair_from_memory] THEN
  CONV_TAC(ONCE_DEPTH_CONV NUM_MULT_CONV) THEN
  CONV_TAC(ONCE_DEPTH_CONV NORMALIZE_RELATIVE_ADDRESS_CONV) THEN

  (*** Modified input arguments, mathematically first ***)

  ABBREV_TAC `n_red = n_input MOD n_256` THEN
  SUBGOAL_THEN `n_red < n_256` ASSUME_TAC THENL
   [EXPAND_TAC "n_red" THEN REWRITE_TAC[n_256] THEN ARITH_TAC; ALL_TAC] THEN

  ABBREV_TAC `sgn <=> 2 EXP 255 <= n_red` THEN
  ABBREV_TAC `n_neg = if sgn then n_256 - n_red else n_red` THEN
  SUBGOAL_THEN `n_neg < 2 EXP 255` ASSUME_TAC THENL
   [MAP_EVERY EXPAND_TAC ["n_neg"; "sgn"] THEN
    UNDISCH_TAC `n_red < n_256` THEN REWRITE_TAC[n_256] THEN ARITH_TAC;
    ALL_TAC] THEN

  ABBREV_TAC
   `recoder =
    0x0888888888888888888888888888888888888888888888888888888888888888` THEN
  ABBREV_TAC `n = n_neg + recoder` THEN
  SUBGOAL_THEN `n < 9 * 2 EXP 252` ASSUME_TAC THENL
   [MAP_EVERY EXPAND_TAC ["n"; "recoder"] THEN
    UNDISCH_TAC `n_neg < 2 EXP 255` THEN ARITH_TAC;
    ALL_TAC] THEN

  ABBREV_TAC `x' = (2 EXP 256 * x) MOD p_256` THEN
  ABBREV_TAC `y'' = (2 EXP 256 * y) MOD p_256` THEN
  SUBGOAL_THEN `y'' < p_256` ASSUME_TAC THENL
   [EXPAND_TAC "y''" THEN REWRITE_TAC[MOD_LT_EQ; p_256; ARITH_EQ];
    ALL_TAC] THEN
  ABBREV_TAC `y':num = if sgn then p_256 - y'' else y''` THEN
  ABBREV_TAC `z' = (2 EXP 256) MOD p_256` THEN

  (*** Main loop invariant setup ***)

  ENSURES_WHILE_DOWN_TAC `63` `pc + 0x454` `pc + 0x7b0`
   `\i s.
      read SP s = word_add stackpointer (word 224) /\
      read X19 s = res /\
      read X20 s = word (4 * i) /\
      bignum_from_memory(word_add stackpointer (word 224),4) s = n /\
      !P. P IN group_carrier p256_group /\ represents2_p256 P (x',y')
          ==> represents_p256
                (group_zpow p256_group P
                    (&(n DIV 2 EXP (4 * i)) - &(recoder DIV 2 EXP (4 * i))))
                (bignum_triple_from_memory
                     (word_add stackpointer (word 256),4) s) /\
              !i. i < 8
                  ==> represents_p256 (group_pow p256_group P (i + 1))
                       (bignum_triple_from_memory
                       (word_add stackpointer (word (96 * i + 448)),4) s)` THEN
  REPEAT CONJ_TAC THENL
   [ARITH_TAC;

    (*** Initial holding of invariant ***)
    (*** First, the input reduced modulo the group order ***)

    REWRITE_TAC[BIGNUM_FROM_MEMORY_BYTES] THEN ENSURES_INIT_TAC "s0" THEN
    BIGNUM_LDIGITIZE_TAC "nin_" `read (memory :> bytes(scalar,8 * 4)) s0` THEN
    ARM_ACCSTEPS_TAC P256_SCALARMUL_ALT_EXEC (15--18) (1--22) THEN
    MAP_EVERY REABBREV_TAC
     [`nr_0 = read X2 s22`;
      `nr_1 = read X3 s22`;
      `nr_2 = read X4 s22`;
      `nr_3 = read X5 s22`] THEN
    SUBGOAL_THEN `bignum_of_wordlist[nr_0; nr_1; nr_2; nr_3] = n_red`
    ASSUME_TAC THENL
     [SUBGOAL_THEN `carry_s18 <=> n_input < n_256` SUBST_ALL_TAC THENL
       [MATCH_MP_TAC FLAG_FROM_CARRY_LT THEN EXISTS_TAC `256` THEN
        EXPAND_TAC "n_input" THEN
        REWRITE_TAC[n_256; GSYM REAL_OF_NUM_CLAUSES] THEN
        ACCUMULATOR_ASSUM_LIST(MP_TAC o end_itlist CONJ o DECARRY_RULE) THEN
        DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN BOUNDER_TAC[];
        ALL_TAC] THEN
      EXPAND_TAC "n_red" THEN
      W(MP_TAC o PART_MATCH (lhand o rand) MOD_CASES o rand o snd) THEN
      ANTS_TAC THENL
       [REWRITE_TAC[n_256] THEN EXPAND_TAC "n_input" THEN BOUNDER_TAC[];
        DISCH_THEN SUBST1_TAC] THEN
      MAP_EVERY EXPAND_TAC ["nr_0"; "nr_1"; "nr_2"; "nr_3"] THEN
      COND_CASES_TAC THEN ASM_REWRITE_TAC[] THEN
      RULE_ASSUM_TAC(REWRITE_RULE[NOT_LT]) THEN
      ASM_SIMP_TAC[GSYM REAL_OF_NUM_EQ; GSYM REAL_OF_NUM_SUB] THEN
      EXPAND_TAC "n_input" THEN REWRITE_TAC[GSYM REAL_OF_NUM_ADD] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_POW; GSYM REAL_OF_NUM_MUL] THEN
      MATCH_MP_TAC EQUAL_FROM_CONGRUENT_REAL THEN
      MAP_EVERY EXISTS_TAC [`256`; `&0:real`] THEN ASM_REWRITE_TAC[] THEN
      CONJ_TAC THENL [BOUNDER_TAC[]; ALL_TAC] THEN CONJ_TAC THENL
       [ASM_REWRITE_TAC[REAL_OF_NUM_ADD; REAL_OF_NUM_POW; REAL_OF_NUM_MUL] THEN
        ASM_REWRITE_TAC[REAL_SUB_LE; REAL_OF_NUM_LE] THEN
        MATCH_MP_TAC(REAL_ARITH `x:real < y ==> x - &n < y`) THEN
        REWRITE_TAC[REAL_OF_NUM_LT] THEN EXPAND_TAC "n_input" THEN
        BOUNDER_TAC[];
        ALL_TAC] THEN
      CONJ_TAC THENL [REAL_INTEGER_TAC; ALL_TAC] THEN
      EXPAND_TAC "n_input" THEN
      REWRITE_TAC[bignum_of_wordlist; n_256; GSYM REAL_OF_NUM_CLAUSES] THEN
      ACCUMULATOR_POP_ASSUM_LIST(MP_TAC o end_itlist CONJ o DESUM_RULE) THEN
      REWRITE_TAC[REAL_BITVAL_NOT; n_256] THEN
      DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN
      CONV_TAC(RAND_CONV REAL_POLY_CONV) THEN REAL_INTEGER_TAC;
      ACCUMULATOR_POP_ASSUM_LIST(K ALL_TAC)] THEN

    (*** Conditional negation of the scalar ***)

    ARM_ACCSTEPS_TAC P256_SCALARMUL_ALT_EXEC (23--26) (23--32) THEN
    MAP_EVERY REABBREV_TAC
     [`nn_0 = read X2 s32`;
      `nn_1 = read X3 s32`;
      `nn_2 = read X4 s32`;
      `nn_3 = read X5 s32`] THEN

    SUBGOAL_THEN
     `val(word_and nr_3 (word 9223372036854775808):int64) = 0 <=> ~sgn`
    SUBST_ALL_TAC THENL
     [MAP_EVERY EXPAND_TAC ["sgn"; "n_red"] THEN
      REWRITE_TAC[ARITH_RULE `~(2 EXP 255 <= b) <=> b DIV 2 EXP 255 = 0`] THEN
      CONV_TAC(ONCE_DEPTH_CONV BIGNUM_OF_WORDLIST_DIV_CONV) THEN
      CONV_TAC WORD_BLAST;
      RULE_ASSUM_TAC(REWRITE_RULE[COND_SWAP; GSYM WORD_BITVAL])] THEN

    SUBGOAL_THEN `bignum_of_wordlist[nn_0; nn_1; nn_2; nn_3] = n_neg`
    ASSUME_TAC THENL
     [MAP_EVERY EXPAND_TAC ["n_neg"; "nn_0"; "nn_1"; "nn_2"; "nn_3"] THEN
      COND_CASES_TAC THEN ASM_REWRITE_TAC[] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES] THEN
      MATCH_MP_TAC EQUAL_FROM_CONGRUENT_REAL THEN
      MAP_EVERY EXISTS_TAC [`256`; `&0:real`] THEN ASM_REWRITE_TAC[] THEN
      CONJ_TAC THENL [BOUNDER_TAC[]; ALL_TAC] THEN CONJ_TAC THENL
       [UNDISCH_TAC `n_red < n_256` THEN
        REWRITE_TAC[REAL_OF_NUM_CLAUSES; LE_0; n_256] THEN ARITH_TAC;
        ALL_TAC] THEN
      CONJ_TAC THENL [REAL_INTEGER_TAC; ALL_TAC] THEN
      ASM_SIMP_TAC[GSYM REAL_OF_NUM_SUB; LT_IMP_LE] THEN
      EXPAND_TAC "n_red" THEN
      REWRITE_TAC[bignum_of_wordlist; n_256; GSYM REAL_OF_NUM_CLAUSES] THEN
      ACCUMULATOR_POP_ASSUM_LIST(MP_TAC o end_itlist CONJ o DESUM_RULE) THEN
      REWRITE_TAC[REAL_BITVAL_NOT; n_256] THEN
      DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN
      CONV_TAC(RAND_CONV REAL_POLY_CONV) THEN REAL_INTEGER_TAC;
      ACCUMULATOR_POP_ASSUM_LIST(K ALL_TAC)] THEN

    (*** Add the recoding constant ***)

    ARM_ACCSTEPS_TAC P256_SCALARMUL_ALT_EXEC [34;35;37;38] (33--40) THEN
    SUBGOAL_THEN
     `read (memory :> bytes(word_add stackpointer (word 224),8 * 4)) s40 = n`
    ASSUME_TAC THENL
     [CONV_TAC(LAND_CONV BIGNUM_LEXPAND_CONV) THEN ASM_REWRITE_TAC[] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES] THEN
      MATCH_MP_TAC EQUAL_FROM_CONGRUENT_REAL THEN
      MAP_EVERY EXISTS_TAC [`256`; `&0:real`] THEN ASM_REWRITE_TAC[] THEN
      CONJ_TAC THENL [BOUNDER_TAC[]; ALL_TAC] THEN CONJ_TAC THENL
       [REWRITE_TAC[REAL_OF_NUM_CLAUSES] THEN
        UNDISCH_TAC `n < 9 * 2 EXP 252` THEN ARITH_TAC;
        ALL_TAC] THEN
      CONJ_TAC THENL [REAL_INTEGER_TAC; ALL_TAC] THEN
      ASM_SIMP_TAC[GSYM REAL_OF_NUM_SUB; LT_IMP_LE] THEN
      MAP_EVERY EXPAND_TAC ["n"; "recoder"; "n_neg"] THEN
      REWRITE_TAC[bignum_of_wordlist; n_256; GSYM REAL_OF_NUM_CLAUSES] THEN
      ACCUMULATOR_POP_ASSUM_LIST(MP_TAC o end_itlist CONJ o DESUM_RULE) THEN
      REWRITE_TAC[REAL_BITVAL_NOT; n_256] THEN
      DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN
      CONV_TAC(RAND_CONV REAL_POLY_CONV) THEN REAL_INTEGER_TAC;
      ACCUMULATOR_POP_ASSUM_LIST(K ALL_TAC)] THEN

    (*** Convert to Montgomery-Jacobian form ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (41--43) THEN LOCAL_TOMONT_TAC 44 THEN
    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (45--47) THEN LOCAL_TOMONT_TAC 48 THEN
    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (49--54) THEN
    SUBGOAL_THEN
     `read(memory :> bytes(word_add stackpointer (word 512),8 * 4)) s54 = z'`
    ASSUME_TAC THENL
     [CONV_TAC(LAND_CONV BIGNUM_LEXPAND_CONV) THEN
      ASM_REWRITE_TAC[bignum_of_wordlist] THEN EXPAND_TAC "z'" THEN
      REWRITE_TAC[p_256] THEN CONV_TAC(DEPTH_CONV WORD_NUM_RED_CONV);
      ALL_TAC] THEN

    (*** The conditional negation of y ***)

    BIGNUM_LDIGITIZE_TAC "ym_"
     `read (memory :> bytes(word_add stackpointer (word 480),8 * 4)) s54` THEN
    ARM_ACCSTEPS_TAC P256_SCALARMUL_ALT_EXEC [58;60;62;63] (55--70) THEN
    SUBGOAL_THEN
     `read(memory :> bytes(word_add stackpointer (word 480),8 * 4)) s70 = y'`
    ASSUME_TAC THENL
     [CONV_TAC(LAND_CONV BIGNUM_LEXPAND_CONV) THEN ASM_REWRITE_TAC[] THEN
      REWRITE_TAC[WORD_SUB_0; VAL_WORD_BITVAL; BITVAL_EQ_0] THEN
      EXPAND_TAC "y'" THEN COND_CASES_TAC THEN ASM_REWRITE_TAC[] THEN
      MATCH_MP_TAC CONG_IMP_EQ THEN EXISTS_TAC `2 EXP 256` THEN
      CONJ_TAC THENL [BOUNDER_TAC[]; ALL_TAC] THEN
      CONJ_TAC THENL [REWRITE_TAC[p_256] THEN ARITH_TAC; ALL_TAC] THEN
      ASM_SIMP_TAC[num_congruent; GSYM INT_OF_NUM_SUB; LT_IMP_LE] THEN
      REWRITE_TAC[REAL_INT_CONGRUENCE] THEN
      REWRITE_TAC[GSYM REAL_OF_INT_CLAUSES] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; REAL_POW_EQ_0] THEN
      REWRITE_TAC[REAL_OF_NUM_EQ; ARITH_EQ] THEN EXPAND_TAC "y''" THEN
      REWRITE_TAC[bignum_of_wordlist; p_256; GSYM REAL_OF_NUM_CLAUSES] THEN
      ACCUMULATOR_POP_ASSUM_LIST(MP_TAC o end_itlist CONJ o DESUM_RULE) THEN
      DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN REAL_INTEGER_TAC;
      ACCUMULATOR_POP_ASSUM_LIST(K ALL_TAC)] THEN

    (*** Computation of 2 * P ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (71--73) THEN
    LOCAL_JDOUBLE_TAC 74 THEN
    MAP_EVERY ABBREV_TAC
     [`x2 = read (memory :> bytes(word_add stackpointer (word 544),8 * 4)) s74`;
      `y2 = read (memory :> bytes(word_add stackpointer (word 576),8 * 4)) s74`;
      `z2 = read (memory :> bytes(word_add stackpointer (word 608),8 * 4)) s74`
     ] THEN

    (*** Computation of 3 * P ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (75--78) THEN LOCAL_JMIXADD_TAC 79 THEN
    MAP_EVERY ABBREV_TAC
     [`x3 = read (memory :> bytes(word_add stackpointer (word 640),8 * 4)) s79`;
      `y3 = read (memory :> bytes(word_add stackpointer (word 672),8 * 4)) s79`;
      `z3 = read (memory :> bytes(word_add stackpointer (word 704),8 * 4)) s79`
     ] THEN

    (*** Computation of 4 * P ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (80--82) THEN LOCAL_JDOUBLE_TAC 83 THEN
    MAP_EVERY ABBREV_TAC
     [`x4 = read (memory :> bytes(word_add stackpointer (word 736),8 * 4)) s83`;
      `y4 = read (memory :> bytes(word_add stackpointer (word 768),8 * 4)) s83`;
      `z4 = read (memory :> bytes(word_add stackpointer (word 800),8 * 4)) s83`
     ] THEN

    (*** Computation of 5 * P ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (84--87) THEN LOCAL_JMIXADD_TAC 88 THEN
    MAP_EVERY ABBREV_TAC
     [`x5 = read (memory :> bytes(word_add stackpointer (word 832),8 * 4)) s88`;
      `y5 = read (memory :> bytes(word_add stackpointer (word 864),8 * 4)) s88`;
      `z5 = read (memory :> bytes(word_add stackpointer (word 896),8 * 4)) s88`
     ] THEN

    (*** Computation of 6 * P ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (89--91) THEN LOCAL_JDOUBLE_TAC 92 THEN
    MAP_EVERY ABBREV_TAC
     [`x6 = read (memory :> bytes(word_add stackpointer (word 928),8 * 4)) s92`;
      `y6 = read (memory :> bytes(word_add stackpointer (word 960),8 * 4)) s92`;
      `z6 = read (memory :> bytes(word_add stackpointer (word 992),8 * 4)) s92`
     ] THEN

    (*** Computation of 7 * P ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (93--96) THEN LOCAL_JMIXADD_TAC 97 THEN
    MAP_EVERY ABBREV_TAC
     [`x7 = read (memory :> bytes(word_add stackpointer (word 1024),8 * 4)) s97`;
      `y7 = read (memory :> bytes(word_add stackpointer (word 1056),8 * 4)) s97`;
      `z7 = read (memory :> bytes(word_add stackpointer (word 1088),8 * 4)) s97`
     ] THEN

    (*** Computation of 8 * P ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (98--100) THEN LOCAL_JDOUBLE_TAC 101 THEN
    MAP_EVERY ABBREV_TAC
     [`x8 = read (memory :> bytes(word_add stackpointer (word 1120),8 * 4)) s101`;
      `y8 = read (memory :> bytes(word_add stackpointer (word 1152),8 * 4)) s101`;
      `z8 = read (memory :> bytes(word_add stackpointer (word 1184),8 * 4)) s101`
     ] THEN

    (*** Selection of the top bitfield ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (102--103) THEN
    SUBGOAL_THEN `word_ushr (sum_s38:int64) 60 = word(n DIV 2 EXP 252)`
    SUBST_ALL_TAC THENL
     [EXPAND_TAC "n" THEN CONV_TAC(ONCE_DEPTH_CONV BIGNUM_LEXPAND_CONV) THEN
      ASM_REWRITE_TAC[] THEN
      CONV_TAC(ONCE_DEPTH_CONV BIGNUM_OF_WORDLIST_DIV_CONV) THEN
      CONV_TAC WORD_BLAST;
      ALL_TAC] THEN

    (*** Constant-time table selection ***)

    BIGNUM_LDIGITIZE_TAC "fab_"
     `read(memory :> bytes(word_add stackpointer (word 448),8 * 96)) s103` THEN
    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (104--283) THEN
    ENSURES_FINAL_STATE_TAC THEN ASM_REWRITE_TAC[] THEN

    REWRITE_TAC[bignum_triple_from_memory] THEN
    CONV_TAC(ONCE_DEPTH_CONV EXPAND_CASES_CONV) THEN
    CONV_TAC(DEPTH_CONV(NUM_ADD_CONV ORELSEC NUM_MULT_CONV)) THEN
    CONV_TAC(ONCE_DEPTH_CONV NORMALIZE_RELATIVE_ADDRESS_CONV) THEN
    ASM_REWRITE_TAC[BIGNUM_FROM_MEMORY_BYTES] THEN

    (*** Final fiddle of the representations ***)

    X_GEN_TAC `P:(int#int)option` THEN STRIP_TAC THEN
    CONV_TAC(ONCE_DEPTH_CONV BIGNUM_LEXPAND_CONV) THEN
    ASM_REWRITE_TAC[] THEN REWRITE_TAC[SYM(NUM_EXP_CONV `2 EXP 252`)] THEN
    SUBGOAL_THEN `n DIV 2 EXP 252 < 9` MP_TAC THENL
     [UNDISCH_TAC `n < 9 * 2 EXP 252` THEN ARITH_TAC;
      SPEC_TAC(`n DIV 2 EXP 252`,`b:num`)] THEN
    REWRITE_TAC[MESON[ARITH_RULE `0 < 9`]
      `(!b. b < 9 ==> P b /\ Q) <=> (Q ==> !b. b < 9 ==> P b) /\ Q`] THEN
    CONJ_TAC THENL
     [EXPAND_TAC "recoder" THEN CONV_TAC NUM_REDUCE_CONV THEN
      REWRITE_TAC[INT_SUB_RZERO; GROUP_NPOW] THEN
      CONV_TAC(RAND_CONV EXPAND_CASES_CONV) THEN
      CONV_TAC(DEPTH_CONV WORD_NUM_RED_CONV) THEN
      ASM_REWRITE_TAC[WORD_SUB_0; VAL_WORD_BITVAL; BITVAL_EQ_0] THEN
      MATCH_MP_TAC(TAUT `q /\ (p ==> r) ==> p ==> q /\ r`) THEN
      CONJ_TAC THENL
       [REWRITE_TAC[group_pow; P256_GROUP; represents_p256; tripled;
                    weierstrass_of_jacobian; montgomery_decode; p_256;
                    bignum_of_wordlist; INTEGER_MOD_RING_CLAUSES] THEN
        CONV_TAC(DEPTH_CONV(WORD_NUM_RED_CONV ORELSEC INVERSE_MOD_CONV));
        REPEAT(MATCH_MP_TAC MONO_AND THEN CONJ_TAC) THEN
        MATCH_MP_TAC EQ_IMP THEN AP_TERM_TAC THEN REWRITE_TAC[PAIR_EQ] THEN
        W(MAP_EVERY EXPAND_TAC o map (name_of o lhs) o conjuncts o snd) THEN
        CONV_TAC(ONCE_DEPTH_CONV BIGNUM_LEXPAND_CONV) THEN
        ASM_REWRITE_TAC[WORD_SUB_0; VAL_WORD_BITVAL; BITVAL_EQ_0]];
      ALL_TAC] THEN

    FIRST_ASSUM(ASSUME_TAC o MATCH_MP REPRESENTS2_P256_NONZERO) THEN
    FIRST_ASSUM(MP_TAC o GEN_REWRITE_RULE I [GSYM REPRESENTS_P256_2]) THEN
    ASM_REWRITE_TAC[] THEN DISCH_TAC THEN

    REPEAT(FIRST_X_ASSUM(MP_TAC o check (is_forall o concl))) THEN
    DISCH_THEN(MP_TAC o SPEC `P:(int#int)option`) THEN ASM_REWRITE_TAC[] THEN
    REWRITE_TAC[IMP_IMP; GSYM CONJ_ASSOC] THEN
    ASM_SIMP_TAC[GROUP_RULE `group_mul G x x = group_pow G x 2`] THEN
    GEN_REWRITE_TAC I [IMP_CONJ] THEN DISCH_TAC THEN
    GEN_REWRITE_TAC I [IMP_CONJ] THEN DISCH_THEN(MP_TAC o SPECL
     [`group_pow p256_group P 2`; `P:(int#int)option`]) THEN
    ASM_SIMP_TAC[GROUP_RULE `group_pow G x 2 = x <=> x = group_id G`] THEN
    ASM_SIMP_TAC[GROUP_RULE
     `group_mul G (group_pow G x 2) x = group_pow G x 3`] THEN
    DISCH_TAC THEN
    GEN_REWRITE_TAC I [IMP_CONJ] THEN DISCH_THEN(MP_TAC o SPEC
     `group_pow p256_group P 2`) THEN
    ASM_SIMP_TAC[GSYM GROUP_POW_ADD] THEN CONV_TAC NUM_REDUCE_CONV THEN
    DISCH_TAC THEN GEN_REWRITE_TAC I [IMP_CONJ] THEN DISCH_THEN(MP_TAC o SPECL
     [`group_pow p256_group P 4`; `P:(int#int)option`]) THEN
    ASM_SIMP_TAC[GROUP_RULE
     `group_mul G (group_pow G x 4) x = group_pow G x 5`] THEN
    ANTS_TAC THENL
     [ASM_SIMP_TAC[GROUP_POW_EQ_ID; P256_GROUP_ELEMENT_ORDER; GROUP_RULE
        `group_pow G x 4 = x <=> group_pow G x 3 = group_id G`] THEN
      REWRITE_TAC[n_256] THEN CONV_TAC(RAND_CONV DIVIDES_CONV) THEN
      REWRITE_TAC[];
      DISCH_TAC] THEN
    GEN_REWRITE_TAC I [IMP_CONJ] THEN DISCH_THEN(MP_TAC o SPEC
     `group_pow p256_group P 3`) THEN
    ASM_SIMP_TAC[GSYM GROUP_POW_ADD] THEN CONV_TAC NUM_REDUCE_CONV THEN
    DISCH_TAC THEN GEN_REWRITE_TAC I [IMP_CONJ] THEN DISCH_THEN(MP_TAC o SPECL
     [`group_pow p256_group P 6`; `P:(int#int)option`]) THEN
    ASM_SIMP_TAC[GROUP_RULE
     `group_mul G (group_pow G x 6) x = group_pow G x 7`] THEN
    ANTS_TAC THENL
     [ASM_SIMP_TAC[GROUP_POW_EQ_ID; P256_GROUP_ELEMENT_ORDER; GROUP_RULE
        `group_pow G x 6 = x <=> group_pow G x 5 = group_id G`] THEN
      REWRITE_TAC[n_256] THEN CONV_TAC(RAND_CONV DIVIDES_CONV) THEN
      REWRITE_TAC[];
      DISCH_TAC] THEN
    DISCH_THEN(MP_TAC o SPEC `group_pow p256_group P 4`) THEN
    ASM_SIMP_TAC[GSYM GROUP_POW_ADD] THEN CONV_TAC NUM_REDUCE_CONV THEN
    DISCH_TAC THEN ASM_SIMP_TAC[GROUP_POW_1];

    (**** Preservation of the loop invariant ***)

    X_GEN_TAC `i:num` THEN STRIP_TAC THEN
    CONV_TAC(RATOR_CONV(LAND_CONV(ONCE_DEPTH_CONV EXPAND_CASES_CONV))) THEN
    CONV_TAC NUM_REDUCE_CONV THEN
    GEN_REWRITE_TAC (RATOR_CONV o LAND_CONV o ONCE_DEPTH_CONV)
     [bignum_triple_from_memory] THEN
    CONV_TAC NUM_REDUCE_CONV THEN
    CONV_TAC(ONCE_DEPTH_CONV NORMALIZE_RELATIVE_ADDRESS_CONV) THEN

    GHOST_INTRO_TAC `Xa:num`
     `bignum_from_memory (word_add stackpointer (word 256),4)` THEN
    GHOST_INTRO_TAC `Ya:num`
     `bignum_from_memory (word_add stackpointer (word 288),4)` THEN
    GHOST_INTRO_TAC `Za:num`
     `bignum_from_memory (word_add stackpointer (word 320),4)` THEN

    (*** Computation of 2 * (Xa,Ya,Za) ***)

    REWRITE_TAC[BIGNUM_FROM_MEMORY_BYTES] THEN ENSURES_INIT_TAC "s0" THEN
    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (1--4) THEN
    RULE_ASSUM_TAC(REWRITE_RULE[WORD_RULE
     `word_sub (word (4 * (i + 1))) (word 4) = word(4 * i)`]) THEN
    LOCAL_JDOUBLE_TAC 5 THEN
    MAP_EVERY ABBREV_TAC
     [`X2a = read (memory :> bytes(word_add stackpointer (word 256),8 * 4)) s5`;
      `Y2a = read (memory :> bytes(word_add stackpointer (word 288),8 * 4)) s5`;
      `Z2a = read (memory :> bytes(word_add stackpointer (word 320),8 * 4)) s5`
     ] THEN

    (*** Computation of 4 * (Xa,Ya,Za) ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (6--8) THEN LOCAL_JDOUBLE_TAC 9 THEN
    MAP_EVERY ABBREV_TAC
     [`X4a = read (memory :> bytes(word_add stackpointer (word 256),8 * 4)) s9`;
      `Y4a = read (memory :> bytes(word_add stackpointer (word 288),8 * 4)) s9`;
      `Z4a = read (memory :> bytes(word_add stackpointer (word 320),8 * 4)) s9`
     ] THEN

    (*** Computation of 8 * (Xa,Ya,Za) ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (10--12) THEN LOCAL_JDOUBLE_TAC 13 THEN
    MAP_EVERY ABBREV_TAC
     [`X8a = read (memory :> bytes(word_add stackpointer (word 256),8 * 4)) s13`;
      `Y8a = read (memory :> bytes(word_add stackpointer (word 288),8 * 4)) s13`;
      `Z8a = read (memory :> bytes(word_add stackpointer (word 320),8 * 4)) s13`
     ] THEN

    (*** Computation of 16 * (Xa,Ya,Za) ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (14--16) THEN LOCAL_JDOUBLE_TAC 17 THEN
    MAP_EVERY ABBREV_TAC
     [`Xha = read (memory :> bytes(word_add stackpointer (word 256),8 * 4)) s17`;
      `Yha = read (memory :> bytes(word_add stackpointer (word 288),8 * 4)) s17`;
      `Zha = read (memory :> bytes(word_add stackpointer (word 320),8 * 4)) s17`
     ] THEN

    (*** Selection of btable nybble ***)

    SUBGOAL_THEN
     `read(memory :> bytes64 (word_add stackpointer
           (word(224 + 8 * val(word_ushr (word (4 * i)) 6:int64))))) s17 =
      word(n DIV 2 EXP (64 * (4 * i) DIV 64) MOD 2 EXP (64 * 1))`
    ASSUME_TAC THENL
     [EXPAND_TAC "n" THEN
      REWRITE_TAC[GSYM BIGNUM_FROM_MEMORY_BYTES] THEN
      REWRITE_TAC[BIGNUM_FROM_MEMORY_DIV; BIGNUM_FROM_MEMORY_MOD] THEN
      ASM_SIMP_TAC[ARITH_RULE
       `i < 63 ==> MIN (4 - (4 * i) DIV 64) 1 = 1`] THEN
      REWRITE_TAC[BIGNUM_FROM_MEMORY_SING; WORD_VAL] THEN
      REWRITE_TAC[GSYM WORD_ADD_ASSOC; GSYM WORD_ADD] THEN
      AP_THM_TAC THEN REPLICATE_TAC 7 AP_TERM_TAC THEN
      REWRITE_TAC[VAL_WORD_USHR] THEN CONV_TAC NUM_REDUCE_CONV THEN
      AP_THM_TAC THEN AP_TERM_TAC THEN MATCH_MP_TAC VAL_WORD_EQ THEN
      REWRITE_TAC[DIMINDEX_64] THEN CONV_TAC NUM_REDUCE_CONV THEN
      ASM BOUNDER_TAC[];
      ALL_TAC] THEN

    (*** Recoding offset to get indexing and negation flag ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (18--24) THEN
    ABBREV_TAC `bf = (n DIV (2 EXP (4 * i))) MOD 16` THEN
    SUBGOAL_THEN
     `word_and
       (word_jushr
        (word ((n DIV 2 EXP (64 * (4 * i) DIV 64)) MOD 2 EXP 64))
       (word (4 * i)))
      (word 15):int64 = word bf` SUBST_ALL_TAC THENL
     [REWRITE_TAC[GSYM VAL_EQ; VAL_WORD_AND_MASK_WORD;
                  ARITH_RULE `15 = 2 EXP 4 - 1`] THEN
      REWRITE_TAC[word_jushr; VAL_WORD_USHR; DIMINDEX_64; MOD_64_CLAUSES] THEN
      EXPAND_TAC "bf" THEN REWRITE_TAC[VAL_WORD; DIMINDEX_64] THEN
      REWRITE_TAC[MOD_MOD_EXP_MIN; ARITH_RULE `16 = 2 EXP 4`] THEN
      CONV_TAC(ONCE_DEPTH_CONV NUM_MIN_CONV) THEN
      REWRITE_TAC[DIV_MOD; MOD_MOD_EXP_MIN; GSYM EXP_ADD; DIV_DIV] THEN
      REWRITE_TAC[ADD_ASSOC; ARITH_RULE `64 * i DIV 64 + i MOD 64 = i`] THEN
      AP_THM_TAC THEN REPLICATE_TAC 3 AP_TERM_TAC THEN
      REWRITE_TAC[ARITH_RULE `MIN a b = b <=> b <= a`] THEN
      REWRITE_TAC[ARITH_RULE
       `x <= 64 * y DIV 64 + z <=> x + y MOD 64 <= y + z`] THEN
      REWRITE_TAC[ARITH_RULE `64 = 4 * 16`; MOD_MULT2] THEN
      UNDISCH_TAC `i < 63` THEN ARITH_TAC;
      ALL_TAC] THEN
    RULE_ASSUM_TAC(REWRITE_RULE[WORD_UNMASK_64; NOT_LE]) THEN
    SUBGOAL_THEN `val(word bf:int64) = bf` SUBST_ALL_TAC THENL
     [MATCH_MP_TAC VAL_WORD_EQ THEN REWRITE_TAC[DIMINDEX_64] THEN
      EXPAND_TAC "bf" THEN ARITH_TAC;
      ALL_TAC] THEN
    ABBREV_TAC `ix = if bf < 8 then 8 - bf else bf - 8` THEN
    FIRST_X_ASSUM(MP_TAC o SPEC `word ix:int64` o MATCH_MP (MESON[]
     `read X14 s = x ==> !x'. x = x' ==> read X14 s = x'`)) THEN
    ANTS_TAC THENL
     [EXPAND_TAC "ix" THEN REWRITE_TAC[GSYM NOT_LT; COND_SWAP] THEN
      REWRITE_TAC[WORD_NEG_SUB] THEN POP_ASSUM_LIST(K ALL_TAC) THEN
      COND_CASES_TAC THEN ASM_REWRITE_TAC[WORD_SUB] THEN ASM_ARITH_TAC;
      DISCH_TAC] THEN

    (*** Constant-time selection from the table ***)

    BIGNUM_LDIGITIZE_TAC "tab_"
     `read(memory :> bytes(word_add stackpointer (word 448),8 * 96)) s24` THEN
    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (25--197) THEN
    MAP_EVERY REABBREV_TAC
     [`tab0 = read X0 s197`;
      `tab1 = read X1 s197`;
      `tab2 = read X2 s197`;
      `tab3 = read X3 s197`;
      `tab4 = read X4 s197`;
      `tab5 = read X5 s197`;
      `tab6 = read X6 s197`;
      `tab7 = read X7 s197`;
      `tab8 = read X8 s197`;
      `tab9 = read X9 s197`;
      `tab10 = read X10 s197`;
      `tab11 = read X11 s197`] THEN

    SUBGOAL_THEN
     `!P. P IN group_carrier p256_group /\ represents2_p256 P (x',y')
          ==> represents_p256 (group_pow p256_group P ix)
                 (bignum_of_wordlist[tab0; tab1; tab2; tab3],
                  bignum_of_wordlist[tab4; tab5; tab6; tab7],
                  bignum_of_wordlist[tab8; tab9; tab10; tab11])`
    ASSUME_TAC THENL
     [X_GEN_TAC `P:(int#int)option` THEN STRIP_TAC THEN
      FIRST_X_ASSUM(MP_TAC o SPEC `P:(int#int)option`) THEN
      CONV_TAC(ONCE_DEPTH_CONV BIGNUM_LEXPAND_CONV) THEN
      ASM_REWRITE_TAC[] THEN DISCH_TAC THEN
      MAP_EVERY EXPAND_TAC
       (map (fun n -> "tab"^string_of_int n) (0--11)) THEN
      SUBGOAL_THEN `ix < 9` MP_TAC THENL
       [MAP_EVERY EXPAND_TAC ["ix"; "bf"] THEN ARITH_TAC;
        SPEC_TAC(`ix:num`,`ix:num`)] THEN
      CONV_TAC EXPAND_CASES_CONV THEN
      CONV_TAC(DEPTH_CONV WORD_NUM_RED_CONV) THEN
      ASM_REWRITE_TAC[CONJUNCT1 group_pow] THEN
      REWRITE_TAC[group_pow; P256_GROUP; represents_p256; tripled;
                  weierstrass_of_jacobian; montgomery_decode; p_256;
                  bignum_of_wordlist; INTEGER_MOD_RING_CLAUSES] THEN
      CONV_TAC(DEPTH_CONV(WORD_NUM_RED_CONV ORELSEC INVERSE_MOD_CONV));
      ALL_TAC] THEN

    (*** Optional negation of the table entry -- we know that y is nonzero ***)

    ARM_ACCSTEPS_TAC P256_SCALARMUL_ALT_EXEC [201;203;205;206] (198--215) THEN
    MAP_EVERY ABBREV_TAC
     [`Xt = read (memory :> bytes(word_add stackpointer (word 352),8 * 4)) s215`;
      `Yt = read (memory :> bytes(word_add stackpointer (word 384),8 * 4)) s215`;
      `Zt = read (memory :> bytes(word_add stackpointer (word 416),8 * 4)) s215`
     ] THEN
    SUBGOAL_THEN
     `!P. P IN group_carrier p256_group /\ represents2_p256 P (x',y')
          ==> represents_p256 (group_zpow p256_group P (&bf - &8)) (Xt,Yt,Zt)`
    ASSUME_TAC THENL
     [X_GEN_TAC `P:(int#int)option` THEN STRIP_TAC THEN
      FIRST_X_ASSUM(K ALL_TAC o SPEC `P:(int#int)option`) THEN
      FIRST_X_ASSUM(MP_TAC o SPEC `P:(int#int)option`) THEN
      ASM_REWRITE_TAC[] THEN DISCH_TAC THEN
      MAP_EVERY EXPAND_TAC ["Xt"; "Yt"; "Zt"] THEN
      CONV_TAC(ONCE_DEPTH_CONV BIGNUM_LEXPAND_CONV) THEN ASM_REWRITE_TAC[] THEN
      SUBGOAL_THEN `&bf - &8:int = if bf < 8 then --(&ix) else &ix`
      SUBST1_TAC THENL
       [EXPAND_TAC "ix" THEN
        SUBGOAL_THEN `bf < 16` MP_TAC THENL
         [EXPAND_TAC "bf" THEN ARITH_TAC; POP_ASSUM_LIST(K ALL_TAC)] THEN
        COND_CASES_TAC THEN ASM_SIMP_TAC[GSYM INT_OF_NUM_SUB; GSYM NOT_LT] THEN
        ASM_SIMP_TAC[GSYM INT_OF_NUM_SUB; LT_IMP_LE] THEN INT_ARITH_TAC;
        ALL_TAC] THEN
      REWRITE_TAC[GSYM NOT_LT; COND_SWAP] THEN COND_CASES_TAC THEN
      CONV_TAC(DEPTH_CONV WORD_NUM_RED_CONV) THEN
      ASM_REWRITE_TAC[GROUP_ZPOW_POW] THEN
      FIRST_ASSUM(ASSUME_TAC o MATCH_MP REPRESENTS2_P256_NONZERO) THEN
      FIRST_ASSUM(MP_TAC o MATCH_MP REPRESENTS_P256_NEGATION) THEN
      ANTS_TAC THENL
       [ASM_SIMP_TAC[GROUP_POW; GROUP_POW_EQ_ID; P256_GROUP_ELEMENT_ORDER] THEN
        DISCH_THEN(MP_TAC o MATCH_MP DIVIDES_LE) THEN UNDISCH_TAC `bf < 8` THEN
        REWRITE_TAC[n_256] THEN MAP_EVERY EXPAND_TAC ["ix"; "bf"] THEN
        ARITH_TAC;
        MATCH_MP_TAC EQ_IMP] THEN
      AP_TERM_TAC THEN REWRITE_TAC[PAIR_EQ] THEN
      MATCH_MP_TAC CONG_IMP_EQ THEN EXISTS_TAC `2 EXP 256` THEN
      CONJ_TAC THENL [REWRITE_TAC[p_256] THEN ARITH_TAC; ALL_TAC] THEN
      CONJ_TAC THENL [BOUNDER_TAC[]; ALL_TAC] THEN
      FIRST_ASSUM(MP_TAC o el 1 o CONJUNCTS o
       GEN_REWRITE_RULE I [represents_p256]) THEN
      SIMP_TAC[num_congruent; GSYM INT_OF_NUM_SUB; LT_IMP_LE] THEN
      DISCH_THEN(K ALL_TAC) THEN REWRITE_TAC[REAL_INT_CONGRUENCE] THEN
      REWRITE_TAC[GSYM REAL_OF_INT_CLAUSES] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; REAL_POW_EQ_0] THEN
      REWRITE_TAC[REAL_OF_NUM_EQ; ARITH_EQ] THEN
      REWRITE_TAC[bignum_of_wordlist; p_256; GSYM REAL_OF_NUM_CLAUSES] THEN
      ACCUMULATOR_POP_ASSUM_LIST(MP_TAC o end_itlist CONJ o DESUM_RULE) THEN
      DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN REAL_INTEGER_TAC;
      ACCUMULATOR_POP_ASSUM_LIST(K ALL_TAC)] THEN

    (*** Final addition of the table entry ***)

    ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (216--219) THEN LOCAL_JADD_TAC 220 THEN
    MAP_EVERY ABBREV_TAC
     [`X' = read (memory :> bytes(word_add stackpointer (word 256),8 * 4)) s220`;
      `Y' = read (memory :> bytes(word_add stackpointer (word 288),8 * 4)) s220`;
      `Z' = read (memory :> bytes(word_add stackpointer (word 320),8 * 4)) s220`
     ] THEN

    (*** The final mathematics ***)

    ENSURES_FINAL_STATE_TAC THEN ASM_REWRITE_TAC[] THEN

    X_GEN_TAC `P:(int#int)option` THEN STRIP_TAC THEN
    CONV_TAC(RAND_CONV EXPAND_CASES_CONV) THEN
    REWRITE_TAC[bignum_triple_from_memory] THEN
    CONV_TAC NUM_REDUCE_CONV THEN
    CONV_TAC(ONCE_DEPTH_CONV NORMALIZE_RELATIVE_ADDRESS_CONV) THEN
    ASM_REWRITE_TAC[BIGNUM_FROM_MEMORY_BYTES] THEN
    FIRST_X_ASSUM(MP_TAC o SPEC `P:(int#int)option`) THEN
    ASM_REWRITE_TAC[] THEN
    DISCH_THEN(CONJUNCTS_THEN2 ASSUME_TAC (fun th -> REWRITE_TAC[th])) THEN

    ABBREV_TAC
     `Q = group_zpow p256_group P
        (&(n DIV 2 EXP (4 * (i + 1))) -
         &(recoder DIV 2 EXP (4 * (i + 1))))` THEN
    SUBGOAL_THEN `Q IN group_carrier p256_group` ASSUME_TAC THENL
     [EXPAND_TAC "Q" THEN MATCH_MP_TAC GROUP_ZPOW THEN ASM_REWRITE_TAC[];
      ALL_TAC] THEN
    UNDISCH_THEN
     `!P. represents_p256 P (Xa,Ya,Za)
          ==> represents_p256 (group_mul p256_group P P) (X2a,Y2a,Z2a)`
     (MP_TAC o SPEC `Q:(int#int)option`) THEN
    ASM_SIMP_TAC[GROUP_RULE `group_mul G x x = group_pow G x 2`] THEN
    DISCH_TAC THEN UNDISCH_THEN
     `!P. represents_p256 P (X2a,Y2a,Z2a)
          ==> represents_p256 (group_mul p256_group P P) (X4a,Y4a,Z4a)`
     (MP_TAC o SPEC `group_pow p256_group Q 2`) THEN
    ASM_SIMP_TAC[GSYM GROUP_POW_ADD] THEN CONV_TAC NUM_REDUCE_CONV THEN
    DISCH_TAC THEN UNDISCH_THEN
     `!P. represents_p256 P (X4a,Y4a,Z4a)
          ==> represents_p256 (group_mul p256_group P P) (X8a,Y8a,Z8a)`
     (MP_TAC o SPEC `group_pow p256_group Q 4`) THEN
    ASM_SIMP_TAC[GSYM GROUP_POW_ADD] THEN CONV_TAC NUM_REDUCE_CONV THEN
    DISCH_TAC THEN UNDISCH_THEN
     `!P. represents_p256 P (X8a,Y8a,Z8a)
          ==> represents_p256 (group_mul p256_group P P) (Xha,Yha,Zha)`
     (MP_TAC o SPEC `group_pow p256_group Q 8`) THEN
    ASM_SIMP_TAC[GSYM GROUP_POW_ADD] THEN CONV_TAC NUM_REDUCE_CONV THEN
    DISCH_TAC THEN FIRST_X_ASSUM(MP_TAC o SPECL
     [`group_pow p256_group Q 16`;
      `group_zpow p256_group P (&bf - &8)`]) THEN
    FIRST_X_ASSUM(MP_TAC o SPEC `P:(int#int)option`) THEN
    ASM_REWRITE_TAC[] THEN DISCH_TAC THEN ASM_REWRITE_TAC[] THEN
    ASM_SIMP_TAC[GSYM GROUP_NPOW] THEN EXPAND_TAC "Q" THEN
    ASM_SIMP_TAC[GSYM GROUP_ZPOW_MUL; GSYM GROUP_ZPOW_ADD] THEN
    ANTS_TAC THENL
     [SUBST1_TAC(SYM(el 1 (CONJUNCTS P256_GROUP))) THEN
      ASM_SIMP_TAC[GROUP_ZPOW_EQ; GROUP_ZPOW_EQ_ID;
                   P256_GROUP_ELEMENT_ORDER] THEN
      COND_CASES_TAC THEN ASM_REWRITE_TAC[INT_DIVIDES_1] THEN
      DISCH_THEN(MP_TAC o MATCH_MP (REWRITE_RULE[IMP_CONJ_ALT]
          INT_CONG_IMP_EQ)) THEN
      ANTS_TAC THENL
       [MATCH_MP_TAC(INT_ARITH
         `abs(&16 * x) + abs(&16 * y) + abs(bf) + &8:int < n
          ==> abs((x - y) * &16 - (bf - &8)) < n`) THEN
        REWRITE_TAC[INT_ABS_NUM; INT_OF_NUM_CLAUSES] THEN
        REWRITE_TAC[ARITH_RULE `4 * (i + 1) = 4 * i + 4`; EXP_ADD] THEN
        REWRITE_TAC[GSYM DIV_DIV] THEN MATCH_MP_TAC(ARITH_RULE
         `a + b + c + d < n
          ==> 16 * a DIV 2 EXP 4 + 16 * b DIV 2 EXP 4 + c + d < n`) THEN
        TRANS_TAC LET_TRANS `n + recoder + bf + 8` THEN
        SIMP_TAC[LE_ADD2; LE_REFL; DIV_LE] THEN
        UNDISCH_TAC `n < 9 * 2 EXP 252` THEN
        MAP_EVERY EXPAND_TAC ["recoder"; "bf"] THEN
        REWRITE_TAC[n_256] THEN ARITH_TAC;
        ALL_TAC] THEN
      ASM_CASES_TAC `&bf - &8:int = &0` THEN ASM_REWRITE_TAC[INT_DIVIDES_0] THEN
      UNDISCH_TAC `~(&bf - &8:int = &0)` THEN
      MATCH_MP_TAC(TAUT `(p ==> ~q) ==> p ==> q ==> r`) THEN
      MATCH_MP_TAC(INT_ARITH
       `abs(y:int) < &16 /\ (~(x = &0) ==> &1 <= abs(x))
        ==> ~(y = &0) ==> ~(x * &16 = y)`) THEN
      CONJ_TAC THENL [EXPAND_TAC "bf"; CONV_TAC INT_ARITH] THEN
      REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES; GSYM INT_OF_NUM_REM] THEN
      CONV_TAC INT_ARITH;
      MATCH_MP_TAC EQ_IMP] THEN

    AP_THM_TAC THEN AP_TERM_TAC THEN AP_TERM_TAC THEN

    SUBGOAL_THEN
     `!n. n DIV 2 EXP (4 * i) =
          16 * (n DIV 2 EXP (4 * (i + 1))) + (n DIV 2 EXP (4 * i)) MOD 16`
    MP_TAC THENL
     [REWRITE_TAC[ARITH_RULE `4 * (i + 1) = 4 * i + 4`; EXP_ADD] THEN
      REWRITE_TAC[GSYM DIV_DIV] THEN ARITH_TAC;
      DISCH_THEN(fun th -> ONCE_REWRITE_TAC[th]) THEN
      ASM_REWRITE_TAC[]] THEN
    SUBGOAL_THEN `(recoder DIV 2 EXP (4 * i)) MOD 16 = 8` SUBST1_TAC THENL
     [UNDISCH_TAC `i < 63` THEN SPEC_TAC(`i:num`,`i:num`) THEN
      EXPAND_TAC "recoder" THEN POP_ASSUM_LIST(K ALL_TAC) THEN
      CONV_TAC EXPAND_CASES_CONV THEN CONV_TAC NUM_REDUCE_CONV;
      ALL_TAC] THEN
    REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES] THEN CONV_TAC INT_ARITH;

    (*** Trivial loop-back goal ***)

    REPEAT STRIP_TAC THEN CONV_TAC(ONCE_DEPTH_CONV EXPAND_CASES_CONV) THEN
    CONV_TAC NUM_REDUCE_CONV THEN
    ARM_SIM_TAC P256_SCALARMUL_ALT_EXEC [1] THEN
    VAL_INT64_TAC `4 * i` THEN
    ASM_REWRITE_TAC[ARITH_RULE `4 * i = 0 <=> ~(0 < i)`];

    ALL_TAC] THEN

  (*** Final conversion and mathematics ***)

  GEN_REWRITE_TAC (RATOR_CONV o LAND_CONV o ONCE_DEPTH_CONV)
   [bignum_triple_from_memory] THEN
  CONV_TAC NUM_REDUCE_CONV THEN REWRITE_TAC[DIV_1] THEN
  CONV_TAC(ONCE_DEPTH_CONV NORMALIZE_RELATIVE_ADDRESS_CONV) THEN

  GHOST_INTRO_TAC `X:num`
   `bignum_from_memory (word_add stackpointer (word 256),4)` THEN
  GHOST_INTRO_TAC `Y:num`
   `bignum_from_memory (word_add stackpointer (word 288),4)` THEN
  GHOST_INTRO_TAC `Z:num`
   `bignum_from_memory (word_add stackpointer (word 320),4)` THEN

  REWRITE_TAC[BIGNUM_FROM_MEMORY_BYTES] THEN ENSURES_INIT_TAC "s0" THEN
  FIRST_X_ASSUM(ASSUME_TAC o MATCH_MP (MESON[]
    `(!x. P x ==> Q x /\ R x) ==> (!x. P x ==> Q x)`)) THEN

  (*** Remaining arithmetic operations to get final answer ***)

  ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (1--4) THEN LOCAL_SQR_TAC 5 THEN
  ABBREV_TAC
   `Z2 = read(memory :> bytes(word_add stackpointer (word 448),8 * 4)) s5` THEN
  ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (6--9) THEN LOCAL_MUL_TAC 10 THEN
  ABBREV_TAC
   `Z3 =
    read(memory :> bytes(word_add stackpointer (word 480),8 * 4)) s10` THEN
  ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (11--13) THEN LOCAL_DEMONT_TAC 14 THEN
  ABBREV_TAC
   `Z3' =
    read(memory :> bytes(word_add stackpointer (word 448),8 * 4)) s14` THEN
  ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (15--17) THEN LOCAL_INV_TAC 18 THEN
  ABBREV_TAC
   `I3 =
    read(memory :> bytes(word_add stackpointer (word 480),8 * 4)) s18` THEN
  ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (19--22) THEN LOCAL_MUL_TAC 23 THEN
  ABBREV_TAC
   `I2 =
    read (memory :> bytes(word_add stackpointer (word 448),8 * 4)) s23` THEN
  ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (24--27) THEN LOCAL_MUL_TAC 28 THEN
  ABBREV_TAC `X' = read (memory :> bytes(res,8 * 4)) s28` THEN
  ARM_STEPS_TAC P256_SCALARMUL_ALT_EXEC (29--32) THEN LOCAL_MUL_TAC 33 THEN
  ABBREV_TAC
   `Y' = read (memory :> bytes(word_add res (word 32),8 * 4)) s33` THEN

  (*** Final mathematics sorting out point negation and last operations ***)

  ENSURES_FINAL_STATE_TAC THEN ASM_REWRITE_TAC[] THEN
  DISCARD_STATE_TAC "s33" THEN X_GEN_TAC `P:(int#int)option` THEN
  STRIP_TAC THEN
  ABBREV_TAC `Q = if sgn then group_inv p256_group P else P` THEN
  SUBGOAL_THEN `Q IN group_carrier p256_group` ASSUME_TAC THENL
   [EXPAND_TAC "Q" THEN COND_CASES_TAC THEN ASM_SIMP_TAC[GROUP_INV];
    ALL_TAC] THEN

  SUBGOAL_THEN `represents2_p256 P (x',y'')` ASSUME_TAC THENL
   [MAP_EVERY EXPAND_TAC ["x'"; "y''"] THEN REWRITE_TAC[represents2_p256] THEN
    REPEAT(CONJ_TAC THENL [REWRITE_TAC[p_256] THEN ARITH_TAC; ALL_TAC]) THEN
    FIRST_X_ASSUM(SUBST1_TAC o GEN_REWRITE_RULE I [affinepoint_p256]) THEN
    AP_TERM_TAC THEN REWRITE_TAC[paired; PAIR_EQ] THEN
    REWRITE_TAC[montgomery_decode; modular_decode] THEN
    REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES; GSYM INT_OF_NUM_REM] THEN
    CONV_TAC INT_REM_DOWN_CONV THEN REWRITE_TAC[INT_REM_EQ] THEN
    MATCH_MP_TAC(INTEGER_RULE
     `(i * p:int == &1) (mod n)
      ==> (i * p * x == x) (mod n) /\ (i * p * y == y) (mod n)`) THEN
    REWRITE_TAC[INT_OF_NUM_CLAUSES; GSYM num_congruent] THEN
    REWRITE_TAC[INVERSE_MOD_LMUL_EQ; COPRIME_REXP; COPRIME_2] THEN
    REWRITE_TAC[p_256] THEN CONV_TAC NUM_REDUCE_CONV;
    ALL_TAC] THEN

  FIRST_ASSUM(ASSUME_TAC o MATCH_MP REPRESENTS2_P256_NONZERO) THEN

  SUBGOAL_THEN `represents2_p256 Q (x',y')` ASSUME_TAC THENL
   [MAP_EVERY EXPAND_TAC ["y'"; "Q"] THEN
    COND_CASES_TAC THEN ASM_REWRITE_TAC[] THEN
    REWRITE_TAC[GSYM REPRESENTS_P256_2] THEN
    FIRST_ASSUM(MP_TAC o GEN_REWRITE_RULE I [GSYM REPRESENTS_P256_2]) THEN
    DISCH_THEN(MATCH_MP_TAC o MATCH_MP REPRESENTS_P256_NEGATION) THEN
    ASM_REWRITE_TAC[];
    ALL_TAC] THEN

  FIRST_X_ASSUM(MP_TAC o SPEC `Q:(int#int)option`) THEN
  ASM_REWRITE_TAC[] THEN
  SUBGOAL_THEN
   `group_zpow p256_group Q (&n - &recoder) = group_pow p256_group P n_input`
  SUBST1_TAC THENL
   [MAP_EVERY EXPAND_TAC ["Q"; "n"; "n_neg"; "n_red"] THEN
    REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[INT_ARITH `(x + y) - y:int = x`] THEN
    COND_CASES_TAC THEN REWRITE_TAC[GSYM INT_OF_NUM_REM; GSYM GROUP_NPOW] THEN
    ASM_SIMP_TAC[GSYM GROUP_INV_ZPOW; GSYM GROUP_ZPOW_NEG; GROUP_ZPOW_EQ] THEN
    ASM_SIMP_TAC[P256_GROUP_ELEMENT_ORDER; INT_CONG_LREM; INT_CONG_REFL] THEN
    EXPAND_TAC "n_red" THEN
    SUBGOAL_THEN `n_input MOD n_256 <= n_256`
     (fun th -> SIMP_TAC[GSYM INT_OF_NUM_SUB; th])
    THENL [REWRITE_TAC[n_256] THEN ARITH_TAC; ALL_TAC] THEN
    REWRITE_TAC[INTEGER_RULE
    `(--(n - x):int == y) (mod n) <=> (x == y) (mod n)`] THEN
    REWRITE_TAC[INT_CONG_LREM; INT_CONG_REFL; GSYM INT_OF_NUM_REM];
    ALL_TAC] THEN

  DISCH_TAC THEN
  FIRST_ASSUM(MP_TAC o MATCH_MP REPRESENTS_P256_Y_NONZERO) THEN
  ASM_SIMP_TAC[GROUP_POW] THEN
  FIRST_X_ASSUM(MP_TAC o GEN_REWRITE_RULE I [represents_p256]) THEN
  REWRITE_TAC[affinepointz_p256] THEN
  REPEAT(DISCH_THEN(CONJUNCTS_THEN2 ASSUME_TAC MP_TAC)) THEN
  DISCH_THEN(SUBST1_TAC o SYM) THEN

  REPEAT(FIRST_X_ASSUM(MP_TAC o check (is_imp o concl))) THEN
  REPEAT(ANTS_TAC THENL
   [TRY(COND_CASES_TAC THEN REWRITE_TAC[MULT_CLAUSES; LE_0] THEN
        W(MP_TAC o PART_MATCH (lhand o lhand) INVERSE_MOD_BOUND o
          rand o lhand o snd) THEN
        REWRITE_TAC[p_256] THEN CONV_TAC NUM_REDUCE_CONV THEN STRIP_TAC) THEN
     REWRITE_TAC[p_256] THEN RULE_ASSUM_TAC(REWRITE_RULE[p_256]) THEN
    CONV_TAC NUM_REDUCE_CONV THEN ASM BOUNDER_TAC[];
    (DISCH_THEN(STRIP_ASSUME_TAC o MATCH_MP unilemma0) ORELSE
     DISCH_THEN(STRIP_ASSUME_TAC o MATCH_MP unilemma1) ORELSE
     STRIP_TAC)]) THEN

  ASM_CASES_TAC `Z = 0` THENL
   [ASM_REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES; INT_MUL_LZERO; INT_MUL_RZERO;
                    INT_REM_ZERO; num_divides; INT_DIVIDES_0;
                    weierstrass_of_jacobian; tripled; INTEGER_MOD_RING_CLAUSES;
                    montgomery_decode];
    ASM_REWRITE_TAC[] THEN DISCH_TAC] THEN

  MP_TAC(SPECL [`p_256`; `2 EXP 256`] INVERSE_MOD_LMUL_EQ) THEN
  REWRITE_TAC[COPRIME_REXP; COPRIME_2] THEN
  DISCH_THEN(MP_TAC o snd o EQ_IMP_RULE) THEN ANTS_TAC THENL
   [REWRITE_TAC[p_256] THEN ARITH_TAC; DISCH_TAC] THEN

  SUBGOAL_THEN `~(p_256 divides Z3')` MP_TAC THENL
   [ASM_REWRITE_TAC[num_divides; GSYM INT_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[GSYM INT_REM_EQ_0] THEN CONV_TAC INT_REM_DOWN_CONV THEN
    MP_TAC(SPECL [`3`; `p_256`; `Z:num`] PRIME_DIVEXP_EQ) THEN
    ASM_SIMP_TAC[DIVIDES_EQ_ZERO; ARITH_EQ; PRIME_P256] THEN
    ASM_REWRITE_TAC[INT_REM_EQ_0] THEN
    FIRST_X_ASSUM(MP_TAC o GEN_REWRITE_RULE I [num_congruent]) THEN
    REWRITE_TAC[num_divides; GSYM INT_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[TAUT `p ==> ~q ==> ~r <=> p /\ r ==> q`] THEN
    CONV_TAC INTEGER_RULE;
    DISCH_THEN(fun th ->
     RULE_ASSUM_TAC(REWRITE_RULE[th]) THEN ASSUME_TAC th)] THEN

  SUBGOAL_THEN `~(p_256 divides Y')` MP_TAC THENL
   [ASM_REWRITE_TAC[num_divides; GSYM INT_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[GSYM INT_REM_EQ_0] THEN CONV_TAC INT_REM_DOWN_CONV THEN
    MP_TAC(SPECL [`p_256`; `Z3':num`] INVERSE_MOD_LMUL_EQ) THEN
    ASM_SIMP_TAC[PRIME_COPRIME_EQ; PRIME_P256] THEN
    MP_TAC(SPECL [`p_256`; `Y:num`] DIVIDES_EQ_ZERO) THEN
    ASM_REWRITE_TAC[INT_REM_EQ_0] THEN
    FIRST_X_ASSUM(MP_TAC o GEN_REWRITE_RULE I [num_congruent]) THEN
    REWRITE_TAC[num_divides; num_congruent; GSYM INT_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[TAUT `p ==> ~q ==> r ==> ~s <=> p /\ r /\ s ==> q`] THEN
    CONV_TAC INTEGER_RULE;
    ASM_SIMP_TAC[DIVIDES_EQ_ZERO] THEN DISCH_TAC] THEN

  ASM_REWRITE_TAC[weierstrass_of_jacobian; tripled; paired] THEN
  COND_CASES_TAC THENL
   [POP_ASSUM MP_TAC THEN REWRITE_TAC[option_DISTINCT] THEN
    REWRITE_TAC[montgomery_decode; INTEGER_MOD_RING_CLAUSES] THEN
    FIRST_X_ASSUM(MP_TAC o GEN_REWRITE_RULE I [num_congruent]) THEN
    SIMP_TAC[INT_REM_EQ_0; PRIME_INT_DIVPROD_EQ; PRIME_P256;
             GSYM INT_OF_NUM_CLAUSES] THEN
    ASM_SIMP_TAC[GSYM num_divides; DIVIDES_EQ_ZERO] THEN
    SIMP_TAC[GSYM PRIME_COPRIME_EQ; PRIME_P256; num_coprime] THEN
    CONV_TAC INTEGER_RULE;
    REWRITE_TAC[option_INJ; PAIR_EQ]] THEN

  RULE_ASSUM_TAC(REWRITE_RULE
   [montgomery_decode; INTEGER_MOD_RING_CLAUSES]) THEN
  CONJ_TAC THEN MATCH_MP_TAC fdivlemma THEN
  SIMP_TAC[INTEGER_MOD_RING_CARRIER_REM; montgomery_decode; modular_decode;
           RING_POW; FIELD_POW_EQ_0; FIELD_INTEGER_MOD_RING; PRIME_P256] THEN
  ASM_REWRITE_TAC[ARITH_EQ; INTEGER_MOD_RING_CLAUSES] THEN
  ASM_REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES] THEN
  CONV_TAC INT_REM_DOWN_CONV THEN REWRITE_TAC[INT_REM_EQ] THEN
  REWRITE_TAC[INT_MUL_ASSOC] THEN MATCH_MP_TAC(INTEGER_RULE
   `!e:int. (i * e == &1) (mod p) /\ (b * e == a) (mod p)
            ==> (a * i == b) (mod p)`) THEN
  EXISTS_TAC `&Z3':int` THEN
  (CONJ_TAC THENL
    [REWRITE_TAC[GSYM num_congruent; INT_OF_NUM_CLAUSES] THEN
     ASM_SIMP_TAC[INVERSE_MOD_LMUL_EQ; PRIME_COPRIME_EQ; PRIME_P256];
     ASM_REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES]]) THEN
  REWRITE_TAC[GSYM INT_REM_EQ] THEN CONV_TAC INT_REM_DOWN_CONV THEN
  REWRITE_TAC[GSYM INT_MUL_ASSOC; INT_REM_EQ] THEN
  CONV_TAC INTEGER_RULE);;

let P256_SCALARMUL_ALT_SUBROUTINE_CORRECT = time prove
 (`!res scalar point n xy pc stackpointer returnaddress.
        aligned 16 stackpointer /\
        ALL (nonoverlapping (word_sub stackpointer (word 1248),1248))
            [(word pc,0x624c); (res,64); (scalar,32); (point,64)] /\
        nonoverlapping (res,64) (word pc,0x624c)
        ==> ensures arm
             (\s. aligned_bytes_loaded s (word pc) p256_scalarmul_alt_mc /\
                  read PC s = word pc /\
                  read SP s = stackpointer /\
                  read X30 s = returnaddress /\
                  C_ARGUMENTS [res;scalar;point] s /\
                  bignum_from_memory (scalar,4) s = n /\
                  bignum_pair_from_memory (point,4) s = xy)
             (\s. read PC s = returnaddress /\
                  !P. P IN group_carrier p256_group /\
                      affinepoint_p256 xy P
                      ==> affinepointz_p256
                            (bignum_pair_from_memory(res,4) s)
                            (group_pow p256_group P n))
          (MAYCHANGE_REGS_AND_FLAGS_PERMITTED_BY_ABI ,,
           MAYCHANGE[memory :> bytes(res,64);
                     memory :> bytes(word_sub stackpointer (word 1248),1248)])`,
   ARM_ADD_RETURN_STACK_TAC P256_SCALARMUL_ALT_EXEC
   P256_SCALARMUL_ALT_CORRECT `[X19; X20; X21; X30]` 1248);;
