(*
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0 OR ISC OR MIT-0
 *)

(* ========================================================================= *)
(* Point addition in Montgomery-Jacobian coordinates for NIST P-384 curve.   *)
(* ========================================================================= *)

needs "arm/proofs/base.ml";;
needs "common/ecencoding.ml";;
needs "EC/jacobian.ml";;
needs "EC/nistp384.ml";;

prioritize_int();;
prioritize_real();;
prioritize_num();;

(**** print_literal_from_elf "arm/p384/p384_montjadd_alt.o";;
 ****)

let p384_montjadd_alt_mc = define_assert_from_elf
  "p384_montjadd_alt_mc" "arm/p384/p384_montjadd_alt.o"
[
  0xa9bf53f3;       (* arm_STP X19 X20 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xa9bf5bf5;       (* arm_STP X21 X22 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xa9bf63f7;       (* arm_STP X23 X24 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xa9bf6bf9;       (* arm_STP X25 X26 SP (Preimmediate_Offset (iword (-- &16))) *)
  0xd10543ff;       (* arm_SUB SP SP (rvalue (word 336)) *)
  0xaa0003f8;       (* arm_MOV X24 X0 *)
  0xaa0103f9;       (* arm_MOV X25 X1 *)
  0xaa0203fa;       (* arm_MOV X26 X2 *)
  0xa9460f22;       (* arm_LDP X2 X3 X25 (Immediate_Offset (iword (&96))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa9471724;       (* arm_LDP X4 X5 X25 (Immediate_Offset (iword (&112))) *)
  0x9b047c48;       (* arm_MUL X8 X2 X4 *)
  0xab08014a;       (* arm_ADDS X10 X10 X8 *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9b047c68;       (* arm_MUL X8 X3 X4 *)
  0xba08016b;       (* arm_ADCS X11 X11 X8 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b057c68;       (* arm_MUL X8 X3 X5 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0xa9481f26;       (* arm_LDP X6 X7 X25 (Immediate_Offset (iword (&128))) *)
  0x9b077c4d;       (* arm_MUL X13 X2 X7 *)
  0x9b067c68;       (* arm_MUL X8 X3 X6 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9bc77c4e;       (* arm_UMULH X14 X2 X7 *)
  0x9b077c68;       (* arm_MUL X8 X3 X7 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9b067caf;       (* arm_MUL X15 X5 X6 *)
  0xba1f01ef;       (* arm_ADCS X15 X15 XZR *)
  0x9bc67cb0;       (* arm_UMULH X16 X5 X6 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0x9bc47c48;       (* arm_UMULH X8 X2 X4 *)
  0xab08016b;       (* arm_ADDS X11 X11 X8 *)
  0x9bc47c68;       (* arm_UMULH X8 X3 X4 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0x9bc57c68;       (* arm_UMULH X8 X3 X5 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9bc67c68;       (* arm_UMULH X8 X3 X6 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc77c68;       (* arm_UMULH X8 X3 X7 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0x9b067c48;       (* arm_MUL X8 X2 X6 *)
  0xab08018c;       (* arm_ADDS X12 X12 X8 *)
  0x9b057c88;       (* arm_MUL X8 X4 X5 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9b067c88;       (* arm_MUL X8 X4 X6 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9b077c88;       (* arm_MUL X8 X4 X7 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9b077ca8;       (* arm_MUL X8 X5 X7 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9b077cd1;       (* arm_MUL X17 X6 X7 *)
  0xba1f0231;       (* arm_ADCS X17 X17 XZR *)
  0x9bc77cd3;       (* arm_UMULH X19 X6 X7 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9bc67c48;       (* arm_UMULH X8 X2 X6 *)
  0xab0801ad;       (* arm_ADDS X13 X13 X8 *)
  0x9bc57c88;       (* arm_UMULH X8 X4 X5 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc67c88;       (* arm_UMULH X8 X4 X6 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9bc77c88;       (* arm_UMULH X8 X4 X7 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9bc77ca8;       (* arm_UMULH X8 X5 X7 *)
  0xba080231;       (* arm_ADCS X17 X17 X8 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0xba0f01ef;       (* arm_ADCS X15 X15 X15 *)
  0xba100210;       (* arm_ADCS X16 X16 X16 *)
  0xba110231;       (* arm_ADCS X17 X17 X17 *)
  0xba130273;       (* arm_ADCS X19 X19 X19 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc27c48;       (* arm_UMULH X8 X2 X2 *)
  0x9b027c42;       (* arm_MUL X2 X2 X2 *)
  0xab080129;       (* arm_ADDS X9 X9 X8 *)
  0x9b037c68;       (* arm_MUL X8 X3 X3 *)
  0xba08014a;       (* arm_ADCS X10 X10 X8 *)
  0x9bc37c68;       (* arm_UMULH X8 X3 X3 *)
  0xba08016b;       (* arm_ADCS X11 X11 X8 *)
  0x9b047c88;       (* arm_MUL X8 X4 X4 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0x9bc47c88;       (* arm_UMULH X8 X4 X4 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9b057ca8;       (* arm_MUL X8 X5 X5 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc57ca8;       (* arm_UMULH X8 X5 X5 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9b067cc8;       (* arm_MUL X8 X6 X6 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9bc67cc8;       (* arm_UMULH X8 X6 X6 *)
  0xba080231;       (* arm_ADCS X17 X17 X8 *)
  0x9b077ce8;       (* arm_MUL X8 X7 X7 *)
  0xba080273;       (* arm_ADCS X19 X19 X8 *)
  0x9bc77ce8;       (* arm_UMULH X8 X7 X7 *)
  0x9a080294;       (* arm_ADC X20 X20 X8 *)
  0xd3607c45;       (* arm_LSL X5 X2 32 *)
  0x8b0200a2;       (* arm_ADD X2 X5 X2 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bc27ca5;       (* arm_UMULH X5 X5 X2 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b027c83;       (* arm_MUL X3 X4 X2 *)
  0x9bc27c84;       (* arm_UMULH X4 X4 X2 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba020084;       (* arm_ADCS X4 X4 X2 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb050129;       (* arm_SUBS X9 X9 X5 *)
  0xfa04014a;       (* arm_SBCS X10 X10 X4 *)
  0xfa03016b;       (* arm_SBCS X11 X11 X3 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f0042;       (* arm_SBC X2 X2 XZR *)
  0xd3607d25;       (* arm_LSL X5 X9 32 *)
  0x8b0900a9;       (* arm_ADD X9 X5 X9 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bc97ca5;       (* arm_UMULH X5 X5 X9 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b097c83;       (* arm_MUL X3 X4 X9 *)
  0x9bc97c84;       (* arm_UMULH X4 X4 X9 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba090084;       (* arm_ADCS X4 X4 X9 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05014a;       (* arm_SUBS X10 X10 X5 *)
  0xfa04016b;       (* arm_SBCS X11 X11 X4 *)
  0xfa03018c;       (* arm_SBCS X12 X12 X3 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xda1f0129;       (* arm_SBC X9 X9 XZR *)
  0xd3607d45;       (* arm_LSL X5 X10 32 *)
  0x8b0a00aa;       (* arm_ADD X10 X5 X10 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bca7ca5;       (* arm_UMULH X5 X5 X10 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0a7c83;       (* arm_MUL X3 X4 X10 *)
  0x9bca7c84;       (* arm_UMULH X4 X4 X10 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0a0084;       (* arm_ADCS X4 X4 X10 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05016b;       (* arm_SUBS X11 X11 X5 *)
  0xfa04018c;       (* arm_SBCS X12 X12 X4 *)
  0xfa0301ad;       (* arm_SBCS X13 X13 X3 *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xfa1f0129;       (* arm_SBCS X9 X9 XZR *)
  0xda1f014a;       (* arm_SBC X10 X10 XZR *)
  0xd3607d65;       (* arm_LSL X5 X11 32 *)
  0x8b0b00ab;       (* arm_ADD X11 X5 X11 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcb7ca5;       (* arm_UMULH X5 X5 X11 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0b7c83;       (* arm_MUL X3 X4 X11 *)
  0x9bcb7c84;       (* arm_UMULH X4 X4 X11 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05018c;       (* arm_SUBS X12 X12 X5 *)
  0xfa0401ad;       (* arm_SBCS X13 X13 X4 *)
  0xfa030042;       (* arm_SBCS X2 X2 X3 *)
  0xfa1f0129;       (* arm_SBCS X9 X9 XZR *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda1f016b;       (* arm_SBC X11 X11 XZR *)
  0xd3607d85;       (* arm_LSL X5 X12 32 *)
  0x8b0c00ac;       (* arm_ADD X12 X5 X12 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ca5;       (* arm_UMULH X5 X5 X12 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0c7c83;       (* arm_MUL X3 X4 X12 *)
  0x9bcc7c84;       (* arm_UMULH X4 X4 X12 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0c0084;       (* arm_ADCS X4 X4 X12 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb0501ad;       (* arm_SUBS X13 X13 X5 *)
  0xfa040042;       (* arm_SBCS X2 X2 X4 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xfa1f016b;       (* arm_SBCS X11 X11 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da5;       (* arm_LSL X5 X13 32 *)
  0x8b0d00ad;       (* arm_ADD X13 X5 X13 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ca5;       (* arm_UMULH X5 X5 X13 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0d7c83;       (* arm_MUL X3 X4 X13 *)
  0x9bcd7c84;       (* arm_UMULH X4 X4 X13 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0d0084;       (* arm_ADCS X4 X4 X13 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb050042;       (* arm_SUBS X2 X2 X5 *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xfa1f016b;       (* arm_SBCS X11 X11 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xab0e0042;       (* arm_ADDS X2 X2 X14 *)
  0xba0f0129;       (* arm_ADCS X9 X9 X15 *)
  0xba10014a;       (* arm_ADCS X10 X10 X16 *)
  0xba11016b;       (* arm_ADCS X11 X11 X17 *)
  0xba13018c;       (* arm_ADCS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xb26083ee;       (* arm_MOV X14 (rvalue (word 18446744069414584321)) *)
  0xb2407fef;       (* arm_MOV X15 (rvalue (word 4294967295)) *)
  0x9a9f21ce;       (* arm_CSEL X14 X14 XZR Condition_CS *)
  0x9a9f21ef;       (* arm_CSEL X15 X15 XZR Condition_CS *)
  0x9a9f37f0;       (* arm_CSET X16 Condition_CS *)
  0xab0e0042;       (* arm_ADDS X2 X2 X14 *)
  0xba0f0129;       (* arm_ADCS X9 X9 X15 *)
  0xba10014a;       (* arm_ADCS X10 X10 X16 *)
  0xba1f016b;       (* arm_ADCS X11 X11 XZR *)
  0xba1f018c;       (* arm_ADCS X12 X12 XZR *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xa90027e2;       (* arm_STP X2 X9 SP (Immediate_Offset (iword (&0))) *)
  0xa9012fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&16))) *)
  0xa90237ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&32))) *)
  0xa9460f42;       (* arm_LDP X2 X3 X26 (Immediate_Offset (iword (&96))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa9471744;       (* arm_LDP X4 X5 X26 (Immediate_Offset (iword (&112))) *)
  0x9b047c48;       (* arm_MUL X8 X2 X4 *)
  0xab08014a;       (* arm_ADDS X10 X10 X8 *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9b047c68;       (* arm_MUL X8 X3 X4 *)
  0xba08016b;       (* arm_ADCS X11 X11 X8 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b057c68;       (* arm_MUL X8 X3 X5 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0xa9481f46;       (* arm_LDP X6 X7 X26 (Immediate_Offset (iword (&128))) *)
  0x9b077c4d;       (* arm_MUL X13 X2 X7 *)
  0x9b067c68;       (* arm_MUL X8 X3 X6 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9bc77c4e;       (* arm_UMULH X14 X2 X7 *)
  0x9b077c68;       (* arm_MUL X8 X3 X7 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9b067caf;       (* arm_MUL X15 X5 X6 *)
  0xba1f01ef;       (* arm_ADCS X15 X15 XZR *)
  0x9bc67cb0;       (* arm_UMULH X16 X5 X6 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0x9bc47c48;       (* arm_UMULH X8 X2 X4 *)
  0xab08016b;       (* arm_ADDS X11 X11 X8 *)
  0x9bc47c68;       (* arm_UMULH X8 X3 X4 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0x9bc57c68;       (* arm_UMULH X8 X3 X5 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9bc67c68;       (* arm_UMULH X8 X3 X6 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc77c68;       (* arm_UMULH X8 X3 X7 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0x9b067c48;       (* arm_MUL X8 X2 X6 *)
  0xab08018c;       (* arm_ADDS X12 X12 X8 *)
  0x9b057c88;       (* arm_MUL X8 X4 X5 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9b067c88;       (* arm_MUL X8 X4 X6 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9b077c88;       (* arm_MUL X8 X4 X7 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9b077ca8;       (* arm_MUL X8 X5 X7 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9b077cd1;       (* arm_MUL X17 X6 X7 *)
  0xba1f0231;       (* arm_ADCS X17 X17 XZR *)
  0x9bc77cd3;       (* arm_UMULH X19 X6 X7 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9bc67c48;       (* arm_UMULH X8 X2 X6 *)
  0xab0801ad;       (* arm_ADDS X13 X13 X8 *)
  0x9bc57c88;       (* arm_UMULH X8 X4 X5 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc67c88;       (* arm_UMULH X8 X4 X6 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9bc77c88;       (* arm_UMULH X8 X4 X7 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9bc77ca8;       (* arm_UMULH X8 X5 X7 *)
  0xba080231;       (* arm_ADCS X17 X17 X8 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0xba0f01ef;       (* arm_ADCS X15 X15 X15 *)
  0xba100210;       (* arm_ADCS X16 X16 X16 *)
  0xba110231;       (* arm_ADCS X17 X17 X17 *)
  0xba130273;       (* arm_ADCS X19 X19 X19 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc27c48;       (* arm_UMULH X8 X2 X2 *)
  0x9b027c42;       (* arm_MUL X2 X2 X2 *)
  0xab080129;       (* arm_ADDS X9 X9 X8 *)
  0x9b037c68;       (* arm_MUL X8 X3 X3 *)
  0xba08014a;       (* arm_ADCS X10 X10 X8 *)
  0x9bc37c68;       (* arm_UMULH X8 X3 X3 *)
  0xba08016b;       (* arm_ADCS X11 X11 X8 *)
  0x9b047c88;       (* arm_MUL X8 X4 X4 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0x9bc47c88;       (* arm_UMULH X8 X4 X4 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9b057ca8;       (* arm_MUL X8 X5 X5 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc57ca8;       (* arm_UMULH X8 X5 X5 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9b067cc8;       (* arm_MUL X8 X6 X6 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9bc67cc8;       (* arm_UMULH X8 X6 X6 *)
  0xba080231;       (* arm_ADCS X17 X17 X8 *)
  0x9b077ce8;       (* arm_MUL X8 X7 X7 *)
  0xba080273;       (* arm_ADCS X19 X19 X8 *)
  0x9bc77ce8;       (* arm_UMULH X8 X7 X7 *)
  0x9a080294;       (* arm_ADC X20 X20 X8 *)
  0xd3607c45;       (* arm_LSL X5 X2 32 *)
  0x8b0200a2;       (* arm_ADD X2 X5 X2 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bc27ca5;       (* arm_UMULH X5 X5 X2 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b027c83;       (* arm_MUL X3 X4 X2 *)
  0x9bc27c84;       (* arm_UMULH X4 X4 X2 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba020084;       (* arm_ADCS X4 X4 X2 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb050129;       (* arm_SUBS X9 X9 X5 *)
  0xfa04014a;       (* arm_SBCS X10 X10 X4 *)
  0xfa03016b;       (* arm_SBCS X11 X11 X3 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f0042;       (* arm_SBC X2 X2 XZR *)
  0xd3607d25;       (* arm_LSL X5 X9 32 *)
  0x8b0900a9;       (* arm_ADD X9 X5 X9 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bc97ca5;       (* arm_UMULH X5 X5 X9 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b097c83;       (* arm_MUL X3 X4 X9 *)
  0x9bc97c84;       (* arm_UMULH X4 X4 X9 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba090084;       (* arm_ADCS X4 X4 X9 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05014a;       (* arm_SUBS X10 X10 X5 *)
  0xfa04016b;       (* arm_SBCS X11 X11 X4 *)
  0xfa03018c;       (* arm_SBCS X12 X12 X3 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xda1f0129;       (* arm_SBC X9 X9 XZR *)
  0xd3607d45;       (* arm_LSL X5 X10 32 *)
  0x8b0a00aa;       (* arm_ADD X10 X5 X10 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bca7ca5;       (* arm_UMULH X5 X5 X10 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0a7c83;       (* arm_MUL X3 X4 X10 *)
  0x9bca7c84;       (* arm_UMULH X4 X4 X10 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0a0084;       (* arm_ADCS X4 X4 X10 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05016b;       (* arm_SUBS X11 X11 X5 *)
  0xfa04018c;       (* arm_SBCS X12 X12 X4 *)
  0xfa0301ad;       (* arm_SBCS X13 X13 X3 *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xfa1f0129;       (* arm_SBCS X9 X9 XZR *)
  0xda1f014a;       (* arm_SBC X10 X10 XZR *)
  0xd3607d65;       (* arm_LSL X5 X11 32 *)
  0x8b0b00ab;       (* arm_ADD X11 X5 X11 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcb7ca5;       (* arm_UMULH X5 X5 X11 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0b7c83;       (* arm_MUL X3 X4 X11 *)
  0x9bcb7c84;       (* arm_UMULH X4 X4 X11 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05018c;       (* arm_SUBS X12 X12 X5 *)
  0xfa0401ad;       (* arm_SBCS X13 X13 X4 *)
  0xfa030042;       (* arm_SBCS X2 X2 X3 *)
  0xfa1f0129;       (* arm_SBCS X9 X9 XZR *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda1f016b;       (* arm_SBC X11 X11 XZR *)
  0xd3607d85;       (* arm_LSL X5 X12 32 *)
  0x8b0c00ac;       (* arm_ADD X12 X5 X12 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ca5;       (* arm_UMULH X5 X5 X12 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0c7c83;       (* arm_MUL X3 X4 X12 *)
  0x9bcc7c84;       (* arm_UMULH X4 X4 X12 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0c0084;       (* arm_ADCS X4 X4 X12 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb0501ad;       (* arm_SUBS X13 X13 X5 *)
  0xfa040042;       (* arm_SBCS X2 X2 X4 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xfa1f016b;       (* arm_SBCS X11 X11 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da5;       (* arm_LSL X5 X13 32 *)
  0x8b0d00ad;       (* arm_ADD X13 X5 X13 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ca5;       (* arm_UMULH X5 X5 X13 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0d7c83;       (* arm_MUL X3 X4 X13 *)
  0x9bcd7c84;       (* arm_UMULH X4 X4 X13 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0d0084;       (* arm_ADCS X4 X4 X13 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb050042;       (* arm_SUBS X2 X2 X5 *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xfa1f016b;       (* arm_SBCS X11 X11 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xab0e0042;       (* arm_ADDS X2 X2 X14 *)
  0xba0f0129;       (* arm_ADCS X9 X9 X15 *)
  0xba10014a;       (* arm_ADCS X10 X10 X16 *)
  0xba11016b;       (* arm_ADCS X11 X11 X17 *)
  0xba13018c;       (* arm_ADCS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xb26083ee;       (* arm_MOV X14 (rvalue (word 18446744069414584321)) *)
  0xb2407fef;       (* arm_MOV X15 (rvalue (word 4294967295)) *)
  0x9a9f21ce;       (* arm_CSEL X14 X14 XZR Condition_CS *)
  0x9a9f21ef;       (* arm_CSEL X15 X15 XZR Condition_CS *)
  0x9a9f37f0;       (* arm_CSET X16 Condition_CS *)
  0xab0e0042;       (* arm_ADDS X2 X2 X14 *)
  0xba0f0129;       (* arm_ADCS X9 X9 X15 *)
  0xba10014a;       (* arm_ADCS X10 X10 X16 *)
  0xba1f016b;       (* arm_ADCS X11 X11 XZR *)
  0xba1f018c;       (* arm_ADCS X12 X12 XZR *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xa90f27e2;       (* arm_STP X2 X9 SP (Immediate_Offset (iword (&240))) *)
  0xa9102fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&256))) *)
  0xa91137ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&272))) *)
  0xa9461343;       (* arm_LDP X3 X4 X26 (Immediate_Offset (iword (&96))) *)
  0xa9431b25;       (* arm_LDP X5 X6 X25 (Immediate_Offset (iword (&48))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9442327;       (* arm_LDP X7 X8 X25 (Immediate_Offset (iword (&64))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9452b29;       (* arm_LDP X9 X10 X25 (Immediate_Offset (iword (&80))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa9471343;       (* arm_LDP X3 X4 X26 (Immediate_Offset (iword (&112))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa9481343;       (* arm_LDP X3 X4 X26 (Immediate_Offset (iword (&128))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa91237ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&288))) *)
  0xa9133fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&304))) *)
  0xa91447f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&320))) *)
  0xa9461323;       (* arm_LDP X3 X4 X25 (Immediate_Offset (iword (&96))) *)
  0xa9431b45;       (* arm_LDP X5 X6 X26 (Immediate_Offset (iword (&48))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9442347;       (* arm_LDP X7 X8 X26 (Immediate_Offset (iword (&64))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9452b49;       (* arm_LDP X9 X10 X26 (Immediate_Offset (iword (&80))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa9471323;       (* arm_LDP X3 X4 X25 (Immediate_Offset (iword (&112))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa9481323;       (* arm_LDP X3 X4 X25 (Immediate_Offset (iword (&128))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90337ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&48))) *)
  0xa9043fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&64))) *)
  0xa90547f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&80))) *)
  0xa94013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&0))) *)
  0xa9401b45;       (* arm_LDP X5 X6 X26 (Immediate_Offset (iword (&0))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9412347;       (* arm_LDP X7 X8 X26 (Immediate_Offset (iword (&16))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9422b49;       (* arm_LDP X9 X10 X26 (Immediate_Offset (iword (&32))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa94113e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&16))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa94213e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&32))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90637ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&96))) *)
  0xa9073fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&112))) *)
  0xa90847f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&128))) *)
  0xa94f13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&240))) *)
  0xa9401b25;       (* arm_LDP X5 X6 X25 (Immediate_Offset (iword (&0))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9412327;       (* arm_LDP X7 X8 X25 (Immediate_Offset (iword (&16))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9422b29;       (* arm_LDP X9 X10 X25 (Immediate_Offset (iword (&32))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa95013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&256))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa95113e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&272))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90c37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&192))) *)
  0xa90d3fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&208))) *)
  0xa90e47f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&224))) *)
  0xa94013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&0))) *)
  0xa9431be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&48))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa94423e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&64))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9452be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&80))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa94113e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&16))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa94213e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&32))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90337ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&48))) *)
  0xa9043fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&64))) *)
  0xa90547f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&80))) *)
  0xa94f13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&240))) *)
  0xa9521be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&288))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa95323e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&304))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9542be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&320))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa95013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&256))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa95113e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&272))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa91237ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&288))) *)
  0xa9133fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&304))) *)
  0xa91447f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&320))) *)
  0xa9461be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&96))) *)
  0xa94c0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&192))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94723e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&112))) *)
  0xa94d0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&208))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xa9482be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&128))) *)
  0xa94e0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&224))) *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xca030084;       (* arm_EOR X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0x92800024;       (* arm_MOVN X4 (word 1) 0 *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9a03014a;       (* arm_ADC X10 X10 X3 *)
  0xa90f1be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&240))) *)
  0xa91023e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&256))) *)
  0xa9112be9;       (* arm_STP X9 X10 SP (Immediate_Offset (iword (&272))) *)
  0xa9431be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&48))) *)
  0xa9520fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&288))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94423e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&64))) *)
  0xa9530fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&304))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xa9452be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&80))) *)
  0xa9540fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&320))) *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xca030084;       (* arm_EOR X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0x92800024;       (* arm_MOVN X4 (word 1) 0 *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9a03014a;       (* arm_ADC X10 X10 X3 *)
  0xa9031be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&48))) *)
  0xa90423e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&64))) *)
  0xa9052be9;       (* arm_STP X9 X10 SP (Immediate_Offset (iword (&80))) *)
  0xa94f0fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&240))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa95017e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&256))) *)
  0x9b047c48;       (* arm_MUL X8 X2 X4 *)
  0xab08014a;       (* arm_ADDS X10 X10 X8 *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9b047c68;       (* arm_MUL X8 X3 X4 *)
  0xba08016b;       (* arm_ADCS X11 X11 X8 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b057c68;       (* arm_MUL X8 X3 X5 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0xa9511fe6;       (* arm_LDP X6 X7 SP (Immediate_Offset (iword (&272))) *)
  0x9b077c4d;       (* arm_MUL X13 X2 X7 *)
  0x9b067c68;       (* arm_MUL X8 X3 X6 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9bc77c4e;       (* arm_UMULH X14 X2 X7 *)
  0x9b077c68;       (* arm_MUL X8 X3 X7 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9b067caf;       (* arm_MUL X15 X5 X6 *)
  0xba1f01ef;       (* arm_ADCS X15 X15 XZR *)
  0x9bc67cb0;       (* arm_UMULH X16 X5 X6 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0x9bc47c48;       (* arm_UMULH X8 X2 X4 *)
  0xab08016b;       (* arm_ADDS X11 X11 X8 *)
  0x9bc47c68;       (* arm_UMULH X8 X3 X4 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0x9bc57c68;       (* arm_UMULH X8 X3 X5 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9bc67c68;       (* arm_UMULH X8 X3 X6 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc77c68;       (* arm_UMULH X8 X3 X7 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0x9b067c48;       (* arm_MUL X8 X2 X6 *)
  0xab08018c;       (* arm_ADDS X12 X12 X8 *)
  0x9b057c88;       (* arm_MUL X8 X4 X5 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9b067c88;       (* arm_MUL X8 X4 X6 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9b077c88;       (* arm_MUL X8 X4 X7 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9b077ca8;       (* arm_MUL X8 X5 X7 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9b077cd1;       (* arm_MUL X17 X6 X7 *)
  0xba1f0231;       (* arm_ADCS X17 X17 XZR *)
  0x9bc77cd3;       (* arm_UMULH X19 X6 X7 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9bc67c48;       (* arm_UMULH X8 X2 X6 *)
  0xab0801ad;       (* arm_ADDS X13 X13 X8 *)
  0x9bc57c88;       (* arm_UMULH X8 X4 X5 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc67c88;       (* arm_UMULH X8 X4 X6 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9bc77c88;       (* arm_UMULH X8 X4 X7 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9bc77ca8;       (* arm_UMULH X8 X5 X7 *)
  0xba080231;       (* arm_ADCS X17 X17 X8 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0xba0f01ef;       (* arm_ADCS X15 X15 X15 *)
  0xba100210;       (* arm_ADCS X16 X16 X16 *)
  0xba110231;       (* arm_ADCS X17 X17 X17 *)
  0xba130273;       (* arm_ADCS X19 X19 X19 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc27c48;       (* arm_UMULH X8 X2 X2 *)
  0x9b027c42;       (* arm_MUL X2 X2 X2 *)
  0xab080129;       (* arm_ADDS X9 X9 X8 *)
  0x9b037c68;       (* arm_MUL X8 X3 X3 *)
  0xba08014a;       (* arm_ADCS X10 X10 X8 *)
  0x9bc37c68;       (* arm_UMULH X8 X3 X3 *)
  0xba08016b;       (* arm_ADCS X11 X11 X8 *)
  0x9b047c88;       (* arm_MUL X8 X4 X4 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0x9bc47c88;       (* arm_UMULH X8 X4 X4 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9b057ca8;       (* arm_MUL X8 X5 X5 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc57ca8;       (* arm_UMULH X8 X5 X5 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9b067cc8;       (* arm_MUL X8 X6 X6 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9bc67cc8;       (* arm_UMULH X8 X6 X6 *)
  0xba080231;       (* arm_ADCS X17 X17 X8 *)
  0x9b077ce8;       (* arm_MUL X8 X7 X7 *)
  0xba080273;       (* arm_ADCS X19 X19 X8 *)
  0x9bc77ce8;       (* arm_UMULH X8 X7 X7 *)
  0x9a080294;       (* arm_ADC X20 X20 X8 *)
  0xd3607c45;       (* arm_LSL X5 X2 32 *)
  0x8b0200a2;       (* arm_ADD X2 X5 X2 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bc27ca5;       (* arm_UMULH X5 X5 X2 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b027c83;       (* arm_MUL X3 X4 X2 *)
  0x9bc27c84;       (* arm_UMULH X4 X4 X2 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba020084;       (* arm_ADCS X4 X4 X2 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb050129;       (* arm_SUBS X9 X9 X5 *)
  0xfa04014a;       (* arm_SBCS X10 X10 X4 *)
  0xfa03016b;       (* arm_SBCS X11 X11 X3 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f0042;       (* arm_SBC X2 X2 XZR *)
  0xd3607d25;       (* arm_LSL X5 X9 32 *)
  0x8b0900a9;       (* arm_ADD X9 X5 X9 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bc97ca5;       (* arm_UMULH X5 X5 X9 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b097c83;       (* arm_MUL X3 X4 X9 *)
  0x9bc97c84;       (* arm_UMULH X4 X4 X9 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba090084;       (* arm_ADCS X4 X4 X9 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05014a;       (* arm_SUBS X10 X10 X5 *)
  0xfa04016b;       (* arm_SBCS X11 X11 X4 *)
  0xfa03018c;       (* arm_SBCS X12 X12 X3 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xda1f0129;       (* arm_SBC X9 X9 XZR *)
  0xd3607d45;       (* arm_LSL X5 X10 32 *)
  0x8b0a00aa;       (* arm_ADD X10 X5 X10 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bca7ca5;       (* arm_UMULH X5 X5 X10 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0a7c83;       (* arm_MUL X3 X4 X10 *)
  0x9bca7c84;       (* arm_UMULH X4 X4 X10 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0a0084;       (* arm_ADCS X4 X4 X10 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05016b;       (* arm_SUBS X11 X11 X5 *)
  0xfa04018c;       (* arm_SBCS X12 X12 X4 *)
  0xfa0301ad;       (* arm_SBCS X13 X13 X3 *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xfa1f0129;       (* arm_SBCS X9 X9 XZR *)
  0xda1f014a;       (* arm_SBC X10 X10 XZR *)
  0xd3607d65;       (* arm_LSL X5 X11 32 *)
  0x8b0b00ab;       (* arm_ADD X11 X5 X11 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcb7ca5;       (* arm_UMULH X5 X5 X11 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0b7c83;       (* arm_MUL X3 X4 X11 *)
  0x9bcb7c84;       (* arm_UMULH X4 X4 X11 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05018c;       (* arm_SUBS X12 X12 X5 *)
  0xfa0401ad;       (* arm_SBCS X13 X13 X4 *)
  0xfa030042;       (* arm_SBCS X2 X2 X3 *)
  0xfa1f0129;       (* arm_SBCS X9 X9 XZR *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda1f016b;       (* arm_SBC X11 X11 XZR *)
  0xd3607d85;       (* arm_LSL X5 X12 32 *)
  0x8b0c00ac;       (* arm_ADD X12 X5 X12 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ca5;       (* arm_UMULH X5 X5 X12 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0c7c83;       (* arm_MUL X3 X4 X12 *)
  0x9bcc7c84;       (* arm_UMULH X4 X4 X12 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0c0084;       (* arm_ADCS X4 X4 X12 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb0501ad;       (* arm_SUBS X13 X13 X5 *)
  0xfa040042;       (* arm_SBCS X2 X2 X4 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xfa1f016b;       (* arm_SBCS X11 X11 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da5;       (* arm_LSL X5 X13 32 *)
  0x8b0d00ad;       (* arm_ADD X13 X5 X13 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ca5;       (* arm_UMULH X5 X5 X13 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0d7c83;       (* arm_MUL X3 X4 X13 *)
  0x9bcd7c84;       (* arm_UMULH X4 X4 X13 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0d0084;       (* arm_ADCS X4 X4 X13 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb050042;       (* arm_SUBS X2 X2 X5 *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xfa1f016b;       (* arm_SBCS X11 X11 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xab0e0042;       (* arm_ADDS X2 X2 X14 *)
  0xba0f0129;       (* arm_ADCS X9 X9 X15 *)
  0xba10014a;       (* arm_ADCS X10 X10 X16 *)
  0xba11016b;       (* arm_ADCS X11 X11 X17 *)
  0xba13018c;       (* arm_ADCS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xb26083ee;       (* arm_MOV X14 (rvalue (word 18446744069414584321)) *)
  0xb2407fef;       (* arm_MOV X15 (rvalue (word 4294967295)) *)
  0x9a9f21ce;       (* arm_CSEL X14 X14 XZR Condition_CS *)
  0x9a9f21ef;       (* arm_CSEL X15 X15 XZR Condition_CS *)
  0x9a9f37f0;       (* arm_CSET X16 Condition_CS *)
  0xab0e0042;       (* arm_ADDS X2 X2 X14 *)
  0xba0f0129;       (* arm_ADCS X9 X9 X15 *)
  0xba10014a;       (* arm_ADCS X10 X10 X16 *)
  0xba1f016b;       (* arm_ADCS X11 X11 XZR *)
  0xba1f018c;       (* arm_ADCS X12 X12 XZR *)
  0x9a1f01ad;       (* arm_ADC X13 X13 XZR *)
  0xa90927e2;       (* arm_STP X2 X9 SP (Immediate_Offset (iword (&144))) *)
  0xa90a2fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&160))) *)
  0xa90b37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&176))) *)
  0xa9430fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&48))) *)
  0x9b037c49;       (* arm_MUL X9 X2 X3 *)
  0x9bc37c4a;       (* arm_UMULH X10 X2 X3 *)
  0xa94417e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&64))) *)
  0x9b047c48;       (* arm_MUL X8 X2 X4 *)
  0xab08014a;       (* arm_ADDS X10 X10 X8 *)
  0x9b057c4b;       (* arm_MUL X11 X2 X5 *)
  0x9b047c68;       (* arm_MUL X8 X3 X4 *)
  0xba08016b;       (* arm_ADCS X11 X11 X8 *)
  0x9bc57c4c;       (* arm_UMULH X12 X2 X5 *)
  0x9b057c68;       (* arm_MUL X8 X3 X5 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0xa9451fe6;       (* arm_LDP X6 X7 SP (Immediate_Offset (iword (&80))) *)
  0x9b077c4d;       (* arm_MUL X13 X2 X7 *)
  0x9b067c68;       (* arm_MUL X8 X3 X6 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9bc77c4e;       (* arm_UMULH X14 X2 X7 *)
  0x9b077c68;       (* arm_MUL X8 X3 X7 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9b067caf;       (* arm_MUL X15 X5 X6 *)
  0xba1f01ef;       (* arm_ADCS X15 X15 XZR *)
  0x9bc67cb0;       (* arm_UMULH X16 X5 X6 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0x9bc47c48;       (* arm_UMULH X8 X2 X4 *)
  0xab08016b;       (* arm_ADDS X11 X11 X8 *)
  0x9bc47c68;       (* arm_UMULH X8 X3 X4 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0x9bc57c68;       (* arm_UMULH X8 X3 X5 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9bc67c68;       (* arm_UMULH X8 X3 X6 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc77c68;       (* arm_UMULH X8 X3 X7 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9a1f0210;       (* arm_ADC X16 X16 XZR *)
  0x9b067c48;       (* arm_MUL X8 X2 X6 *)
  0xab08018c;       (* arm_ADDS X12 X12 X8 *)
  0x9b057c88;       (* arm_MUL X8 X4 X5 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9b067c88;       (* arm_MUL X8 X4 X6 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9b077c88;       (* arm_MUL X8 X4 X7 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9b077ca8;       (* arm_MUL X8 X5 X7 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9b077cd1;       (* arm_MUL X17 X6 X7 *)
  0xba1f0231;       (* arm_ADCS X17 X17 XZR *)
  0x9bc77cd3;       (* arm_UMULH X19 X6 X7 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9bc67c48;       (* arm_UMULH X8 X2 X6 *)
  0xab0801ad;       (* arm_ADDS X13 X13 X8 *)
  0x9bc57c88;       (* arm_UMULH X8 X4 X5 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc67c88;       (* arm_UMULH X8 X4 X6 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9bc77c88;       (* arm_UMULH X8 X4 X7 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9bc77ca8;       (* arm_UMULH X8 X5 X7 *)
  0xba080231;       (* arm_ADCS X17 X17 X8 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0xab090129;       (* arm_ADDS X9 X9 X9 *)
  0xba0a014a;       (* arm_ADCS X10 X10 X10 *)
  0xba0b016b;       (* arm_ADCS X11 X11 X11 *)
  0xba0c018c;       (* arm_ADCS X12 X12 X12 *)
  0xba0d01ad;       (* arm_ADCS X13 X13 X13 *)
  0xba0e01ce;       (* arm_ADCS X14 X14 X14 *)
  0xba0f01ef;       (* arm_ADCS X15 X15 X15 *)
  0xba100210;       (* arm_ADCS X16 X16 X16 *)
  0xba110231;       (* arm_ADCS X17 X17 X17 *)
  0xba130273;       (* arm_ADCS X19 X19 X19 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc27c48;       (* arm_UMULH X8 X2 X2 *)
  0x9b027c42;       (* arm_MUL X2 X2 X2 *)
  0xab080129;       (* arm_ADDS X9 X9 X8 *)
  0x9b037c68;       (* arm_MUL X8 X3 X3 *)
  0xba08014a;       (* arm_ADCS X10 X10 X8 *)
  0x9bc37c68;       (* arm_UMULH X8 X3 X3 *)
  0xba08016b;       (* arm_ADCS X11 X11 X8 *)
  0x9b047c88;       (* arm_MUL X8 X4 X4 *)
  0xba08018c;       (* arm_ADCS X12 X12 X8 *)
  0x9bc47c88;       (* arm_UMULH X8 X4 X4 *)
  0xba0801ad;       (* arm_ADCS X13 X13 X8 *)
  0x9b057ca8;       (* arm_MUL X8 X5 X5 *)
  0xba0801ce;       (* arm_ADCS X14 X14 X8 *)
  0x9bc57ca8;       (* arm_UMULH X8 X5 X5 *)
  0xba0801ef;       (* arm_ADCS X15 X15 X8 *)
  0x9b067cc8;       (* arm_MUL X8 X6 X6 *)
  0xba080210;       (* arm_ADCS X16 X16 X8 *)
  0x9bc67cc8;       (* arm_UMULH X8 X6 X6 *)
  0xba080231;       (* arm_ADCS X17 X17 X8 *)
  0x9b077ce8;       (* arm_MUL X8 X7 X7 *)
  0xba080273;       (* arm_ADCS X19 X19 X8 *)
  0x9bc77ce8;       (* arm_UMULH X8 X7 X7 *)
  0x9a080294;       (* arm_ADC X20 X20 X8 *)
  0xd3607c45;       (* arm_LSL X5 X2 32 *)
  0x8b0200a2;       (* arm_ADD X2 X5 X2 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bc27ca5;       (* arm_UMULH X5 X5 X2 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b027c83;       (* arm_MUL X3 X4 X2 *)
  0x9bc27c84;       (* arm_UMULH X4 X4 X2 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba020084;       (* arm_ADCS X4 X4 X2 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb050129;       (* arm_SUBS X9 X9 X5 *)
  0xfa04014a;       (* arm_SBCS X10 X10 X4 *)
  0xfa03016b;       (* arm_SBCS X11 X11 X3 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f0042;       (* arm_SBC X2 X2 XZR *)
  0xd3607d25;       (* arm_LSL X5 X9 32 *)
  0x8b0900a9;       (* arm_ADD X9 X5 X9 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bc97ca5;       (* arm_UMULH X5 X5 X9 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b097c83;       (* arm_MUL X3 X4 X9 *)
  0x9bc97c84;       (* arm_UMULH X4 X4 X9 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba090084;       (* arm_ADCS X4 X4 X9 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05014a;       (* arm_SUBS X10 X10 X5 *)
  0xfa04016b;       (* arm_SBCS X11 X11 X4 *)
  0xfa03018c;       (* arm_SBCS X12 X12 X3 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xda1f0129;       (* arm_SBC X9 X9 XZR *)
  0xd3607d45;       (* arm_LSL X5 X10 32 *)
  0x8b0a00aa;       (* arm_ADD X10 X5 X10 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bca7ca5;       (* arm_UMULH X5 X5 X10 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0a7c83;       (* arm_MUL X3 X4 X10 *)
  0x9bca7c84;       (* arm_UMULH X4 X4 X10 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0a0084;       (* arm_ADCS X4 X4 X10 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05016b;       (* arm_SUBS X11 X11 X5 *)
  0xfa04018c;       (* arm_SBCS X12 X12 X4 *)
  0xfa0301ad;       (* arm_SBCS X13 X13 X3 *)
  0xfa1f0042;       (* arm_SBCS X2 X2 XZR *)
  0xfa1f0129;       (* arm_SBCS X9 X9 XZR *)
  0xda1f014a;       (* arm_SBC X10 X10 XZR *)
  0xd3607d65;       (* arm_LSL X5 X11 32 *)
  0x8b0b00ab;       (* arm_ADD X11 X5 X11 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcb7ca5;       (* arm_UMULH X5 X5 X11 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0b7c83;       (* arm_MUL X3 X4 X11 *)
  0x9bcb7c84;       (* arm_UMULH X4 X4 X11 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0b0084;       (* arm_ADCS X4 X4 X11 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb05018c;       (* arm_SUBS X12 X12 X5 *)
  0xfa0401ad;       (* arm_SBCS X13 X13 X4 *)
  0xfa030042;       (* arm_SBCS X2 X2 X3 *)
  0xfa1f0129;       (* arm_SBCS X9 X9 XZR *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xda1f016b;       (* arm_SBC X11 X11 XZR *)
  0xd3607d85;       (* arm_LSL X5 X12 32 *)
  0x8b0c00ac;       (* arm_ADD X12 X5 X12 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ca5;       (* arm_UMULH X5 X5 X12 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0c7c83;       (* arm_MUL X3 X4 X12 *)
  0x9bcc7c84;       (* arm_UMULH X4 X4 X12 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0c0084;       (* arm_ADCS X4 X4 X12 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb0501ad;       (* arm_SUBS X13 X13 X5 *)
  0xfa040042;       (* arm_SBCS X2 X2 X4 *)
  0xfa030129;       (* arm_SBCS X9 X9 X3 *)
  0xfa1f014a;       (* arm_SBCS X10 X10 XZR *)
  0xfa1f016b;       (* arm_SBCS X11 X11 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da5;       (* arm_LSL X5 X13 32 *)
  0x8b0d00ad;       (* arm_ADD X13 X5 X13 *)
  0xb26083e5;       (* arm_MOV X5 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ca5;       (* arm_UMULH X5 X5 X13 *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x9b0d7c83;       (* arm_MUL X3 X4 X13 *)
  0x9bcd7c84;       (* arm_UMULH X4 X4 X13 *)
  0xab0300a5;       (* arm_ADDS X5 X5 X3 *)
  0xba0d0084;       (* arm_ADCS X4 X4 X13 *)
  0x9a1f03e3;       (* arm_ADC X3 XZR XZR *)
  0xeb050042;       (* arm_SUBS X2 X2 X5 *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xfa1f016b;       (* arm_SBCS X11 X11 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xab0e0042;       (* arm_ADDS X2 X2 X14 *)
  0xba0f0129;       (* arm_ADCS X9 X9 X15 *)
  0xba10014a;       (* arm_ADCS X10 X10 X16 *)
  0xba11016b;       (* arm_ADCS X11 X11 X17 *)
  0xba13018c;       (* arm_ADCS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0x9a1f03e6;       (* arm_ADC X6 XZR XZR *)
  0xb26083e8;       (* arm_MOV X8 (rvalue (word 18446744069414584321)) *)
  0xab08004e;       (* arm_ADDS X14 X2 X8 *)
  0xb2407fe8;       (* arm_MOV X8 (rvalue (word 4294967295)) *)
  0xba08012f;       (* arm_ADCS X15 X9 X8 *)
  0xd2800028;       (* arm_MOV X8 (rvalue (word 1)) *)
  0xba080150;       (* arm_ADCS X16 X10 X8 *)
  0xba1f0171;       (* arm_ADCS X17 X11 XZR *)
  0xba1f0193;       (* arm_ADCS X19 X12 XZR *)
  0xba1f01b4;       (* arm_ADCS X20 X13 XZR *)
  0xba1f00c6;       (* arm_ADCS X6 X6 XZR *)
  0x9a8e0042;       (* arm_CSEL X2 X2 X14 Condition_EQ *)
  0x9a8f0129;       (* arm_CSEL X9 X9 X15 Condition_EQ *)
  0x9a90014a;       (* arm_CSEL X10 X10 X16 Condition_EQ *)
  0x9a91016b;       (* arm_CSEL X11 X11 X17 Condition_EQ *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0xa90027e2;       (* arm_STP X2 X9 SP (Immediate_Offset (iword (&0))) *)
  0xa9012fea;       (* arm_STP X10 X11 SP (Immediate_Offset (iword (&16))) *)
  0xa90237ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&32))) *)
  0xa94913e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&144))) *)
  0xa94c1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&192))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa94d23e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&208))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa94e2be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&224))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa94b13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&176))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90c37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&192))) *)
  0xa90d3fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&208))) *)
  0xa90e47f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&224))) *)
  0xa94913e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&144))) *)
  0xa9461be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&96))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa94723e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&112))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9482be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&128))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa94b13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&176))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90637ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&96))) *)
  0xa9073fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&112))) *)
  0xa90847f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&128))) *)
  0xa9401be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa94c0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&192))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94123e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa94d0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&208))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xa9422be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&32))) *)
  0xa94e0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&224))) *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xca030084;       (* arm_EOR X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0x92800024;       (* arm_MOVN X4 (word 1) 0 *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9a03014a;       (* arm_ADC X10 X10 X3 *)
  0xa9001be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa90123e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9022be9;       (* arm_STP X9 X10 SP (Immediate_Offset (iword (&32))) *)
  0xa9461be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&96))) *)
  0xa94c0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&192))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94723e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&112))) *)
  0xa94d0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&208))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xa9482be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&128))) *)
  0xa94e0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&224))) *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xca030084;       (* arm_EOR X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0x92800024;       (* arm_MOVN X4 (word 1) 0 *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9a03014a;       (* arm_ADC X10 X10 X3 *)
  0xa9091be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&144))) *)
  0xa90a23e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&160))) *)
  0xa90b2be9;       (* arm_STP X9 X10 SP (Immediate_Offset (iword (&176))) *)
  0xa94f13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&240))) *)
  0xa9461b25;       (* arm_LDP X5 X6 X25 (Immediate_Offset (iword (&96))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9472327;       (* arm_LDP X7 X8 X25 (Immediate_Offset (iword (&112))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9482b29;       (* arm_LDP X9 X10 X25 (Immediate_Offset (iword (&128))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa95013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&256))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa95113e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&272))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90f37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&240))) *)
  0xa9103fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&256))) *)
  0xa91147f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&272))) *)
  0xa9401be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa9460fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&96))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94123e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9470fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&112))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xa9422be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&32))) *)
  0xa9480fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&128))) *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xca030084;       (* arm_EOR X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0x92800024;       (* arm_MOVN X4 (word 1) 0 *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9a03014a;       (* arm_ADC X10 X10 X3 *)
  0xa9001be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&0))) *)
  0xa90123e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&16))) *)
  0xa9022be9;       (* arm_STP X9 X10 SP (Immediate_Offset (iword (&32))) *)
  0xa94c1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&192))) *)
  0xa9400fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&0))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94d23e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&208))) *)
  0xa9410fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&16))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xa94e2be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&224))) *)
  0xa9420fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&32))) *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xca030084;       (* arm_EOR X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0x92800024;       (* arm_MOVN X4 (word 1) 0 *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9a03014a;       (* arm_ADC X10 X10 X3 *)
  0xa90c1be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&192))) *)
  0xa90d23e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&208))) *)
  0xa90e2be9;       (* arm_STP X9 X10 SP (Immediate_Offset (iword (&224))) *)
  0xa94913e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&144))) *)
  0xa9521be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&288))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa95323e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&304))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9542be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&320))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa94a13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&160))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa94b13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&176))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90937ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&144))) *)
  0xa90a3fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&160))) *)
  0xa90b47f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&176))) *)
  0xa94f13e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&240))) *)
  0xa9461b45;       (* arm_LDP X5 X6 X26 (Immediate_Offset (iword (&96))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa9472347;       (* arm_LDP X7 X8 X26 (Immediate_Offset (iword (&112))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa9482b49;       (* arm_LDP X9 X10 X26 (Immediate_Offset (iword (&128))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa95013e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&256))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa95113e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&272))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90f37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&240))) *)
  0xa9103fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&256))) *)
  0xa91147f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&272))) *)
  0xa94313e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&48))) *)
  0xa94c1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&192))) *)
  0x9b057c6c;       (* arm_MUL X12 X3 X5 *)
  0x9bc57c6d;       (* arm_UMULH X13 X3 X5 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0x9bc67c6e;       (* arm_UMULH X14 X3 X6 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0xa94d23e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&208))) *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0x9bc77c6f;       (* arm_UMULH X15 X3 X7 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0x9bc87c70;       (* arm_UMULH X16 X3 X8 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0xa94e2be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&224))) *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0x9bc97c71;       (* arm_UMULH X17 X3 X9 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0x9bca7c73;       (* arm_UMULH X19 X3 X10 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9a1f0273;       (* arm_ADC X19 X19 XZR *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ad;       (* arm_ADDS X13 X13 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b01ce;       (* arm_ADCS X14 X14 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9a9f37f4;       (* arm_CSET X20 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0294;       (* arm_ADC X20 X20 X11 *)
  0xa94413e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&64))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b01ce;       (* arm_ADDS X14 X14 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b01ef;       (* arm_ADCS X15 X15 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9a9f37f5;       (* arm_CSET X21 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b02b5;       (* arm_ADC X21 X21 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b01ef;       (* arm_ADDS X15 X15 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0210;       (* arm_ADCS X16 X16 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9a9f37f6;       (* arm_CSET X22 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b02d6;       (* arm_ADC X22 X22 X11 *)
  0xa94513e3;       (* arm_LDP X3 X4 SP (Immediate_Offset (iword (&80))) *)
  0x9b057c6b;       (* arm_MUL X11 X3 X5 *)
  0xab0b0210;       (* arm_ADDS X16 X16 X11 *)
  0x9b067c6b;       (* arm_MUL X11 X3 X6 *)
  0xba0b0231;       (* arm_ADCS X17 X17 X11 *)
  0x9b077c6b;       (* arm_MUL X11 X3 X7 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b087c6b;       (* arm_MUL X11 X3 X8 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b097c6b;       (* arm_MUL X11 X3 X9 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b0a7c6b;       (* arm_MUL X11 X3 X10 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9a9f37e2;       (* arm_CSET X2 Condition_CS *)
  0x9bc57c6b;       (* arm_UMULH X11 X3 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9bc67c6b;       (* arm_UMULH X11 X3 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9bc77c6b;       (* arm_UMULH X11 X3 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc87c6b;       (* arm_UMULH X11 X3 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc97c6b;       (* arm_UMULH X11 X3 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bca7c6b;       (* arm_UMULH X11 X3 X10 *)
  0x9a0b0042;       (* arm_ADC X2 X2 X11 *)
  0x9b057c8b;       (* arm_MUL X11 X4 X5 *)
  0xab0b0231;       (* arm_ADDS X17 X17 X11 *)
  0x9b067c8b;       (* arm_MUL X11 X4 X6 *)
  0xba0b0273;       (* arm_ADCS X19 X19 X11 *)
  0x9b077c8b;       (* arm_MUL X11 X4 X7 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9b087c8b;       (* arm_MUL X11 X4 X8 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9b097c8b;       (* arm_MUL X11 X4 X9 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9b0a7c8b;       (* arm_MUL X11 X4 X10 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9a9f37e1;       (* arm_CSET X1 Condition_CS *)
  0x9bc57c8b;       (* arm_UMULH X11 X4 X5 *)
  0xab0b0273;       (* arm_ADDS X19 X19 X11 *)
  0x9bc67c8b;       (* arm_UMULH X11 X4 X6 *)
  0xba0b0294;       (* arm_ADCS X20 X20 X11 *)
  0x9bc77c8b;       (* arm_UMULH X11 X4 X7 *)
  0xba0b02b5;       (* arm_ADCS X21 X21 X11 *)
  0x9bc87c8b;       (* arm_UMULH X11 X4 X8 *)
  0xba0b02d6;       (* arm_ADCS X22 X22 X11 *)
  0x9bc97c8b;       (* arm_UMULH X11 X4 X9 *)
  0xba0b0042;       (* arm_ADCS X2 X2 X11 *)
  0x9bca7c8b;       (* arm_UMULH X11 X4 X10 *)
  0x9a0b0021;       (* arm_ADC X1 X1 X11 *)
  0xd3607d87;       (* arm_LSL X7 X12 32 *)
  0x8b0c00ec;       (* arm_ADD X12 X7 X12 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcc7ce7;       (* arm_UMULH X7 X7 X12 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0c7cc5;       (* arm_MUL X5 X6 X12 *)
  0x9bcc7cc6;       (* arm_UMULH X6 X6 X12 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0c00c6;       (* arm_ADCS X6 X6 X12 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ad;       (* arm_SUBS X13 X13 X7 *)
  0xfa0601ce;       (* arm_SBCS X14 X14 X6 *)
  0xfa0501ef;       (* arm_SBCS X15 X15 X5 *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xda1f018c;       (* arm_SBC X12 X12 XZR *)
  0xd3607da7;       (* arm_LSL X7 X13 32 *)
  0x8b0d00ed;       (* arm_ADD X13 X7 X13 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcd7ce7;       (* arm_UMULH X7 X7 X13 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0d7cc5;       (* arm_MUL X5 X6 X13 *)
  0x9bcd7cc6;       (* arm_UMULH X6 X6 X13 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0d00c6;       (* arm_ADCS X6 X6 X13 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ce;       (* arm_SUBS X14 X14 X7 *)
  0xfa0601ef;       (* arm_SBCS X15 X15 X6 *)
  0xfa050210;       (* arm_SBCS X16 X16 X5 *)
  0xfa1f0231;       (* arm_SBCS X17 X17 XZR *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xda1f01ad;       (* arm_SBC X13 X13 XZR *)
  0xd3607dc7;       (* arm_LSL X7 X14 32 *)
  0x8b0e00ee;       (* arm_ADD X14 X7 X14 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bce7ce7;       (* arm_UMULH X7 X7 X14 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0e7cc5;       (* arm_MUL X5 X6 X14 *)
  0x9bce7cc6;       (* arm_UMULH X6 X6 X14 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0e00c6;       (* arm_ADCS X6 X6 X14 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb0701ef;       (* arm_SUBS X15 X15 X7 *)
  0xfa060210;       (* arm_SBCS X16 X16 X6 *)
  0xfa050231;       (* arm_SBCS X17 X17 X5 *)
  0xfa1f018c;       (* arm_SBCS X12 X12 XZR *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xda1f01ce;       (* arm_SBC X14 X14 XZR *)
  0xd3607de7;       (* arm_LSL X7 X15 32 *)
  0x8b0f00ef;       (* arm_ADD X15 X7 X15 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bcf7ce7;       (* arm_UMULH X7 X7 X15 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b0f7cc5;       (* arm_MUL X5 X6 X15 *)
  0x9bcf7cc6;       (* arm_UMULH X6 X6 X15 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba0f00c6;       (* arm_ADCS X6 X6 X15 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070210;       (* arm_SUBS X16 X16 X7 *)
  0xfa060231;       (* arm_SBCS X17 X17 X6 *)
  0xfa05018c;       (* arm_SBCS X12 X12 X5 *)
  0xfa1f01ad;       (* arm_SBCS X13 X13 XZR *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xda1f01ef;       (* arm_SBC X15 X15 XZR *)
  0xd3607e07;       (* arm_LSL X7 X16 32 *)
  0x8b1000f0;       (* arm_ADD X16 X7 X16 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd07ce7;       (* arm_UMULH X7 X7 X16 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b107cc5;       (* arm_MUL X5 X6 X16 *)
  0x9bd07cc6;       (* arm_UMULH X6 X6 X16 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1000c6;       (* arm_ADCS X6 X6 X16 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb070231;       (* arm_SUBS X17 X17 X7 *)
  0xfa06018c;       (* arm_SBCS X12 X12 X6 *)
  0xfa0501ad;       (* arm_SBCS X13 X13 X5 *)
  0xfa1f01ce;       (* arm_SBCS X14 X14 XZR *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xda1f0210;       (* arm_SBC X16 X16 XZR *)
  0xd3607e27;       (* arm_LSL X7 X17 32 *)
  0x8b1100f1;       (* arm_ADD X17 X7 X17 *)
  0xb26083e7;       (* arm_MOV X7 (rvalue (word 18446744069414584321)) *)
  0x9bd17ce7;       (* arm_UMULH X7 X7 X17 *)
  0xb2407fe6;       (* arm_MOV X6 (rvalue (word 4294967295)) *)
  0x9b117cc5;       (* arm_MUL X5 X6 X17 *)
  0x9bd17cc6;       (* arm_UMULH X6 X6 X17 *)
  0xab0500e7;       (* arm_ADDS X7 X7 X5 *)
  0xba1100c6;       (* arm_ADCS X6 X6 X17 *)
  0x9a1f03e5;       (* arm_ADC X5 XZR XZR *)
  0xeb07018c;       (* arm_SUBS X12 X12 X7 *)
  0xfa0601ad;       (* arm_SBCS X13 X13 X6 *)
  0xfa0501ce;       (* arm_SBCS X14 X14 X5 *)
  0xfa1f01ef;       (* arm_SBCS X15 X15 XZR *)
  0xfa1f0210;       (* arm_SBCS X16 X16 XZR *)
  0xda1f0231;       (* arm_SBC X17 X17 XZR *)
  0xab13018c;       (* arm_ADDS X12 X12 X19 *)
  0xba1401ad;       (* arm_ADCS X13 X13 X20 *)
  0xba1501ce;       (* arm_ADCS X14 X14 X21 *)
  0xba1601ef;       (* arm_ADCS X15 X15 X22 *)
  0xba020210;       (* arm_ADCS X16 X16 X2 *)
  0xba010231;       (* arm_ADCS X17 X17 X1 *)
  0x9a1f03ea;       (* arm_ADC X10 XZR XZR *)
  0xb26083eb;       (* arm_MOV X11 (rvalue (word 18446744069414584321)) *)
  0xab0b0193;       (* arm_ADDS X19 X12 X11 *)
  0xb2407feb;       (* arm_MOV X11 (rvalue (word 4294967295)) *)
  0xba0b01b4;       (* arm_ADCS X20 X13 X11 *)
  0xd280002b;       (* arm_MOV X11 (rvalue (word 1)) *)
  0xba0b01d5;       (* arm_ADCS X21 X14 X11 *)
  0xba1f01f6;       (* arm_ADCS X22 X15 XZR *)
  0xba1f0202;       (* arm_ADCS X2 X16 XZR *)
  0xba1f0221;       (* arm_ADCS X1 X17 XZR *)
  0xba1f014a;       (* arm_ADCS X10 X10 XZR *)
  0x9a93018c;       (* arm_CSEL X12 X12 X19 Condition_EQ *)
  0x9a9401ad;       (* arm_CSEL X13 X13 X20 Condition_EQ *)
  0x9a9501ce;       (* arm_CSEL X14 X14 X21 Condition_EQ *)
  0x9a9601ef;       (* arm_CSEL X15 X15 X22 Condition_EQ *)
  0x9a820210;       (* arm_CSEL X16 X16 X2 Condition_EQ *)
  0x9a810231;       (* arm_CSEL X17 X17 X1 Condition_EQ *)
  0xa90c37ec;       (* arm_STP X12 X13 SP (Immediate_Offset (iword (&192))) *)
  0xa90d3fee;       (* arm_STP X14 X15 SP (Immediate_Offset (iword (&208))) *)
  0xa90e47f0;       (* arm_STP X16 X17 SP (Immediate_Offset (iword (&224))) *)
  0xa94c1be5;       (* arm_LDP X5 X6 SP (Immediate_Offset (iword (&192))) *)
  0xa9490fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&144))) *)
  0xeb0400a5;       (* arm_SUBS X5 X5 X4 *)
  0xfa0300c6;       (* arm_SBCS X6 X6 X3 *)
  0xa94d23e7;       (* arm_LDP X7 X8 SP (Immediate_Offset (iword (&208))) *)
  0xa94a0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&160))) *)
  0xfa0400e7;       (* arm_SBCS X7 X7 X4 *)
  0xfa030108;       (* arm_SBCS X8 X8 X3 *)
  0xa94e2be9;       (* arm_LDP X9 X10 SP (Immediate_Offset (iword (&224))) *)
  0xa94b0fe4;       (* arm_LDP X4 X3 SP (Immediate_Offset (iword (&176))) *)
  0xfa040129;       (* arm_SBCS X9 X9 X4 *)
  0xfa03014a;       (* arm_SBCS X10 X10 X3 *)
  0xda9f23e3;       (* arm_CSETM X3 Condition_CC *)
  0xb2407fe4;       (* arm_MOV X4 (rvalue (word 4294967295)) *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xab0400a5;       (* arm_ADDS X5 X5 X4 *)
  0xca030084;       (* arm_EOR X4 X4 X3 *)
  0xba0400c6;       (* arm_ADCS X6 X6 X4 *)
  0x92800024;       (* arm_MOVN X4 (word 1) 0 *)
  0x8a030084;       (* arm_AND X4 X4 X3 *)
  0xba0400e7;       (* arm_ADCS X7 X7 X4 *)
  0xba030108;       (* arm_ADCS X8 X8 X3 *)
  0xba030129;       (* arm_ADCS X9 X9 X3 *)
  0x9a03014a;       (* arm_ADC X10 X10 X3 *)
  0xa90c1be5;       (* arm_STP X5 X6 SP (Immediate_Offset (iword (&192))) *)
  0xa90d23e7;       (* arm_STP X7 X8 SP (Immediate_Offset (iword (&208))) *)
  0xa90e2be9;       (* arm_STP X9 X10 SP (Immediate_Offset (iword (&224))) *)
  0xa9460720;       (* arm_LDP X0 X1 X25 (Immediate_Offset (iword (&96))) *)
  0xa9470f22;       (* arm_LDP X2 X3 X25 (Immediate_Offset (iword (&112))) *)
  0xa9481724;       (* arm_LDP X4 X5 X25 (Immediate_Offset (iword (&128))) *)
  0xaa010014;       (* arm_ORR X20 X0 X1 *)
  0xaa030055;       (* arm_ORR X21 X2 X3 *)
  0xaa050096;       (* arm_ORR X22 X4 X5 *)
  0xaa150294;       (* arm_ORR X20 X20 X21 *)
  0xaa160294;       (* arm_ORR X20 X20 X22 *)
  0xeb1f029f;       (* arm_CMP X20 XZR *)
  0x9a9f07f4;       (* arm_CSET X20 Condition_NE *)
  0xa9461f46;       (* arm_LDP X6 X7 X26 (Immediate_Offset (iword (&96))) *)
  0xa9472748;       (* arm_LDP X8 X9 X26 (Immediate_Offset (iword (&112))) *)
  0xa9482f4a;       (* arm_LDP X10 X11 X26 (Immediate_Offset (iword (&128))) *)
  0xaa0700d5;       (* arm_ORR X21 X6 X7 *)
  0xaa090116;       (* arm_ORR X22 X8 X9 *)
  0xaa0b0157;       (* arm_ORR X23 X10 X11 *)
  0xaa1602b5;       (* arm_ORR X21 X21 X22 *)
  0xaa1702b5;       (* arm_ORR X21 X21 X23 *)
  0xeb1f02bf;       (* arm_CMP X21 XZR *)
  0x9a9f07f5;       (* arm_CSET X21 Condition_NE *)
  0xeb1402bf;       (* arm_CMP X21 X20 *)
  0xa94f37ec;       (* arm_LDP X12 X13 SP (Immediate_Offset (iword (&240))) *)
  0x9a8c300c;       (* arm_CSEL X12 X0 X12 Condition_CC *)
  0x9a8d302d;       (* arm_CSEL X13 X1 X13 Condition_CC *)
  0x9a8c80cc;       (* arm_CSEL X12 X6 X12 Condition_HI *)
  0x9a8d80ed;       (* arm_CSEL X13 X7 X13 Condition_HI *)
  0xa9503fee;       (* arm_LDP X14 X15 SP (Immediate_Offset (iword (&256))) *)
  0x9a8e304e;       (* arm_CSEL X14 X2 X14 Condition_CC *)
  0x9a8f306f;       (* arm_CSEL X15 X3 X15 Condition_CC *)
  0x9a8e810e;       (* arm_CSEL X14 X8 X14 Condition_HI *)
  0x9a8f812f;       (* arm_CSEL X15 X9 X15 Condition_HI *)
  0xa95147f0;       (* arm_LDP X16 X17 SP (Immediate_Offset (iword (&272))) *)
  0x9a903090;       (* arm_CSEL X16 X4 X16 Condition_CC *)
  0x9a9130b1;       (* arm_CSEL X17 X5 X17 Condition_CC *)
  0x9a908150;       (* arm_CSEL X16 X10 X16 Condition_HI *)
  0x9a918171;       (* arm_CSEL X17 X11 X17 Condition_HI *)
  0xa9405734;       (* arm_LDP X20 X21 X25 (Immediate_Offset (iword (&0))) *)
  0xa94007e0;       (* arm_LDP X0 X1 SP (Immediate_Offset (iword (&0))) *)
  0x9a803280;       (* arm_CSEL X0 X20 X0 Condition_CC *)
  0x9a8132a1;       (* arm_CSEL X1 X21 X1 Condition_CC *)
  0xa9405754;       (* arm_LDP X20 X21 X26 (Immediate_Offset (iword (&0))) *)
  0x9a808280;       (* arm_CSEL X0 X20 X0 Condition_HI *)
  0x9a8182a1;       (* arm_CSEL X1 X21 X1 Condition_HI *)
  0xa9415734;       (* arm_LDP X20 X21 X25 (Immediate_Offset (iword (&16))) *)
  0xa9410fe2;       (* arm_LDP X2 X3 SP (Immediate_Offset (iword (&16))) *)
  0x9a823282;       (* arm_CSEL X2 X20 X2 Condition_CC *)
  0x9a8332a3;       (* arm_CSEL X3 X21 X3 Condition_CC *)
  0xa9415754;       (* arm_LDP X20 X21 X26 (Immediate_Offset (iword (&16))) *)
  0x9a828282;       (* arm_CSEL X2 X20 X2 Condition_HI *)
  0x9a8382a3;       (* arm_CSEL X3 X21 X3 Condition_HI *)
  0xa9425734;       (* arm_LDP X20 X21 X25 (Immediate_Offset (iword (&32))) *)
  0xa94217e4;       (* arm_LDP X4 X5 SP (Immediate_Offset (iword (&32))) *)
  0x9a843284;       (* arm_CSEL X4 X20 X4 Condition_CC *)
  0x9a8532a5;       (* arm_CSEL X5 X21 X5 Condition_CC *)
  0xa9425754;       (* arm_LDP X20 X21 X26 (Immediate_Offset (iword (&32))) *)
  0x9a848284;       (* arm_CSEL X4 X20 X4 Condition_HI *)
  0x9a8582a5;       (* arm_CSEL X5 X21 X5 Condition_HI *)
  0xa9435734;       (* arm_LDP X20 X21 X25 (Immediate_Offset (iword (&48))) *)
  0xa94c1fe6;       (* arm_LDP X6 X7 SP (Immediate_Offset (iword (&192))) *)
  0x9a863286;       (* arm_CSEL X6 X20 X6 Condition_CC *)
  0x9a8732a7;       (* arm_CSEL X7 X21 X7 Condition_CC *)
  0xa9435754;       (* arm_LDP X20 X21 X26 (Immediate_Offset (iword (&48))) *)
  0x9a868286;       (* arm_CSEL X6 X20 X6 Condition_HI *)
  0x9a8782a7;       (* arm_CSEL X7 X21 X7 Condition_HI *)
  0xa9445734;       (* arm_LDP X20 X21 X25 (Immediate_Offset (iword (&64))) *)
  0xa94d27e8;       (* arm_LDP X8 X9 SP (Immediate_Offset (iword (&208))) *)
  0x9a883288;       (* arm_CSEL X8 X20 X8 Condition_CC *)
  0x9a8932a9;       (* arm_CSEL X9 X21 X9 Condition_CC *)
  0xa9445754;       (* arm_LDP X20 X21 X26 (Immediate_Offset (iword (&64))) *)
  0x9a888288;       (* arm_CSEL X8 X20 X8 Condition_HI *)
  0x9a8982a9;       (* arm_CSEL X9 X21 X9 Condition_HI *)
  0xa9455734;       (* arm_LDP X20 X21 X25 (Immediate_Offset (iword (&80))) *)
  0xa94e2fea;       (* arm_LDP X10 X11 SP (Immediate_Offset (iword (&224))) *)
  0x9a8a328a;       (* arm_CSEL X10 X20 X10 Condition_CC *)
  0x9a8b32ab;       (* arm_CSEL X11 X21 X11 Condition_CC *)
  0xa9455754;       (* arm_LDP X20 X21 X26 (Immediate_Offset (iword (&80))) *)
  0x9a8a828a;       (* arm_CSEL X10 X20 X10 Condition_HI *)
  0x9a8b82ab;       (* arm_CSEL X11 X21 X11 Condition_HI *)
  0xa9000700;       (* arm_STP X0 X1 X24 (Immediate_Offset (iword (&0))) *)
  0xa9010f02;       (* arm_STP X2 X3 X24 (Immediate_Offset (iword (&16))) *)
  0xa9021704;       (* arm_STP X4 X5 X24 (Immediate_Offset (iword (&32))) *)
  0xa9031f06;       (* arm_STP X6 X7 X24 (Immediate_Offset (iword (&48))) *)
  0xa9042708;       (* arm_STP X8 X9 X24 (Immediate_Offset (iword (&64))) *)
  0xa9052f0a;       (* arm_STP X10 X11 X24 (Immediate_Offset (iword (&80))) *)
  0xa906370c;       (* arm_STP X12 X13 X24 (Immediate_Offset (iword (&96))) *)
  0xa9073f0e;       (* arm_STP X14 X15 X24 (Immediate_Offset (iword (&112))) *)
  0xa9084710;       (* arm_STP X16 X17 X24 (Immediate_Offset (iword (&128))) *)
  0x910543ff;       (* arm_ADD SP SP (rvalue (word 336)) *)
  0xa8c16bf9;       (* arm_LDP X25 X26 SP (Postimmediate_Offset (iword (&16))) *)
  0xa8c163f7;       (* arm_LDP X23 X24 SP (Postimmediate_Offset (iword (&16))) *)
  0xa8c15bf5;       (* arm_LDP X21 X22 SP (Postimmediate_Offset (iword (&16))) *)
  0xa8c153f3;       (* arm_LDP X19 X20 SP (Postimmediate_Offset (iword (&16))) *)
  0xd65f03c0        (* arm_RET X30 *)
];;

let P384_MONTJADD_ALT_EXEC = ARM_MK_EXEC_RULE p384_montjadd_alt_mc;;

(* ------------------------------------------------------------------------- *)
(* Common supporting definitions and lemmas for component proofs.            *)
(* ------------------------------------------------------------------------- *)

let swlemma = WORD_RULE
  `word_add (word_shl x 32) x:int64 = word(4294967297 * val x)`;;

let mmlemma = prove
 (`!h (l:int64) (x:int64).
        &2 pow 64 * &h + &(val(l:int64)):real =
        &18446744069414584321 *
        &(val(word_add (word_shl x 32) x:int64))
        ==> &2 pow 64 * &h + &(val(x:int64)):real =
            &18446744069414584321 *
            &(val(word_add (word_shl x 32) x:int64))`,
  REPEAT GEN_TAC THEN REWRITE_TAC[REAL_OF_NUM_CLAUSES] THEN
  REPEAT STRIP_TAC THEN FIRST_ASSUM(SUBST1_TAC o SYM) THEN
  AP_TERM_TAC THEN AP_TERM_TAC THEN
  REWRITE_TAC[GSYM VAL_CONG; DIMINDEX_64] THEN FIRST_X_ASSUM(MATCH_MP_TAC o
   MATCH_MP (NUMBER_RULE
    `p * h + l:num = y ==> (y == x) (mod p) ==> (x == l) (mod p)`)) THEN
  REWRITE_TAC[WORD_RULE
   `word_add (word_shl x 32) x:int64 = word(4294967297 * val x)`] THEN
  REWRITE_TAC[CONG; VAL_WORD; DIMINDEX_64] THEN CONV_TAC MOD_DOWN_CONV THEN
  REWRITE_TAC[GSYM CONG] THEN MATCH_MP_TAC(NUMBER_RULE
   `(a * b == 1) (mod p) ==> (a * (b * x)  == x) (mod p)`) THEN
  REWRITE_TAC[CONG] THEN CONV_TAC NUM_REDUCE_CONV);;

let lvs =
 ["x_1",[`X25`;`0`];
  "y_1",[`X25`;`48`];
  "z_1",[`X25`;`96`];
  "x_2",[`X26`;`0`];
  "y_2",[`X26`;`48`];
  "z_2",[`X26`;`96`];
  "x_3",[`X24`;`0`];
  "y_3",[`X24`;`48`];
  "z_3",[`X24`;`96`];
  "z1sq",[`SP`;`0`];
  "ww",[`SP`;`0`];
  "resx",[`SP`;`0`];
  "yd",[`SP`;`48`];
  "y2a",[`SP`;`48`];
  "x2a",[`SP`;`96`];
  "zzx2",[`SP`;`96`];
  "zz",[`SP`;`144`];
  "t1",[`SP`;`144`];
  "t2",[`SP`;`192`];
  "x1a",[`SP`;`192`];
  "zzx1",[`SP`;`192`];
  "resy",[`SP`;`192`];
  "xd",[`SP`;`240`];
  "z2sq",[`SP`;`240`];
  "resz",[`SP`;`240`];
  "y1a",[`SP`;`288`]];;

let DESUM_RULE' = cache DESUM_RULE and DECARRY_RULE' = cache DECARRY_RULE;;

(* ------------------------------------------------------------------------- *)
(* Instances of montsqr.                                                     *)
(* ------------------------------------------------------------------------- *)

let LOCAL_MONTSQR_P384_TAC =
  ARM_MACRO_SIM_ABBREV_TAC p384_montjadd_alt_mc 215 lvs
  `!(t:armstate) pcin pcout p3 n3 p1 n1.
    !a. read(memory :> bytes(word_add (read p1 t) (word n1),8 * 6)) t = a
    ==>
    aligned 16 (read SP t) /\
    nonoverlapping (word pc,0x4480) (word_add (read p3 t) (word n3),48)
    ==> ensures arm
         (\s. aligned_bytes_loaded s (word pc) p384_montjadd_alt_mc /\
              read PC s = pcin /\
              read SP s = read SP t /\
              read X24 s = read X24 t /\
              read X25 s = read X25 t /\
              read X26 s = read X26 t /\
              read(memory :> bytes(word_add (read p1 t) (word n1),8 * 6)) s =
              a)
             (\s. read PC s = pcout /\
                  (a EXP 2 <= 2 EXP 384 * p_384
                   ==> read(memory :> bytes(word_add (read p3 t) (word n3),
                        8 * 6)) s =
                       (inverse_mod p_384 (2 EXP 384) * a EXP 2) MOD p_384))
           (MAYCHANGE [PC; X2; X3; X4; X5; X6; X7; X8; X9; X10;
                       X11; X12; X13; X14; X15; X16; X17; X19; X20] ,,
              MAYCHANGE
               [memory :> bytes(word_add (read p3 t) (word n3),8 * 6)] ,,
              MAYCHANGE SOME_FLAGS)`
 (REWRITE_TAC[C_ARGUMENTS; C_RETURN; SOME_FLAGS; NONOVERLAPPING_CLAUSES] THEN
  DISCH_THEN(REPEAT_TCL CONJUNCTS_THEN ASSUME_TAC) THEN

  (*** Globalize the a EXP 2 <= 2 EXP 384 * p_384  assumption ***)

  ASM_CASES_TAC `a EXP 2 <= 2 EXP 384 * p_384` THENL
   [ASM_REWRITE_TAC[]; ARM_SIM_TAC P384_MONTJADD_ALT_EXEC (1--215)] THEN
  ENSURES_INIT_TAC "s0" THEN
  FIRST_ASSUM(BIGNUM_DIGITIZE_TAC "x_" o lhand o concl) THEN

  (*** Main squaring block ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC (1--93) (1--93) THEN

  (*** Main Montgomery reduction block ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC
   (allpairs (fun i j -> 16 * i + j) (0--5)
             [97;99;101;102;104;105;106;107;108;109] @
    (190--196))
   (94--196) THEN
  RULE_ASSUM_TAC(REWRITE_RULE[ADD_CLAUSES; COND_SWAP; GSYM WORD_BITVAL]) THEN
  RULE_ASSUM_TAC(fun th -> try MATCH_MP mmlemma th with Failure _ -> th) THEN

  (*** Key properties of pre-reduced result ***)

  ABBREV_TAC
   `t = bignum_of_wordlist
         [sum_s190; sum_s191; sum_s192; sum_s193; sum_s194; sum_s195;
          word(bitval carry_s195)]` THEN
  SUBGOAL_THEN
   `t < 2 * p_384 /\ (2 EXP 384 * t == a EXP 2) (mod p_384)`
  STRIP_ASSUME_TAC THENL
   [RULE_ASSUM_TAC(REWRITE_RULE[VAL_WORD_BITVAL]) THEN
    ACCUMULATOR_POP_ASSUM_LIST
     (STRIP_ASSUME_TAC o end_itlist CONJ o DECARRY_RULE') THEN
    CONJ_TAC THENL
     [FIRST_X_ASSUM(MATCH_MP_TAC o MATCH_MP (ARITH_RULE
        `ab <= 2 EXP 384 * p
         ==> 2 EXP 384 * t < ab + 2 EXP 384 * p ==> t < 2 * p`)) THEN
      MAP_EVERY EXPAND_TAC ["a"; "t"] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; bignum_of_wordlist] THEN
      REWRITE_TAC[p_384; REAL_ARITH `a:real < b + c <=> a - b < c`] THEN
      ASM_REWRITE_TAC[VAL_WORD_BITVAL] THEN BOUNDER_TAC[];
      REWRITE_TAC[REAL_CONGRUENCE; p_384] THEN CONV_TAC NUM_REDUCE_CONV THEN
      MAP_EVERY EXPAND_TAC ["a"; "t"] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; bignum_of_wordlist] THEN
      ASM_REWRITE_TAC[VAL_WORD_BITVAL] THEN REAL_INTEGER_TAC];
    ACCUMULATOR_POP_ASSUM_LIST(K ALL_TAC)] THEN

 (*** Final correction stage ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC
   [198;200;202;203;204;205;206] (197--215) THEN
  RULE_ASSUM_TAC(REWRITE_RULE
   [GSYM WORD_BITVAL; COND_SWAP; REAL_BITVAL_NOT]) THEN
  ENSURES_FINAL_STATE_TAC THEN ASM_REWRITE_TAC[] THEN
  CONV_TAC(LAND_CONV BIGNUM_EXPAND_CONV) THEN ASM_REWRITE_TAC[] THEN
  TRANS_TAC EQ_TRANS `t MOD p_384` THEN CONJ_TAC THENL
   [ALL_TAC;
    REWRITE_TAC[GSYM CONG] THEN FIRST_X_ASSUM(MATCH_MP_TAC o MATCH_MP
     (NUMBER_RULE
       `(e * t == a EXP 2) (mod p)
        ==> (e * i == 1) (mod p) ==> (t == i * a EXP 2) (mod p)`)) THEN
    REWRITE_TAC[INVERSE_MOD_RMUL_EQ; COPRIME_REXP; COPRIME_2] THEN
    REWRITE_TAC[p_384] THEN CONV_TAC NUM_REDUCE_CONV] THEN
  CONV_TAC SYM_CONV THEN MATCH_MP_TAC EQUAL_FROM_CONGRUENT_MOD_MOD THEN
  MAP_EVERY EXISTS_TAC
   [`384`; `if t < p_384 then &t:real else &t - &p_384`] THEN
  REPEAT CONJ_TAC THENL
   [REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES] THEN BOUNDER_TAC[];
    REWRITE_TAC[p_384] THEN ARITH_TAC;
    REWRITE_TAC[p_384] THEN ARITH_TAC;
    ALL_TAC;
    ASM_SIMP_TAC[MOD_CASES] THEN
    GEN_REWRITE_TAC LAND_CONV [COND_RAND] THEN
    SIMP_TAC[REAL_OF_NUM_SUB; GSYM NOT_LT]] THEN
  SUBGOAL_THEN
   `val(word_add (word(bitval carry_s195))
                 (word(bitval carry_s205)):int64) = 0 <=>
    t < p_384`
  SUBST_ALL_TAC THENL
   [REWRITE_TAC[VAL_WORD_ADD_CASES; VAL_WORD_BITVAL; DIMINDEX_64] THEN
    SIMP_TAC[ADD_EQ_0; BITVAL_EQ_0; BITVAL_BOUND; ARITH_RULE
     `a <= 1 /\ b <= 1 ==> a + b < 2 EXP 64`] THEN
    EXPAND_TAC "t" THEN REWRITE_TAC[bignum_of_wordlist; VAL_WORD_BITVAL] THEN
    ASM_CASES_TAC `carry_s195:bool` THEN ASM_REWRITE_TAC[BITVAL_CLAUSES] THENL
     [REWRITE_TAC[p_384] THEN ARITH_TAC; ALL_TAC] THEN
    REWRITE_TAC[ADD_CLAUSES; MULT_CLAUSES; GSYM NOT_LE] THEN AP_TERM_TAC THEN
    MATCH_MP_TAC FLAG_FROM_CARRY_LE THEN EXISTS_TAC `384` THEN
    REWRITE_TAC[bignum_of_wordlist; MULT_CLAUSES; ADD_CLAUSES] THEN
    REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; p_384] THEN
    ACCUMULATOR_ASSUM_LIST(MP_TAC o end_itlist CONJ o DECARRY_RULE') THEN
    DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN BOUNDER_TAC[];
    ALL_TAC] THEN
  COND_CASES_TAC THEN ASM_REWRITE_TAC[] THEN EXPAND_TAC "t" THEN
  REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; bignum_of_wordlist; p_384] THEN
  ACCUMULATOR_POP_ASSUM_LIST(MP_TAC o end_itlist CONJ o DESUM_RULE') THEN
  ASM_REWRITE_TAC[BITVAL_CLAUSES; VAL_WORD_BITVAL] THEN
  DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN REAL_INTEGER_TAC);;

(* ------------------------------------------------------------------------- *)
(* Instances of montmul.                                                     *)
(* ------------------------------------------------------------------------- *)

let LOCAL_MONTMUL_P384_TAC =
  ARM_MACRO_SIM_ABBREV_TAC p384_montjadd_alt_mc 271 lvs
  `!(t:armstate) pcin pcout p3 n3 p1 n1 p2 n2.
    !a. read(memory :> bytes(word_add (read p1 t) (word n1),8 * 6)) t = a
    ==>
    !b. read(memory :> bytes(word_add (read p2 t) (word n2),8 * 6)) t = b
    ==>
    aligned 16 (read SP t) /\
    nonoverlapping (word pc,0x4480) (word_add (read p3 t) (word n3),48)
    ==> ensures arm
         (\s. aligned_bytes_loaded s (word pc) p384_montjadd_alt_mc /\
              read PC s = pcin /\
              read SP s = read SP t /\
              read X24 s = read X24 t /\
              read X25 s = read X25 t /\
              read X26 s = read X26 t /\
              read(memory :> bytes(word_add (read p1 t) (word n1),8 * 6)) s =
              a /\
              read(memory :> bytes(word_add (read p2 t) (word n2),8 * 6)) s =
              b)
             (\s. read PC s = pcout /\
                  (a * b <= 2 EXP 384 * p_384
                   ==> read(memory :> bytes(word_add (read p3 t) (word n3),
                            8 * 6)) s =
                       (inverse_mod p_384 (2 EXP 384) * a * b) MOD p_384))
            (MAYCHANGE [PC; X1; X2; X3; X4; X5; X6; X7; X8; X9;
                         X10; X11; X12; X13; X14; X15; X16; X17; X19;
                         X20; X21; X22] ,,
              MAYCHANGE
               [memory :> bytes(word_add (read p3 t) (word n3),8 * 6)] ,,
              MAYCHANGE SOME_FLAGS)`
 (REWRITE_TAC[C_ARGUMENTS; C_RETURN; SOME_FLAGS; NONOVERLAPPING_CLAUSES] THEN
  DISCH_THEN(REPEAT_TCL CONJUNCTS_THEN ASSUME_TAC) THEN

  (*** Globalize the a * b <= 2 EXP 384 * p_384  assumption ***)

  ASM_CASES_TAC `a * b <= 2 EXP 384 * p_384` THENL
   [ASM_REWRITE_TAC[]; ARM_SIM_TAC P384_MONTJADD_ALT_EXEC (1--271)] THEN
  ENSURES_INIT_TAC "s0" THEN

  FIRST_ASSUM(BIGNUM_DIGITIZE_TAC "y_" o lhand o concl) THEN
  FIRST_ASSUM(BIGNUM_DIGITIZE_TAC "x_" o lhand o concl) THEN

  (*** Main multiplication block ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC (1--149) (1--149) THEN

  (*** Main Montgomery reduction block ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC
   (allpairs (fun i j -> 16 * i + j) (0--5)
             [153;155;157;158;160;161;162;163;164;165] @
    (246--252))
   (150--252) THEN
  RULE_ASSUM_TAC(REWRITE_RULE[ADD_CLAUSES; COND_SWAP; GSYM WORD_BITVAL]) THEN
  RULE_ASSUM_TAC(fun th -> try MATCH_MP mmlemma th with Failure _ -> th) THEN

  (*** Key properties of pre-reduced result ***)

  ABBREV_TAC
   `t = bignum_of_wordlist
         [sum_s246; sum_s247; sum_s248; sum_s249; sum_s250; sum_s251;
          word(bitval carry_s251)]` THEN
  SUBGOAL_THEN
   `t < 2 * p_384 /\ (2 EXP 384 * t == a * b) (mod p_384)`
  STRIP_ASSUME_TAC THENL
   [RULE_ASSUM_TAC(REWRITE_RULE[VAL_WORD_BITVAL]) THEN
    ACCUMULATOR_POP_ASSUM_LIST
     (STRIP_ASSUME_TAC o end_itlist CONJ o DECARRY_RULE') THEN
    CONJ_TAC THENL
     [FIRST_X_ASSUM(MATCH_MP_TAC o MATCH_MP (ARITH_RULE
        `ab <= 2 EXP 384 * p
         ==> 2 EXP 384 * t < ab + 2 EXP 384 * p ==> t < 2 * p`)) THEN
      MAP_EVERY EXPAND_TAC ["a"; "b"; "t"] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; bignum_of_wordlist] THEN
      REWRITE_TAC[p_384; REAL_ARITH `a:real < b + c <=> a - b < c`] THEN
      ASM_REWRITE_TAC[VAL_WORD_BITVAL] THEN BOUNDER_TAC[];
      REWRITE_TAC[REAL_CONGRUENCE; p_384] THEN CONV_TAC NUM_REDUCE_CONV THEN
      MAP_EVERY EXPAND_TAC ["a"; "b"; "t"] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; bignum_of_wordlist] THEN
      ASM_REWRITE_TAC[VAL_WORD_BITVAL] THEN REAL_INTEGER_TAC];
    ACCUMULATOR_POP_ASSUM_LIST(K ALL_TAC)] THEN

 (*** Final correction stage ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC
   [254;256;258;259;260;261;262] (253--271) THEN
  RULE_ASSUM_TAC(REWRITE_RULE
   [GSYM WORD_BITVAL; COND_SWAP; REAL_BITVAL_NOT]) THEN
  ENSURES_FINAL_STATE_TAC THEN ASM_REWRITE_TAC[] THEN
  CONV_TAC(LAND_CONV BIGNUM_EXPAND_CONV) THEN ASM_REWRITE_TAC[] THEN
  TRANS_TAC EQ_TRANS `t MOD p_384` THEN CONJ_TAC THENL
   [ALL_TAC;
    REWRITE_TAC[GSYM CONG] THEN FIRST_X_ASSUM(MATCH_MP_TAC o MATCH_MP
     (NUMBER_RULE
       `(e * t == a * b) (mod p)
        ==> (e * i == 1) (mod p) ==> (t == i * a * b) (mod p)`)) THEN
    REWRITE_TAC[INVERSE_MOD_RMUL_EQ; COPRIME_REXP; COPRIME_2] THEN
    REWRITE_TAC[p_384] THEN CONV_TAC NUM_REDUCE_CONV] THEN
  CONV_TAC SYM_CONV THEN MATCH_MP_TAC EQUAL_FROM_CONGRUENT_MOD_MOD THEN
  MAP_EVERY EXISTS_TAC
   [`384`; `if t < p_384 then &t:real else &t - &p_384`] THEN
  REPEAT CONJ_TAC THENL
   [REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES] THEN BOUNDER_TAC[];
    REWRITE_TAC[p_384] THEN ARITH_TAC;
    REWRITE_TAC[p_384] THEN ARITH_TAC;
    ALL_TAC;
    ASM_SIMP_TAC[MOD_CASES] THEN
    GEN_REWRITE_TAC LAND_CONV [COND_RAND] THEN
    SIMP_TAC[REAL_OF_NUM_SUB; GSYM NOT_LT]] THEN
  SUBGOAL_THEN
   `val(word_add (word(bitval carry_s251))
                 (word(bitval carry_s261)):int64) = 0 <=>
    t < p_384`
  SUBST_ALL_TAC THENL
   [REWRITE_TAC[VAL_WORD_ADD_CASES; VAL_WORD_BITVAL; DIMINDEX_64] THEN
    SIMP_TAC[ADD_EQ_0; BITVAL_EQ_0; BITVAL_BOUND; ARITH_RULE
     `a <= 1 /\ b <= 1 ==> a + b < 2 EXP 64`] THEN
    EXPAND_TAC "t" THEN REWRITE_TAC[bignum_of_wordlist; VAL_WORD_BITVAL] THEN
    ASM_CASES_TAC `carry_s251:bool` THEN ASM_REWRITE_TAC[BITVAL_CLAUSES] THENL
     [REWRITE_TAC[p_384] THEN ARITH_TAC; ALL_TAC] THEN
    REWRITE_TAC[ADD_CLAUSES; MULT_CLAUSES; GSYM NOT_LE] THEN AP_TERM_TAC THEN
    MATCH_MP_TAC FLAG_FROM_CARRY_LE THEN EXISTS_TAC `384` THEN
    REWRITE_TAC[bignum_of_wordlist; MULT_CLAUSES; ADD_CLAUSES] THEN
    REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; p_384] THEN
    ACCUMULATOR_ASSUM_LIST(MP_TAC o end_itlist CONJ o DECARRY_RULE') THEN
    DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN BOUNDER_TAC[];
    ALL_TAC] THEN
  COND_CASES_TAC THEN ASM_REWRITE_TAC[] THEN EXPAND_TAC "t" THEN
  REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; bignum_of_wordlist; p_384] THEN
  ACCUMULATOR_POP_ASSUM_LIST(MP_TAC o end_itlist CONJ o DESUM_RULE') THEN
  ASM_REWRITE_TAC[BITVAL_CLAUSES; VAL_WORD_BITVAL] THEN
  DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN REAL_INTEGER_TAC);;

(* ------------------------------------------------------------------------- *)
(* Instances of sub.                                                         *)
(* ------------------------------------------------------------------------- *)

let LOCAL_SUB_P384_TAC =
  ARM_MACRO_SIM_ABBREV_TAC p384_montjadd_alt_mc 27 lvs
  `!(t:armstate) pcin pcout p3 n3 p1 n1 p2 n2.
    !m. read(memory :> bytes(word_add (read p1 t) (word n1),8 * 6)) t = m
    ==>
    !n. read(memory :> bytes(word_add (read p2 t) (word n2),8 * 6)) t = n
    ==>
    aligned 16 (read SP t) /\
    nonoverlapping (word pc,0x4480) (word_add (read p3 t) (word n3),48)
    ==> ensures arm
         (\s. aligned_bytes_loaded s (word pc) p384_montjadd_alt_mc /\
              read PC s = pcin /\
              read SP s = read SP t /\
              read X24 s = read X24 t /\
              read X25 s = read X25 t /\
              read X26 s = read X26 t /\
              read(memory :> bytes(word_add (read p1 t) (word n1),8 * 6)) s =
              m /\
              read(memory :> bytes(word_add (read p2 t) (word n2),8 * 6)) s =
              n)
             (\s. read PC s = pcout /\
                  (m < p_384 /\ n < p_384
                   ==> &(read(memory :> bytes(word_add (read p3 t) (word n3),
                            8 * 6)) s) = (&m - &n) rem &p_384))
            (MAYCHANGE [PC; X3; X4; X5; X6; X7; X8; X9; X10] ,,
             MAYCHANGE
               [memory :> bytes(word_add (read p3 t) (word n3),8 * 6)] ,,
              MAYCHANGE SOME_FLAGS)`
 (REWRITE_TAC[C_ARGUMENTS; C_RETURN; SOME_FLAGS; NONOVERLAPPING_CLAUSES] THEN
  DISCH_THEN(REPEAT_TCL CONJUNCTS_THEN ASSUME_TAC) THEN

  REWRITE_TAC[BIGNUM_FROM_MEMORY_BYTES] THEN ENSURES_INIT_TAC "s0" THEN
  FIRST_ASSUM(BIGNUM_DIGITIZE_TAC "n_" o lhand o concl) THEN
  FIRST_ASSUM(BIGNUM_DIGITIZE_TAC "m_" o lhand o concl) THEN

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC (1--12) (1--12) THEN

  SUBGOAL_THEN `carry_s12 <=> m < n` SUBST_ALL_TAC THENL
   [MATCH_MP_TAC FLAG_FROM_CARRY_LT THEN EXISTS_TAC `384` THEN
    MAP_EVERY EXPAND_TAC ["m"; "n"] THEN REWRITE_TAC[GSYM REAL_OF_NUM_ADD] THEN
    REWRITE_TAC[GSYM REAL_OF_NUM_MUL; GSYM REAL_OF_NUM_POW] THEN
    ACCUMULATOR_ASSUM_LIST(MP_TAC o end_itlist CONJ o DECARRY_RULE') THEN
    DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN BOUNDER_TAC[];
    ALL_TAC] THEN

  ARM_STEPS_TAC P384_MONTJADD_ALT_EXEC [13] THEN
  RULE_ASSUM_TAC(REWRITE_RULE[WORD_UNMASK_64; NOT_LE]) THEN

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC (14--27) (14--27) THEN

  ENSURES_FINAL_STATE_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN
  CONV_TAC(LAND_CONV(RAND_CONV BIGNUM_EXPAND_CONV)) THEN
  ASM_REWRITE_TAC[] THEN DISCARD_STATE_TAC "s27" THEN

  CONV_TAC SYM_CONV THEN MATCH_MP_TAC INT_REM_UNIQ THEN
  EXISTS_TAC `--(&(bitval(m < n))):int` THEN REWRITE_TAC[INT_ABS_NUM] THEN
  REWRITE_TAC[INT_ARITH `m - n:int = --b * p + z <=> z = b * p + m - n`] THEN
  REWRITE_TAC[int_eq; int_le; int_lt] THEN
  REWRITE_TAC[int_add_th; int_mul_th; int_of_num_th; int_sub_th] THEN
  REWRITE_TAC[GSYM REAL_OF_NUM_ADD; GSYM REAL_OF_NUM_MUL;
              GSYM REAL_OF_NUM_POW] THEN
  MATCH_MP_TAC(REAL_ARITH
  `!t:real. p < t /\
            (&0 <= a /\ a < p) /\
            (&0 <= z /\ z < t) /\
            ((&0 <= z /\ z < t) /\ (&0 <= a /\ a < t) ==> z = a)
            ==> z = a /\ &0 <= z /\ z < p`) THEN
  EXISTS_TAC `(&2:real) pow 384` THEN

  CONJ_TAC THENL [REWRITE_TAC[p_384] THEN REAL_ARITH_TAC; ALL_TAC] THEN
  CONJ_TAC THENL
   [MAP_EVERY UNDISCH_TAC [`m < p_384`; `n < p_384`] THEN
    REWRITE_TAC[GSYM REAL_OF_NUM_LT] THEN
    ASM_CASES_TAC `&m:real < &n` THEN ASM_REWRITE_TAC[BITVAL_CLAUSES] THEN
    POP_ASSUM MP_TAC THEN REWRITE_TAC[p_384] THEN REAL_ARITH_TAC;
    ALL_TAC] THEN
  CONJ_TAC THENL [BOUNDER_TAC[]; STRIP_TAC] THEN

  MATCH_MP_TAC EQUAL_FROM_CONGRUENT_REAL THEN
  MAP_EVERY EXISTS_TAC [`384`; `&0:real`] THEN
  ASM_REWRITE_TAC[] THEN
  CONJ_TAC THENL [REAL_INTEGER_TAC; ALL_TAC] THEN
  ACCUMULATOR_POP_ASSUM_LIST(MP_TAC o end_itlist CONJ o DESUM_RULE') THEN
  REWRITE_TAC[WORD_AND_MASK] THEN
  COND_CASES_TAC THEN ASM_REWRITE_TAC[BITVAL_CLAUSES] THEN
  CONV_TAC WORD_REDUCE_CONV THEN
  MAP_EVERY EXPAND_TAC ["m"; "n"] THEN REWRITE_TAC[GSYM REAL_OF_NUM_ADD] THEN
  REWRITE_TAC[GSYM REAL_OF_NUM_MUL; GSYM REAL_OF_NUM_POW; p_384] THEN
  DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN
  CONV_TAC(RAND_CONV REAL_POLY_CONV) THEN REAL_INTEGER_TAC);;

(* ------------------------------------------------------------------------- *)
(* Instances of amontsqr.                                                    *)
(* ------------------------------------------------------------------------- *)

let LOCAL_AMONTSQR_P384_TAC =
  ARM_MACRO_SIM_ABBREV_TAC p384_montjadd_alt_mc 209 lvs
  `!(t:armstate) pcin pcout p3 n3 p1 n1.
    !a. read(memory :> bytes(word_add (read p1 t) (word n1),8 * 6)) t = a
    ==>
    aligned 16 (read SP t) /\
    nonoverlapping (word pc,0x4480) (word_add (read p3 t) (word n3),48)
    ==> ensures arm
         (\s. aligned_bytes_loaded s (word pc) p384_montjadd_alt_mc /\
              read PC s = pcin /\
              read SP s = read SP t /\
              read X24 s = read X24 t /\
              read X25 s = read X25 t /\
              read X26 s = read X26 t /\
              read(memory :> bytes(word_add (read p1 t) (word n1),8 * 6)) s =
              a)
         (\s. read PC s = pcout /\
              read(memory :> bytes(word_add (read p3 t) (word n3),8 * 6)) s
              < 2 EXP 384 /\
              (read(memory :> bytes(word_add (read p3 t) (word n3),8 * 6)) s ==
               inverse_mod p_384 (2 EXP 384) * a EXP 2) (mod p_384))
             (MAYCHANGE [PC; X2; X3; X4; X5; X6; X7; X8; X9; X10;
                         X11; X12; X13; X14; X15; X16; X17; X19; X20] ,,
              MAYCHANGE
               [memory :> bytes(word_add (read p3 t) (word n3),8 * 6)] ,,
              MAYCHANGE SOME_FLAGS)`
 (REWRITE_TAC[C_ARGUMENTS; C_RETURN; SOME_FLAGS; NONOVERLAPPING_CLAUSES] THEN
  DISCH_THEN(REPEAT_TCL CONJUNCTS_THEN ASSUME_TAC) THEN
  ENSURES_INIT_TAC "s0" THEN
  FIRST_ASSUM(BIGNUM_DIGITIZE_TAC "x_" o lhand o concl) THEN

  (*** Main squaring block ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC (1--93) (1--93) THEN

  (*** Main Montgomery reduction block ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC
   (allpairs (fun i j -> 16 * i + j) (0--5)
             [97;99;101;102;104;105;106;107;108;109] @
    (190--196))
   (94--196) THEN
  RULE_ASSUM_TAC(REWRITE_RULE[ADD_CLAUSES; COND_SWAP; GSYM WORD_BITVAL]) THEN
  RULE_ASSUM_TAC(fun th -> try MATCH_MP mmlemma th with Failure _ -> th) THEN

  (*** Key properties of pre-reduced result ***)

  ABBREV_TAC
   `t = bignum_of_wordlist
         [sum_s190; sum_s191; sum_s192; sum_s193; sum_s194; sum_s195;
          word(bitval carry_s195)]` THEN
  SUBGOAL_THEN
   `t < 2 EXP 384 + p_384 /\ (2 EXP 384 * t == a EXP 2) (mod p_384)`
  STRIP_ASSUME_TAC THENL
   [RULE_ASSUM_TAC(REWRITE_RULE[VAL_WORD_BITVAL]) THEN
    ACCUMULATOR_POP_ASSUM_LIST
     (STRIP_ASSUME_TAC o end_itlist CONJ o DECARRY_RULE') THEN
    CONJ_TAC THENL
     [MATCH_MP_TAC(ARITH_RULE
       `2 EXP 384 * t <= (2 EXP 384 - 1) EXP 2 + (2 EXP 384 - 1) * p
        ==> t < 2 EXP 384 + p`) THEN
      REWRITE_TAC[p_384] THEN CONV_TAC NUM_REDUCE_CONV THEN
      MAP_EVERY EXPAND_TAC ["a"; "t"] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; bignum_of_wordlist] THEN
      REWRITE_TAC[p_384; REAL_ARITH `a:real < b + c <=> a - b < c`] THEN
      ASM_REWRITE_TAC[VAL_WORD_BITVAL] THEN BOUNDER_TAC[];
      REWRITE_TAC[REAL_CONGRUENCE; p_384] THEN CONV_TAC NUM_REDUCE_CONV THEN
      MAP_EVERY EXPAND_TAC ["a"; "t"] THEN
      REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES; bignum_of_wordlist] THEN
      ASM_REWRITE_TAC[VAL_WORD_BITVAL] THEN REAL_INTEGER_TAC];
    ACCUMULATOR_POP_ASSUM_LIST(K ALL_TAC)] THEN

  (*** Final correction stage ***)

  ARM_ACCSTEPS_TAC P384_MONTJADD_ALT_EXEC (201--206) (197--209) THEN
  ENSURES_FINAL_STATE_TAC THEN ASM_REWRITE_TAC[] THEN
  CONJ_TAC THENL
   [CONV_TAC(ONCE_DEPTH_CONV BIGNUM_EXPAND_CONV) THEN
    ASM_REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES] THEN BOUNDER_TAC[];
    ALL_TAC] THEN

  FIRST_X_ASSUM(MATCH_MP_TAC o MATCH_MP
     (NUMBER_RULE
       `(e * t == ab) (mod p)
        ==> (e * i == 1) (mod p) /\ (s == t) (mod p)
            ==> (s == i * ab) (mod p)`)) THEN
  REWRITE_TAC[INVERSE_MOD_RMUL_EQ; COPRIME_REXP; COPRIME_2] THEN
  CONJ_TAC THENL
   [REWRITE_TAC[p_384] THEN CONV_TAC NUM_REDUCE_CONV; ALL_TAC] THEN
  SUBGOAL_THEN `carry_s195 <=> 2 EXP 384 <= t` SUBST_ALL_TAC THENL
   [MATCH_MP_TAC FLAG_FROM_CARRY_LE THEN EXISTS_TAC `384` THEN
    EXPAND_TAC "t" THEN
    REWRITE_TAC[p_384; bignum_of_wordlist; GSYM REAL_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[VAL_WORD_BITVAL] THEN BOUNDER_TAC[];
    ABBREV_TAC `b <=> 2 EXP 384 <= t`] THEN
  MATCH_MP_TAC(NUMBER_RULE `!b:num. x + b * p = y ==> (x == y) (mod p)`) THEN
  EXISTS_TAC `bitval b` THEN REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES] THEN
  ONCE_REWRITE_TAC[REAL_ARITH `a + b:real = c <=> c - b = a`] THEN
  MATCH_MP_TAC EQUAL_FROM_CONGRUENT_REAL THEN
  MAP_EVERY EXISTS_TAC [`384`; `&0:real`] THEN CONJ_TAC THENL
   [EXPAND_TAC "b" THEN UNDISCH_TAC `t < 2 EXP 384 + p_384` THEN
    REWRITE_TAC[bitval; p_384; GSYM REAL_OF_NUM_CLAUSES] THEN REAL_ARITH_TAC;
    REWRITE_TAC[INTEGER_CLOSED]] THEN
  CONJ_TAC THENL
   [CONV_TAC(ONCE_DEPTH_CONV BIGNUM_EXPAND_CONV) THEN
    REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES] THEN BOUNDER_TAC[];
    ALL_TAC] THEN
  CONV_TAC(ONCE_DEPTH_CONV BIGNUM_EXPAND_CONV) THEN
  EXPAND_TAC "t" THEN REWRITE_TAC[bignum_of_wordlist] THEN
  ASM_REWRITE_TAC[GSYM REAL_OF_NUM_CLAUSES] THEN
  ACCUMULATOR_POP_ASSUM_LIST (MP_TAC o end_itlist CONJ o DESUM_RULE') THEN
  DISCH_THEN(fun th -> REWRITE_TAC[th]) THEN
  BOOL_CASES_TAC `b:bool` THEN REWRITE_TAC[BITVAL_CLAUSES; p_384] THEN
  CONV_TAC WORD_REDUCE_CONV THEN REAL_INTEGER_TAC);;

(* ------------------------------------------------------------------------- *)
(* Overall point operation proof.                                            *)
(* ------------------------------------------------------------------------- *)

let unilemma0 = prove
 (`x = a MOD p_384 ==> x < p_384 /\ &x = &a rem &p_384`,
  REWRITE_TAC[INT_OF_NUM_REM; p_384] THEN ARITH_TAC);;

let unilemma1 = prove
 (`&x = a rem &p_384 ==> x < p_384 /\ &x = a rem &p_384`,
  SIMP_TAC[GSYM INT_OF_NUM_LT; INT_LT_REM_EQ; p_384] THEN INT_ARITH_TAC);;

let lemont = prove
 (`(&i * x * y) rem &p_384 = (&i * x rem &p_384 * y rem &p_384) rem &p_384`,
  CONV_TAC INT_REM_DOWN_CONV THEN REWRITE_TAC[]);;

let demont = prove
 (`(&(NUMERAL n) * &x) rem &p_384 = (&(NUMERAL n) * &x rem &p_384) rem &p_384`,
  CONV_TAC INT_REM_DOWN_CONV THEN REWRITE_TAC[]);;

let pumont = prove
 (`(&(inverse_mod p_384 (2 EXP 384)) *
    (&2 pow 384 * x) rem &p_384 * (&2 pow 384 * y) rem &p_384) rem &p_384 =
   (&2 pow 384 * x * y) rem &p_384`,
  CONV_TAC INT_REM_DOWN_CONV THEN REWRITE_TAC[INT_REM_EQ] THEN
  MATCH_MP_TAC(INTEGER_RULE
   `(i * t:int == &1) (mod p)
    ==> (i * (t * x) * (t * y) == t * x * y) (mod p)`) THEN
  REWRITE_TAC[GSYM num_congruent; INT_OF_NUM_CLAUSES] THEN
  REWRITE_TAC[INVERSE_MOD_LMUL_EQ; COPRIME_REXP; COPRIME_2; p_384] THEN
  CONV_TAC NUM_REDUCE_CONV);;

let dismont = prove
 (`((&2 pow 384 * x) rem &p_384 + (&2 pow 384 * y) rem &p_384) rem &p_384 =
   (&2 pow 384 * (x + y)) rem &p_384 /\
   ((&2 pow 384 * x) rem &p_384 - (&2 pow 384 * y) rem &p_384) rem &p_384 =
   (&2 pow 384 * (x - y)) rem &p_384 /\
  (&(NUMERAL n) * (&2 pow 384 * x) rem &p_384) rem &p_384 =
   (&2 pow 384 * (&(NUMERAL n) * x)) rem &p_384`,
  REPEAT CONJ_TAC THEN CONV_TAC INT_REM_DOWN_CONV THEN
  AP_THM_TAC THEN AP_TERM_TAC THEN INT_ARITH_TAC);;

let unmont = prove
 (`(&(inverse_mod p_384 (2 EXP 384)) * (&2 pow 384 * x) rem &p_384) rem &p_384 =
   x rem &p_384`,
  CONV_TAC INT_REM_DOWN_CONV THEN REWRITE_TAC[INT_REM_EQ] THEN
  MATCH_MP_TAC(INTEGER_RULE
   `(i * e:int == &1) (mod p) ==> (i * e * x == x) (mod p)`) THEN
  REWRITE_TAC[INT_OF_NUM_CLAUSES; GSYM num_congruent; INVERSE_MOD_LMUL_EQ] THEN
  REWRITE_TAC[COPRIME_REXP; COPRIME_2; p_384] THEN CONV_TAC NUM_REDUCE_CONV);;

let unreplemma = prove
 (`!x. x < p_384
         ==> x =
             (2 EXP 384 * (inverse_mod p_384 (2 EXP 384) * x) MOD p_384) MOD
             p_384`,
  REPEAT STRIP_TAC THEN CONV_TAC SYM_CONV THEN
  ASM_REWRITE_TAC[MOD_UNIQUE] THEN
  REWRITE_TAC[CONG] THEN CONV_TAC MOD_DOWN_CONV THEN
  REWRITE_TAC[GSYM CONG] THEN MATCH_MP_TAC(NUMBER_RULE
   `(i * e == 1) (mod p) ==> (i * e * x == x) (mod p)`) THEN
  REWRITE_TAC[INVERSE_MOD_RMUL_EQ] THEN
  REWRITE_TAC[COPRIME_REXP; COPRIME_2; p_384] THEN CONV_TAC NUM_REDUCE_CONV);;

let weierstrass_of_jacobian_p384_add = prove
 (`!P1 P2 x1 y1 z1 x2 y2 z2 x3 y3 z3.
        ~(weierstrass_of_jacobian (integer_mod_ring p_384)
           (x1 rem &p_384,y1 rem &p_384,z1 rem &p_384) =
          weierstrass_of_jacobian (integer_mod_ring p_384)
           (x2 rem &p_384,y2 rem &p_384,z2 rem &p_384)) /\
        jacobian_add_unexceptional nistp384
         (x1 rem &p_384,y1 rem &p_384,z1 rem &p_384)
         (x2 rem &p_384,y2 rem &p_384,z2 rem &p_384) =
        (x3 rem &p_384,y3 rem &p_384,z3 rem &p_384)
        ==> weierstrass_of_jacobian (integer_mod_ring p_384)
                (x1 rem &p_384,y1 rem &p_384,z1 rem &p_384) = P1 /\
            weierstrass_of_jacobian (integer_mod_ring p_384)
                (x2 rem &p_384,y2 rem &p_384,z2 rem &p_384) = P2
            ==> weierstrass_of_jacobian (integer_mod_ring p_384)
                  (x3 rem &p_384,y3 rem &p_384,z3 rem &p_384) =
                group_mul p384_group P1 P2`,
  REPEAT GEN_TAC THEN
  DISCH_THEN(CONJUNCTS_THEN2 ASSUME_TAC (SUBST1_TAC o SYM)) THEN
  DISCH_THEN(CONJUNCTS_THEN(SUBST1_TAC o SYM)) THEN
  REWRITE_TAC[nistp384; P384_GROUP] THEN
  MATCH_MP_TAC WEIERSTRASS_OF_JACOBIAN_ADD_UNEXCEPTIONAL THEN
  REWRITE_TAC[CONJ_ASSOC] THEN CONJ_TAC THENL
   [ALL_TAC;
    W(MP_TAC o PART_MATCH (rand o rand) WEIERSTRASS_OF_JACOBIAN_EQ o
      rand o snd) THEN
    ASM_REWRITE_TAC[] THEN DISCH_THEN MATCH_MP_TAC] THEN
  ASM_REWRITE_TAC[FIELD_INTEGER_MOD_RING; PRIME_P384] THEN
  ASM_REWRITE_TAC[jacobian_point; INTEGER_MOD_RING_CHAR;
                  INTEGER_MOD_RING_CLAUSES; IN_INTEGER_MOD_RING_CARRIER] THEN
  REWRITE_TAC[INT_REM_POS_EQ; INT_LT_REM_EQ; GSYM INT_OF_NUM_CLAUSES] THEN
  REWRITE_TAC[p_384; b_384] THEN CONV_TAC INT_REDUCE_CONV);;

let represents_p384 = new_definition
 `represents_p384 P (x,y,z) <=>
        x < p_384 /\ y < p_384 /\ z < p_384 /\
        weierstrass_of_jacobian (integer_mod_ring p_384)
         (tripled (montgomery_decode (384,p_384)) (x,y,z)) = P`;;

let P384_MONTJADD_ALT_CORRECT = time prove
 (`!p3 p1 t1 p2 t2 pc stackpointer.
        aligned 16 stackpointer /\
        ALL (nonoverlapping (stackpointer,336))
            [(word pc,0x4480); (p1,144); (p2,144); (p3,144)] /\
        nonoverlapping (p3,144) (word pc,0x4480)
        ==> ensures arm
             (\s. aligned_bytes_loaded s (word pc) p384_montjadd_alt_mc /\
                  read PC s = word(pc + 0x14) /\
                  read SP s = stackpointer /\
                  C_ARGUMENTS [p3; p1; p2] s /\
                  bignum_triple_from_memory (p1,6) s = t1 /\
                  bignum_triple_from_memory (p2,6) s = t2)
             (\s. read PC s = word (pc + 0x4468) /\
                  !P1 P2. represents_p384 P1 t1 /\ represents_p384 P2 t2 /\
                          (P1 = P2 ==> P2 = NONE)
                          ==> represents_p384 (group_mul p384_group P1 P2)
                               (bignum_triple_from_memory(p3,6) s))
          (MAYCHANGE [PC; X0; X1; X2; X3; X4; X5; X6; X7; X8; X9; X10;
                      X11; X12; X13; X14; X15; X16; X17; X19; X20;
                      X21; X22; X23; X24; X25; X26] ,,
           MAYCHANGE SOME_FLAGS ,,
           MAYCHANGE [memory :> bytes(p3,144);
                      memory :> bytes(stackpointer,336)])`,
  REWRITE_TAC[FORALL_PAIR_THM] THEN
  MAP_EVERY X_GEN_TAC
   [`p3:int64`; `p1:int64`; `x1:num`; `y1:num`; `z1:num`; `p2:int64`;
    `x2:num`; `y2:num`; `z2:num`; `pc:num`; `stackpointer:int64`] THEN
  REWRITE_TAC[ALLPAIRS; ALL; NONOVERLAPPING_CLAUSES] THEN STRIP_TAC THEN
  REWRITE_TAC[C_ARGUMENTS; SOME_FLAGS; PAIR_EQ; bignum_triple_from_memory] THEN
  CONV_TAC(ONCE_DEPTH_CONV NUM_MULT_CONV) THEN
  CONV_TAC(ONCE_DEPTH_CONV NORMALIZE_RELATIVE_ADDRESS_CONV) THEN
  REWRITE_TAC[BIGNUM_FROM_MEMORY_BYTES] THEN ENSURES_INIT_TAC "s0" THEN

  LOCAL_AMONTSQR_P384_TAC 3 ["z1sq";"z_1"] THEN
  LOCAL_AMONTSQR_P384_TAC 0 ["z2sq";"z_2"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["y1a";"z_2";"y_1"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["y2a";"z_1";"y_2"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["x2a";"z1sq";"x_2"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["x1a";"z2sq";"x_1"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["y2a";"z1sq";"y2a"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["y1a";"z2sq";"y1a"] THEN
  LOCAL_SUB_P384_TAC 0 ["xd";"x2a";"x1a"] THEN
  LOCAL_SUB_P384_TAC 0 ["yd";"y2a";"y1a"] THEN
  LOCAL_AMONTSQR_P384_TAC 0 ["zz";"xd"] THEN
  LOCAL_MONTSQR_P384_TAC 0 ["ww";"yd"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["zzx1";"zz";"x1a"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["zzx2";"zz";"x2a"] THEN
  LOCAL_SUB_P384_TAC 0 ["resx";"ww";"zzx1"] THEN
  LOCAL_SUB_P384_TAC 0 ["t1";"zzx2";"zzx1"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["xd";"xd";"z_1"] THEN
  LOCAL_SUB_P384_TAC 0 ["resx";"resx";"zzx2"] THEN
  LOCAL_SUB_P384_TAC 0 ["t2";"zzx1";"resx"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["t1";"t1";"y1a"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["resz";"xd";"z_2"] THEN
  LOCAL_MONTMUL_P384_TAC 0 ["t2";"yd";"t2"] THEN
  LOCAL_SUB_P384_TAC 0 ["resy";"t2";"t1"] THEN

  BIGNUM_LDIGITIZE_TAC "x1_"
   `read (memory :> bytes (p1,8 * 6)) s26` THEN
  BIGNUM_LDIGITIZE_TAC "y1_"
   `read (memory :> bytes (word_add p1 (word 48),8 * 6)) s26` THEN
  BIGNUM_LDIGITIZE_TAC "z1_"
   `read (memory :> bytes (word_add p1 (word 96),8 * 6)) s26` THEN
  BIGNUM_LDIGITIZE_TAC "x2_"
   `read (memory :> bytes (p2,8 * 6)) s26` THEN
  BIGNUM_LDIGITIZE_TAC "y2_"
   `read (memory :> bytes (word_add p2 (word 48),8 * 6)) s26` THEN
  BIGNUM_LDIGITIZE_TAC "z2_"
   `read (memory :> bytes (word_add p2 (word 96),8 * 6)) s26` THEN
  BIGNUM_LDIGITIZE_TAC "resx_"
   `read (memory :> bytes (stackpointer,8 * 6)) s26` THEN
  BIGNUM_LDIGITIZE_TAC "resy_"
   `read (memory :> bytes (word_add stackpointer (word 192),8 * 6)) s26` THEN
  BIGNUM_LDIGITIZE_TAC "resz_"
   `read (memory :> bytes (word_add stackpointer (word 240),8 * 6)) s26` THEN
  ARM_STEPS_TAC P384_MONTJADD_ALT_EXEC (27--113) THEN
  CONV_TAC(ONCE_DEPTH_CONV BIGNUM_LEXPAND_CONV) THEN
  ENSURES_FINAL_STATE_TAC THEN ASM_REWRITE_TAC[] THEN
  DISCARD_STATE_TAC "s113" THEN
  DISCARD_MATCHING_ASSUMPTIONS [`nonoverlapping_modulo a b c`] THEN

  MAP_EVERY X_GEN_TAC [`P1:(int#int)option`; `P2:(int#int)option`] THEN
  REWRITE_TAC[represents_p384; tripled] THEN
  REWRITE_TAC[montgomery_decode; INT_OF_NUM_CLAUSES; INT_OF_NUM_REM] THEN
  STRIP_TAC THEN
  REPEAT(FIRST_X_ASSUM(MP_TAC o check (is_imp o concl))) THEN
  REPEAT(ANTS_TAC THENL
   [REWRITE_TAC[p_384] THEN RULE_ASSUM_TAC(REWRITE_RULE[p_384]) THEN
    CONV_TAC NUM_REDUCE_CONV THEN ASM BOUNDER_TAC[];
    (DISCH_THEN(STRIP_ASSUME_TAC o MATCH_MP unilemma0) ORELSE
     DISCH_THEN(STRIP_ASSUME_TAC o MATCH_MP unilemma1) ORELSE
     STRIP_TAC)]) THEN

  REWRITE_TAC[WORD_SUB_0; VAL_EQ_0; INT_OF_NUM_EQ; WORD_OR_EQ_0] THEN
  REWRITE_TAC[GSYM CONJ_ASSOC] THEN
  MP_TAC(GEN_ALL(SPEC `[x0:int64;x1;x2;x3;x4;x5]`
    BIGNUM_OF_WORDLIST_EQ_0)) THEN
  ASM_REWRITE_TAC[ALL; GSYM INT_OF_NUM_EQ] THEN
  DISCH_THEN(fun th -> REWRITE_TAC[GSYM th]) THEN ASM_REWRITE_TAC[] THEN
  MAP_EVERY ASM_CASES_TAC [`&z1:int = &0`; `&z2:int = &0`] THEN
  ASM_REWRITE_TAC[] THEN CONV_TAC(DEPTH_CONV WORD_NUM_RED_CONV) THEN
  REWRITE_TAC[GSYM(NUM_REDUCE_CONV `2 EXP 384`)] THENL
   [ASM_REWRITE_TAC[] THEN MAP_EVERY EXPAND_TAC ["P1"; "P2"] THEN
    REWRITE_TAC[GSYM INT_OF_NUM_REM; GSYM INT_OF_NUM_CLAUSES] THEN
    ASM_REWRITE_TAC[INT_MUL_LZERO; INT_MUL_RZERO; INT_REM_ZERO;
                    GSYM INT_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[weierstrass_of_jacobian; INTEGER_MOD_RING_CLAUSES] THEN
    REWRITE_TAC[P384_GROUP; weierstrass_add];
    ASM_REWRITE_TAC[] THEN EXPAND_TAC "P1" THEN
    REWRITE_TAC[GSYM INT_OF_NUM_REM] THEN
    ASM_REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[INT_MUL_RZERO; INT_REM_ZERO] THEN
    REWRITE_TAC[weierstrass_of_jacobian; INTEGER_MOD_RING_CLAUSES] THEN
    REWRITE_TAC[P384_GROUP; weierstrass_add];
    ASM_REWRITE_TAC[] THEN EXPAND_TAC "P2" THEN
    REWRITE_TAC[GSYM INT_OF_NUM_REM] THEN
    ASM_REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES] THEN
    REWRITE_TAC[INT_MUL_RZERO; INT_REM_ZERO] THEN
    REWRITE_TAC[weierstrass_of_jacobian; INTEGER_MOD_RING_CLAUSES] THEN
    REWRITE_TAC[P384_GROUP; weierstrass_add];
    ALL_TAC] THEN

  ASM_REWRITE_TAC[] THEN
  MAP_EVERY (MP_TAC o C SPEC unreplemma)
   [`z2:num`; `y2:num`; `x2:num`; `z1:num`; `y1:num`; `x1:num`] THEN
  MAP_EVERY (fun t -> ABBREV_TAC t THEN POP_ASSUM(K ALL_TAC))
   [`x1d = inverse_mod p_384 (2 EXP 384) * x1`;
    `y1d = inverse_mod p_384 (2 EXP 384) * y1`;
    `z1d = inverse_mod p_384 (2 EXP 384) * z1`;
    `x2d = inverse_mod p_384 (2 EXP 384) * x2`;
    `y2d = inverse_mod p_384 (2 EXP 384) * y2`;
    `z2d = inverse_mod p_384 (2 EXP 384) * z2`] THEN
  ASM_REWRITE_TAC[] THEN REPEAT DISCH_TAC THEN
  REPEAT(FIRST_X_ASSUM(K ALL_TAC o GEN_REWRITE_RULE I [GSYM NOT_LE])) THEN
  RULE_ASSUM_TAC(REWRITE_RULE
   [num_congruent; GSYM INT_OF_NUM_CLAUSES; GSYM INT_OF_NUM_REM]) THEN
  RULE_ASSUM_TAC(REWRITE_RULE[GSYM INT_REM_EQ]) THEN
  RULE_ASSUM_TAC(CONV_RULE INT_REM_DOWN_CONV) THEN
  RULE_ASSUM_TAC(REWRITE_RULE[INT_POW_2]) THEN
  RULE_ASSUM_TAC(ONCE_REWRITE_RULE[GSYM INT_ADD_REM; GSYM INT_SUB_REM]) THEN
  RULE_ASSUM_TAC(ONCE_REWRITE_RULE[lemont; demont]) THEN
  ASM_REWRITE_TAC[GSYM INT_OF_NUM_CLAUSES; GSYM INT_OF_NUM_REM] THEN
  REWRITE_TAC[INT_REM_REM] THEN
  REWRITE_TAC[pumont; dismont; unmont] THEN
  REPEAT(FIRST_X_ASSUM(MP_TAC o
    check(can (term_match [] `weierstrass_of_jacobian f j = p`) o concl))) THEN
  REWRITE_TAC[IMP_IMP] THEN
  SUBGOAL_THEN `~(&z1d rem &p_384 = &0) /\ ~(&z2d rem &p_384 = &0)`
  STRIP_ASSUME_TAC THENL
   [CONJ_TAC THENL
     [UNDISCH_TAC `~(&z1:int = &0)`; UNDISCH_TAC `~(&z2:int = &0)`] THEN
    ASM_REWRITE_TAC[CONTRAPOS_THM] THEN
    REWRITE_TAC[INT_REM_EQ_0] THEN CONV_TAC INTEGER_RULE;
    ALL_TAC] THEN
  DISCH_THEN(fun th -> STRIP_ASSUME_TAC th THEN MP_TAC th) THEN
  MATCH_MP_TAC weierstrass_of_jacobian_p384_add THEN ASM_REWRITE_TAC[] THEN
  FIRST_X_ASSUM(MP_TAC o GEN_REWRITE_RULE I [GSYM CONTRAPOS_THM]) THEN
  ANTS_TAC THENL
   [EXPAND_TAC "P2" THEN REWRITE_TAC[weierstrass_of_jacobian] THEN
    ASM_REWRITE_TAC[INTEGER_MOD_RING_CLAUSES; OPTION_DISTINCT];
    DISCH_TAC] THEN
  ASM_REWRITE_TAC[jacobian_add_unexceptional; nistp384;
                  INTEGER_MOD_RING_CLAUSES] THEN
  CONV_TAC(TOP_DEPTH_CONV let_CONV) THEN REWRITE_TAC[PAIR_EQ] THEN
  CONV_TAC INT_REM_DOWN_CONV THEN
  REPEAT CONJ_TAC THEN AP_THM_TAC THEN AP_TERM_TAC THEN INT_ARITH_TAC);;

let P384_MONTJADD_ALT_SUBROUTINE_CORRECT = time prove
 (`!p3 p1 t1 p2 t2 pc stackpointer returnaddress.
        aligned 16 stackpointer /\
        ALL (nonoverlapping (word_sub stackpointer (word 400),400))
            [(word pc,0x4480); (p1,144); (p2,144); (p3,144)] /\
        nonoverlapping (p3,144) (word pc,0x4480)
        ==> ensures arm
             (\s. aligned_bytes_loaded s (word pc) p384_montjadd_alt_mc /\
                  read PC s = word pc /\
                  read SP s = stackpointer /\
                  read X30 s = returnaddress /\
                  C_ARGUMENTS [p3; p1; p2] s /\
                  bignum_triple_from_memory (p1,6) s = t1 /\
                  bignum_triple_from_memory (p2,6) s = t2)
             (\s. read PC s = returnaddress /\
                  !P1 P2. represents_p384 P1 t1 /\ represents_p384 P2 t2 /\
                          (P1 = P2 ==> P2 = NONE)
                          ==> represents_p384 (group_mul p384_group P1 P2)
                               (bignum_triple_from_memory(p3,6) s))
          (MAYCHANGE_REGS_AND_FLAGS_PERMITTED_BY_ABI ,,
           MAYCHANGE [memory :> bytes(p3,144);
                      memory :> bytes(word_sub stackpointer (word 400),400)])`,
  ARM_ADD_RETURN_STACK_TAC P384_MONTJADD_ALT_EXEC
   P384_MONTJADD_ALT_CORRECT
    `[X19; X20; X21; X22; X23; X24; X25; X26]` 400);;
